Require Import Arith.
Require Import NPeano.
Require Import Omega. 
Require Import Program.
Require Import List.
Require Classes.EquivDec.
Require Import LIFO.List.Length.
Require Import LIFO.List.Misc.
Require Import LIFO.List.InSig.
Require Import LIFO.List.Map.
Require Import LIFO.List.SeqNat.
Require Import LIFO.List.Mem.
Require Import LIFO.List.Assoc.
Require Import LIFO.Option.Option.
Require CertifiedBSML.Primitives.Specification.
Require CertifiedBSML.Primitives.Properties.
Require CertifiedBSML.Support.ParSig.
Require CertifiedBSML.StandardLibrary.Base.

Set Implicit Arguments.

Module Type TYPE
  (BsmlPrimitives: Primitives.Specification.PRIMITIVES)
  (PropertiesM : CertifiedBSML.Primitives.Properties.TYPE(BsmlPrimitives))
  (ParSigM : Support.ParSig.TYPE(BsmlPrimitives)(PropertiesM))
  (BaseM : CertifiedBSML.StandardLibrary.Base.TYPE(BsmlPrimitives)(PropertiesM)(ParSigM)).

  Import BsmlPrimitives PropertiesM ParSigM BaseM.

Require Import Coq.Numbers.Natural.Peano.NPeano.

Lemma  p_mod : forall dec x y , 
  y<>0 ->
  x < y ->
  x =(((y - dec mod y + x) mod y + dec) mod y).
Proof.
  intros dec x y H H0.
  rewrite (add_mod_idemp_l _ _ _ H).
    rewrite <- (mod_small x y H0).
    rewrite  <- (add_mod_idemp_r _ _ _ H).
    replace  (y - dec mod y + x mod y + dec mod y)  with   
      (y - dec mod y  + dec mod y + x mod y).
    replace    (y - dec mod y  + dec mod y) with (y) .
    rewrite <- (add_mod_idemp_l _ _ _ H).
    rewrite  (mod_same _ H).
    rewrite  (add_mod_idemp_r _ _ _ H).
    reflexivity.
    pose (NDivMod.mod_upper_bound dec y H).
    omega.
    rewrite (AlgebraClasses.associative) .
    rewrite (AlgebraClasses.commutative (dec mod y )) .
    rewrite <- (AlgebraClasses.associative) .
    reflexivity.
  Qed.

  Program Definition shift A dec (v: par A) : {vr:par A| forall i, get v i = get vr ( (i+dec) mod bsp_p) } :=
  let received := put (apply (mkpar (fun (i:processor) l (j:processor) => 
    if  (eq_nat_dec j (( i + dec) mod bsp_p)) then  Some l else None )) v)
    in parfun (@noSome _)  
    (parSig 
      (apply 
        (mkpar (fun (i:processor) (f : processor -> _) =>  
          f ((bsp_p - (dec mod bsp_p) + i) mod bsp_p)))  
        received) 
      (fun a => a <> None) _)  .
  Next Obligation. (* Proof that  (`i + dec) mod bsp_p < bsp_p *)
    intros A dec v vr i.        
    destruct i.  
    simpl.
    apply ND.mod_bound; auto with arith bsml.
  Qed.

  Next Obligation. (* Proof that  (bsp_p - dec mod bsp_p + `i) mod bsp_p < bsp_p *)
    intros A dec v received i f.
    destruct i.  
      simpl.
      apply ND.mod_bound; auto with arith bsml.
  Defined.

Next Obligation.                (* proof that at each processor, the received value we keep is not None , needed by ParSig *)
  intros A dec v received i.
  autorewrite with bsml bsmlbase.
    destruct i.  
    simpl.
    destruct (eq_nat_dec x (((bsp_p - dec mod bsp_p + x) mod bsp_p + dec) mod bsp_p)).
    discriminate.
    pose (p_mod dec bsp_pNotZero l).
    contradiction.
Qed.
 
Next Obligation. (* proof of the specification  *)
  intros A dec v received i.
  match goal with [ |- _ = get ?V ?p]  => set (j:=p) end.
  match goal with [ |- _ = get (proj1_sig (parfun ?f ?V)) ?p]  => set (f':=f) ; set (V':=V) end.
  rewrite (lparfun f' V').
 
  
  assert( Some  (get v i) = Some  ( f' (get V' j))).
  subst f' V'.
  autorewrite with  par_sig.
  
  rewrite someNoSomeId.
  subst received.
  autorewrite with bsml.
  simpl.
  (* preparing cases *)
  destruct i as [x l].
  assert ( x = (bsp_p - dec mod bsp_p + (x + dec) mod bsp_p) mod bsp_p).
    rewrite  (add_mod_idemp_r _ _ _ bsp_pNotZero).
    rewrite <- (AlgebraClasses.associative ) .
    rewrite <-  (add_mod_idemp_l _ _ _ bsp_pNotZero).
    rewrite <- (p_mod dec bsp_pNotZero l).
    reflexivity.  
    simpl.
    destruct ( eq_nat_dec ((x + dec) mod bsp_p)
         (((bsp_p - dec mod bsp_p + (x + dec) mod bsp_p) mod bsp_p + dec)
          mod bsp_p)).
  f_equal.
  f_equal.
  apply processor_eq. 
  simpl.      
  auto.
  rewrite <- H in n;  elim n; reflexivity.
  injection H; auto.
Qed.

Implicit Arguments shift [A].

Program Definition shift_right A (v: par A) := shift 1 v.

Program Definition shift_left A (v: par A)  := shift (bsp_p - 1) v.

(* TODO: lemma shift_left property *)

(* PROBLEM: there is no specification for this function: this should
be the total_exchange function including in the real BSML standard
library *)

Lemma nthSeq: 
  forall (position start length : nat), 
    position < S length ->
    forall  default:nat, nth position (seq start (S length)) default = start + position.
Proof.
  induction position; intros start length H default; [
    trivial |
    simpl; destruct length; [ contradict H | rewrite IHposition]; omega].
Qed.

Program Definition total_exchange A (v: par A) :
  { res : par(list A) | forall (i j : processor), forall default,
    nth j (get res i) default = get v j } :=  
  let values := List.map (proj v) procs in replicate values.
Next Obligation.
  intros A v values i j default;
    unfold values;
      destruct j as [j Hj]; 
        autorewrite with bsml;
          rewrite map_nth' with (d':=firstProcessor);
            autorewrite with bsml.
  f_equal; simpl; apply nthInSigPI; 
    simpl; destruct bsp_pSMaxPid as [maxPid HmaxPid];
      rewrite HmaxPid; rewrite nthSeq; [ trivial | rewrite <- HmaxPid; assumption].
  simpl; unfold processor; autorewrite with length; trivial.
Qed.

End TYPE.

Module Make
  (BsmlPrimitives: Primitives.Specification.PRIMITIVES)
  (PropertiesM : CertifiedBSML.Primitives.Properties.TYPE(BsmlPrimitives))
  (ParSigM : Support.ParSig.TYPE(BsmlPrimitives)(PropertiesM))
  (BaseM : CertifiedBSML.StandardLibrary.Base.TYPE(BsmlPrimitives)(PropertiesM)(ParSigM)).
Include TYPE BsmlPrimitives PropertiesM ParSigM BaseM.
End Make.

(* *)
