open Bsml
open Stdlib.Base
open Stdlib.Comm

let heatEquationFormula dt dx gamma uI uIMinusOne uIPlusOne =
    gamma *. dt /. (dx *. dx) *. ( uIPlusOne +. uIMinusOne -. uI-. uI ) +. uI

let rec heatEquationSeq leftBound rightBound dt dx gamma (u : float list) : float list =
    match u with 
      | [] -> []
      | uI :: u'  -> 
        match u' with 
          | [] -> [ heatEquationFormula dt dx gamma uI leftBound rightBound ]
          | uIPlusOne:: _ ->
            (heatEquationFormula dt dx gamma uI leftBound uIPlusOne)::
            (heatEquationSeq uI rightBound dt dx gamma u')            

let first l = List.hd l

let last  l = List.hd(List.rev l)

let getBounds l_bound r_bound u =
  let msg = put << fun dst ->
    if dst=($this$+1) && $this$<>(bsp_p-1)
    then [last $u$] 
    else if dst=($this$-1) && $this$<>0
    then [first $u$] 
    else [] >> in
    << if $this$=0
       then l_bound 
       else List.hd ($msg$ ($this$-1)) >>,
    << if $this$=(bsp_p-1) 
       then r_bound 
       else List.hd ($msg$ ($this$+1)) >>
;;

let heat leftBound rightBound dt dx gamma (u : (float list) par ) : (float list) par =
  let leftBounds,rightBounds = getBounds leftBound rightBound u in 
  << heatEquationSeq $leftBounds$ $rightBounds$ dt dx gamma $u$ >>
;;
