open Specification
open Datatypes
open Flatten
open InSig
open EqNat
open List0
open MinMax
open Option
open Peano
open NPeano
open Peano_dec
open Heat1D
open NatTools
open BsmlNat.Primitives

module Number : NUMBER with type number = float = 
struct
  type number = float 
  let add = ( +. )
  let mul = ( *. )
  let sub = ( -. )
  let div = ( /. )
  let opp = ( ~-. )
end

module Processors = 
struct 
  type __ = Obj.t
  let __ = let rec f _ = Obj.repr f in Obj.repr f

  let max_pid =  nat_of_int  (Bsml.bsp_p - 1)
  let firstProcessor = nat_of_int 0
  let lastProcessor = nat_of_int (Bsml.bsp_p - 1)
  let processorOfNatMod n = modulo n (nat_of_int Bsml.bsp_p)
  let natOfProcessor p =p
  let processorOfNat i = i
  let processorEqDec p1 p2 = eq_nat_dec p1 p2
  let processor_beq p1 p2 = eq_nat_dec p1 p2
  let processors = List.map nat_of_int (Tools.from_to 0 (Bsml.bsp_p-1))
  let processorsUpTo i = List.map nat_of_int (Tools.from_to 0 (int_of_nat i))
  let processorsStartingAt i =  List.map nat_of_int (Tools.from_to (int_of_nat i) (Bsml.bsp_p-1))
  
  let processorsRevInduction_obligation_2 x x0 =
    let rec f n x1 =
      match n with
      | O -> x1
      | S n0 -> let sp = S n0 in f n0 (x n0 sp __ __ x1)
    in f lastProcessor x0

  let processorsRevInduction x x0 =
    processorsRevInduction_obligation_2 x x0

  let listOfParList plst =  failwith "listOfParList is for Specification purpose only" 
end

module ParSig = 
struct
  let parSig v = v 
  let parInSig lv = lv 
end

module Base = Base.Make(BsmlNat.Primitives)(Processors)(ParSig)

include Heat1D.Make(BsmlNat.Primitives)(Number)(Processors)(ParSig)(Base)

let heat (leftBound:float) (rightBound:float) (dt:float) (dx:float) (gamma:float) (u:(float list)par) = 
  heatEquationPar leftBound rightBound dt dx gamma u

open Stdlib.Base

let _ = 
  let iter = try int_of_string (Sys.argv.(1)) with _ -> 200 in
  let size = try int_of_string (Sys.argv.(2)) with _ -> 10000 in
  let leftBound = 20. 
  and rightBound = 30. 
  and gamma = 0.1
  and f x = (Random.float 60.) -. 30.
  and dx = 10. /. (float size) in
  let dt = (dx*.dx)/.(10.*.gamma) in
  let u = mkpar(fun pid -> List.map f (Tools.from_to (if pid < nat_of_int (size mod Bsml.bsp_p) then 0 else 1) (size/Bsml.bsp_p))) in
  begin
    Gc.set { (Gc.get ()) with Gc.minor_heap_size = 1024*1024*1024*100};
    let start_minor_collection_vector= << let stat =Gc.stat() in stat.Gc.minor_collections>> in
    let gc_cost= << ref 0. >> in 
    let cost = << ref 0. >> in
			  Bsml.start_timing();
			  let u' = ref u in 
			  for i=1 to iter do
			    u' := heat leftBound rightBound dt dx gamma !u';
			    Bsml.stop_timing();
			    let cost_vec=Bsml.get_cost() in
			    ignore(<< $cost$:=( $cost_vec$ +. ! ($cost$)) >>);
			    Bsml.start_timing();
			    Gc.full_major();
			    Bsml.stop_timing();
			    ignore( << $gc_cost$ := $Bsml.get_cost()$ +.  ! $gc_cost$>>);
			    Bsml.start_timing();
			  done;
			  Bsml.stop_timing();
			  ignore(<< $cost$:=( $Bsml.get_cost()$ +. ! ($cost$)) >>);
			  ignore (<< if ( $this$ = 0 ) then (print_float ! $cost$ ; print_string ";")else () >> );
			  let word_vector= << let stat =Gc.stat() in stat.Gc.minor_words +.  stat.Gc.major_words -. stat.Gc.promoted_words>> in
			  let minor_collection_vector= << let stat =Gc.stat() in stat.Gc.minor_collections>> in
			  ignore (<< if ( $this$ = 0 ) then ( print_float $word_vector$; print_string ";";print_int ($minor_collection_vector$ - $start_minor_collection_vector$); print_string ";") else () >> );
			  Bsml.start_timing();
			  Gc.full_major();
			  Bsml.stop_timing();
			    ignore( << $gc_cost$ := $Bsml.get_cost()$ +.  ! $gc_cost$>>);
			    ignore (<< if ( $this$ = 0 ) then ( print_float ! $gc_cost$ )  else () >> )
  end   


