Require Import Coq.Lists.List.
Require Import ZArith.
Require Import Omega.
Require Import LIFO.Option.Option.
Require Import LIFO.List.Notations.

Set Implicit Arguments.
(** Option version of [hd] 
   The std lib has hd_error but  we prefer the notation Option, more common to Ocaml programmer,
   this is just a Notation for hd_error so that they are perfectly convertible (error | value a) are juste notation for None | Some
   
   *)
Global Notation hd_option := hd_error.

(** [strongHead] returns the head of a non-empty list together with the proof that it returns the same value as the [List.head] function.  **)

Program Definition strongHead (A:Type)(l:list A | l<>nil) : 
    { a : A | Some a = head l } := 
    noSome(head l).
Next Obligation.
  induction l.
    elim H; trivial.
    discriminate.
Defined.

(** [sHead] returns the head of a non-empty list. *)

Definition sHead (A:Type)(l:list A | l<>nil) : A := proj1_sig (strongHead l).

Lemma sHeadProp:
  forall (A:Type)(l:list A | l<>nil),
    Some( sHead l ) = hd_option (proj1_sig l).
Proof.
  intros A l; unfold sHead, strongHead; apply someNoSomeId.
Qed.

(** Facts about strandard [hd]  *)
Lemma hd_hd_option (A:Type) :
  forall (d:A) l,
    hd d l = (match hd_option l with None => d | Some a => a end).
  destruct l;
  reflexivity.
Qed.
  
(* *)