Require Import List.
Require Import Min.
Require Import Arith.

Require Import LIFO.List.Firstn_skipn.

Section list_partitioning.
  Variable A:Type.
  Set Implicit Arguments.
  (** List partitioning, 
     example : [list_part [x0,x1,x2,x3,x4] 1 3 = [x1,x2,x3]] *) 
  Definition list_part (l: list A) (from to: nat) :=
    firstn (to - from) (skipn from l).
  Hint Unfold list_part.

(*
  Lemma list_part_sublist (l: list A) (from to: nat) : sub_list l (list_part l from to).
  Proof.
    intros. unfold sub_list.
    exists (firstn from l); exists (skipn (to - from) (skipn from l)).
    unfold list_part.
    rewrite firstn_skipn. apply firstn_skipn.
  Qed.
*)

  (** merging/splitting of/into list subpart*) 
  Lemma list_part_merge (l: list A) (i1 i2 i3: nat) :
    i1 <= i2 <= i3 /\ i2 <= length l ->
    list_part l i1 i2 ++ list_part l i2 i3 = list_part l i1 i3.
  Proof.
    intros.
    destruct H.
    unfold list_part.
    replace (skipn i2 l) with (skipn (i2 - i1) (skipn i1 l)).
    set (skl1 := skipn i1 l).
    rewrite firstn_skipn_inv.
    replace (i2 - i1 + (i3 - i2)) with (i3 - i1).
    replace (firstn (i2 - i1) skl1) with (firstn (min (i2 - i1) (i3 - i1)) skl1).
    rewrite <- firstn_compose.
    apply firstn_skipn.
    rewrite min_l. reflexivity.
    omega. omega.
    unfold skl1. rewrite skipn_length. omega.
    replace (skipn i2 l) with (skipn ((i2 - i1) + i1) l).
    apply skipn_compose.
    f_equal.
    omega.
  Qed.    

  (** splitting into three sublists, with existantial quatification on right and left lists, and length of sublists *)
  Lemma list_part_sublist_size (l: list A) (from to: nat) (pre: from <= to <= length l):
    exists l1, exists l2, l1 ++ (list_part l from to) ++ l2 = l
      /\ length l1 = from /\ length l2 = (length l - to).
  Proof.
    intros.
    exists (firstn from l); exists (skipn (to - from) (skipn from l)).
    split.
    unfold list_part. rewrite firstn_skipn. apply firstn_skipn.
    split.
    apply firstn_length1. apply le_trans with (m := to). intuition. intuition.
    replace (skipn (to - from) (skipn from l)) with (skipn to l).
    apply skipn_length. intuition.
    rewrite skipn_compose.
    replace (to - from + from) with to. reflexivity.
    rewrite plus_comm.
    apply le_plus_minus.
    intuition.
  Qed.

  (** splitting/merging into/from three sublists *)
  Theorem merge_parts (l: list A) (from to: nat) (pre: from <= to <= length l):
    (list_part l 0 from)++(list_part l from to)++(list_part l to (length l)) = l.
  Proof.
    intros.
    unfold list_part at 1. simpl skipn at 1. rewrite <- minus_n_O.
    unfold list_part at 2. rewrite firstn_skipn_inv. rewrite <- le_plus_minus.
    unfold list_part.
    replace (firstn (length l) l) with l.
    replace (skipn to l) with (skipn (to - from) (skipn from l)).
    repeat (rewrite firstn_skipn). reflexivity.
    rewrite skipn_compose. rewrite plus_comm. rewrite <- le_plus_minus. reflexivity.
    intuition.
    symmetry. apply firstlength.
    intuition. intuition.
  Qed.

End list_partitioning.
