Require Import List.
Require Import Arith.
(* Require Import Relation_Definitions.
Require Import Setoid.
Require Import Sorting.PermutEq.
Require Import Sorting.Sorting.
Require Import Sorting.Permutation.
Require Import Sorting.PermutSetoid. *)
Require Import Orders.
Require Import Mergesort.
Require Import Program.

Require Import LIFO.List.InsertSorting.
Require Import LIFO.List.InSig.
Require Import LIFO.List.Nth.
Require Import LIFO.List.StrongNth.
Require Import LIFO.List.SeqNat.
Require Import LIFO.List.Map.
Require Import LIFO.List.Sorted.
Require Import LIFO.Nat.BoundedNat.
Require Import LIFO.Nat.ProofEquality.
Require Import LIFO.List.Length.
Require Import LIFO.List.Merge.
Require Import Permutation.

Set Implicit Arguments.

Hint Resolve NoDup_nil NoDup_cons NoDup_remove_1 NoDup_remove_2 noDupInj noDupFun noDupSeq :noDup.

Module Type PermuteType (T:Typ).
  
  Parameter permute :forall l : list T.t, {f : boundedNat (length l) -> boundedNat (length l) | bijective f} -> list T.t.

  Parameter permuteLength:     
    forall (l:list T.t)(f: boundedNat (length l) -> boundedNat(length l) | bijective f),
      length(permute l f) = length l.

  Parameter  permuteProperty:
    forall (l:list T.t)(f: boundedNat (length l) -> boundedNat(length l) | bijective f)
     (position: boundedNat (length l)),
      let position' := proj1_sig(boundedNatConversion position (eq_sym(permuteLength l f))) in
      proj1_sig (strongNth l position) = proj1_sig (strongNth (permute l f) position').

Parameter permuteWeakProperty:  forall l  (f: {f : boundedNat (length l) -> boundedNat (length l) | bijective f}), Permutation l (permute l f).


  (** permutation on list of element coupled to their destination position   *)
  Parameter permuteIndexedList : list (nat * T.t) -> list T.t.
  
  Parameter permuteIndexedListLength: 
    forall l,
      length (permuteIndexedList l) = length l.


Parameter permuteIndexedListProperty:
    forall l,
      forall (position : boundedNat (length l)),
        let elemAtPosition:=  proj1_sig (strongNth l position) in
          forall (H:fst elemAtPosition <  length (permuteIndexedList l) ),
          snd elemAtPosition = proj1_sig (strongNth (permuteIndexedList l) ( buildBoundedNat H)).

Parameter permuteIndexedListWeakProperty:  forall l, Permutation (map snd l) (permuteIndexedList l).


End PermuteType.

Module Make(T:Typ) : PermuteType T.

  Module PositionOrder <: TotalLeBool.
    Definition t : Type := prod nat T.t.
    Definition leb (x y: t) := NatOrder.leb (fst x) (fst y).
    Local Coercion is_true : bool >-> Sortclass.
    Theorem leb_total: forall x y, leb x y \/ leb y x.
    Proof.
      intros x y; destruct x; destruct y; unfold leb; simpl; apply NatOrder.leb_total.
    Qed.
  End PositionOrder.

  Module Import PositionSort := Sort PositionOrder.

  Module Import MergeProperties := LIFO.List.Merge.Make PositionOrder.
  
  Definition permute 
    (l:list T.t)(f: boundedNat (length l) -> boundedNat(length l) | bijective f) :=
    List.map 
    snd
    (PositionSort.sort 
      (List.map (fun v:boundedNat (length l)*T.t => let (p,x) := v in (proj1_sig(proj1_sig f p),x))
        (List.combine (inSig _ (seq 0 (length l)) (inSeq0Lt (length  l))) l))).
  
  Lemma permuteLength: 
    forall (l:list T.t)(f: boundedNat (length l) -> boundedNat(length l) | bijective f),
      length(permute l f) = length l.
  Proof.
    intros l f; unfold permute; autorewrite with length; apply Min.min_idempotent.
  Qed.
Hint Rewrite permuteLength : length.

  Definition permuteProperty:
    forall (l:list T.t)(f: boundedNat (length l) -> boundedNat(length l) | bijective f)
     (position: boundedNat (length l)),
      let position' := proj1_sig(boundedNatConversion position (eq_sym(permuteLength l f))) in
      proj1_sig (strongNth l position) = proj1_sig (strongNth (permute l f) position').
  Proof.
    admit.
  Qed.

  Definition permuteWeakProperty :  forall l  (f: {f : boundedNat (length l) -> boundedNat (length l) | bijective f}), Permutation l (permute l f).
  Proof.
    admit.
  Qed.

  (** permutation on list of element coupled to their destination position   *)
  Definition permuteIndexedList
    (l:list (nat*T.t))  :=
    List.map 
    snd
    (PositionSort.sort l).
  
  Lemma permuteIndexedListLength: 
    forall l,
      length (permuteIndexedList l) = length l.
  Proof.
    intros l; unfold permuteIndexedList; autorewrite with length; reflexivity.
  Qed.

  Hint Rewrite permuteIndexedListLength : length.

Require Import Permutation.
Definition permuteIndexedListProperty:
    forall l,
      forall (position : boundedNat (length l)),
        let elemAtPosition:=  proj1_sig (strongNth l position) in
          forall (H:fst elemAtPosition <  length (permuteIndexedList l) ),
          snd elemAtPosition = proj1_sig (strongNth (permuteIndexedList l) ( buildBoundedNat H)).
  Proof.
    admit.
  Qed.
Parameter permuteIndexedListWeakProperty:  forall l, Permutation (map snd l) (permuteIndexedList l).


End Make.

(*

(**
 Same as inSigMapProj1Sig using projT1
*)
Lemma inSigMapProjT1 :  forall (A : Type) (P : A -> Prop) (l : list A)
    (H : forall a : A, In a l -> P a), map (fun x => projT1 (sigT_of_sig x)) (@inSig A P l H) = l.
Proof.
induction l ; intros.
 intuition.
 simpl. rewrite IHl. reflexivity.
Qed.

Lemma sigT_of_sig_sig_of_sigT : forall A H (n:A|H), sigT_of_sig (sig_of_sigT n) = n.
Proof.
destruct n.
simpl.
intuition.
Qed.

Lemma projT1InSig : forall len (n:boundedNat len) x, 
  (nth (projT1 n) (inSig (fun i : nat => i < len) (seq 0 len) (inSeq0Lt len)) x) = n.
Proof.
intros.
destruct n.
destruct x.
assert (nth x0 (seq 0 len) x = x0). apply seq_nth. assumption. 
simpl projT1.
pose (P := fun n : nat => n < len).
fold P.
apply nthInSigPI.
intuition.
Qed.

(**
 le_refl does not exist with this form
*)
Lemma le_refl1 : forall n m, n <= m -> m <= n -> n = m.
Proof.
intuition.
Qed.

Section PermuteLemmas.
Variable A:Type.

(**
  To me, this is probably difficult to reuse the lemmas of this section
*)
Lemma positionsInsig : forall m l (n : boundedNat m) c (positions:=@combine (boundedNat m) A
              (@inSig nat (fun i : nat => i < m)
                 (seq 0 m) (inSeq0Lt m))),
              m <= length l ->  l <> nil -> projT1 (fst (nth (projT1 n) (positions l) c)) = projT1 n.
Proof.
intros.
unfold positions.
rewrite <- map_nth. rewrite map_fst_combine.
rewrite <- (map_nth _ _ (fst c) _). rewrite <- map_nth.
rewrite map_map. 
destruct n ; simpl.
unfold boundedNat. rewrite inSigMapProjT1. rewrite seq_nth ; intuition.
unfold boundedNat. rewrite inSigLength. rewrite seq_length. intuition.
Qed. 

Lemma permute_lemma1 : forall m (f:boundedNat m -> boundedNat m) (Hbij : bijective f) (n:boundedNat m) c l L 
(positions:=@combine (boundedNat m) A
              (@inSig nat (fun i : nat => i < m)
                 (seq 0 m) (inSeq0Lt m)))
(sortinglist := map
                 (fun v : (boundedNat m) * A =>
                  let (p, x) := v in (f p, x)) (positions l)), l <> nil -> m <= length l  ->
(map (fun x : (boundedNat m) * A => projT1 (fst x)) L) = 
(map (fun x : (boundedNat m) * A => projT1 (fst x)) (positions l)) ->
(@eqnA A m) (nth (projT1 n) sortinglist c) (nth (projT1 (f n)) L c).
Proof. 
intros.
assert (H2 : forall (n:boundedNat m), projT1 (fst (nth (projT1 n) (positions l) c)) = projT1 n).
intros. apply positionsInsig ; intuition.
assert (forall (n:boundedNat m), projT1 (fst (nth (projT1 n) (positions l) c)) = projT1 (fst (nth (projT1 n) L c))).
intros n0. do 2 rewrite <- map_nth ; symmetry ; do 2 rewrite <- map_nth.
rewrite <- map_nth. rewrite <- (map_nth _ _ (sigT_of_sig (fst c)) _). 
do 4 rewrite map_map. rewrite H1. reflexivity.

unfold eqnA.
rewrite <- H3 with (f n). rewrite H2.
unfold sortinglist. destruct c. 
assert (exists n, f n = b). apply Hbij. destruct H4.
rewrite <- map_nth. rewrite map_map.
assert (forall x, (fun x0 : boundedNat m * A => fst (let (p, x1) := x0 in (f p, x1))) x
  = (fun x => f (fst x)) x). intuition.
erewrite map_ext ; [ | apply H5].
rewrite <- map_map. simpl. rewrite <- H4. rewrite map_nth.
unfold positions. unfold boundedNat. rewrite map_fst_combine ; [|rewrite inSigLength ; rewrite seq_length ; assumption].
rewrite projT1InSig.
reflexivity.
Qed.

Lemma proj_nth1_compatibility : 
forall (A B : Type) L1 L2 n fn, Permutation L1 L2 -> n < length L1 ->
  (forall c1, nth n (map (@fst A B) L1) (@fst A B c1) = nth fn (map (@fst A B) L2) (fst c1)) ->
      NoDup (map (@fst A B) L2) -> 
        (forall a0 : A * B, exists b : A * B, fst a0 <> fst b) -> 
          (forall c1, nth n L1 c1 = nth fn L2 c1).
Proof.
intros.
apply inMapNoDupEqual1. apply Permutation_in with L1. apply H.
apply nth_In. apply H0. apply H2.
intros. rewrite<- H1. rewrite map_nth.
rewrite (@nth_indep _ L1 n c1 c). reflexivity.
assumption.
assumption.
Qed.

Lemma projNthCompatibility : 
forall (A B : Type) l1 l2 n fn L1 L2, l1 = (map (@snd A B) L1) -> l2 = (map (@snd A B) L2) -> 
  (forall c2, nth n (map (@fst A B) L1) c2 = nth fn (map (@fst A B) L2) c2) ->
    Permutation L1 L2 -> Permutation l1 l2 -> NoDup (map (@fst A B) L2) -> 
      (forall a0 : A * B, exists b : A * B, fst a0 <> fst b) -> n < length L1 ->
        (forall c1, nth n l1 (@snd A B c1) = nth fn l2 (snd c1)).
Proof.
intros.
assert (nth n L1 (c1) = nth fn L2 (c1)).
apply (proj_nth1_compatibility ) ; intuition.
subst.
rewrite map_nth. rewrite H7. symmetry. rewrite map_nth. reflexivity.
Qed.

End PermuteLemmas.

Section Permute.
Variable A : Type.

(**
 permuted asserts that l2 is the permutation of l1 and f 
 is the function of permutation
*)
Definition permuted (l1 l2 : list A) (f:boundedNat (length l1)->boundedNat (length l1)) : Prop := 
 (forall (a : A) (n:boundedNat (length l1)), 

(**
  Injectivity conserves NoDup
*)

Definition locallyInjective  ( B : Type)  (f: A-> B)  l := forall x y : A,In x l-> In y l ->  f x = f y -> x = y.
Implicit Arguments locallyInjective [B].

Lemma noDupLocallyInjective : forall ( B : Type) l (f : A -> B) (f_loc_inj :locallyInjective f l), 
 NoDup l -> NoDup (map f l).
Proof.
  intros B l f f_loc_inj H.
assert (length l = length (map f l)).  autorewrite  with length. reflexivity.
induction l.
 constructor.
 simpl map.
 inversion H.
 constructor. 
 intros SH.   
contradict H3. 
rewrite in_map_iff in SH. destruct SH as [a' [fx_fa in_l]]. 
assert  (a_a' :a = a'). apply f_loc_inj; simpl ;auto.
rewrite a_a'; assumption.
 apply IHl; auto.
 unfold locallyInjective in *;  intuition.
Qed.

Next Obligation.
 unfold listWithPositions.
 (* Notations*)
 pose (positions:=@combine {a : nat | a < @length A l} A
              (@inSig nat (fun i : nat => i < @length A l)
                 (seq 0 (@length A l)) (inSeq0Lt (@length A l)))).
 fold positions.
 pose (sortinglist:=(@map (boundedNat (length l) * A)
           (boundedNat (length l) * A)
           (fun v : boundedNat (length l) * A =>
            let (p, x) := v in (f p, x)) (positions l))).
 fold sortinglist.
 destruct (insertSorting
           (@SortNatA A (length l))
           sortinglist) as (L,LH).
 destruct LH as [permut [sorte  HL]].
 (* end Notations *)

(* New hypothesis : length l >= 2 *)
assert (Hpertinent : length l >= 2 \/ permuted l (map (@snd _ A) L) f).
destruct l. right. split. destruct n. inversion l.
simpl in *. rewrite HL. simpl. constructor.
destruct l. right.
split. intros.
assert (f n = n). destruct n. inversion l.
apply projT1BoundedNatInjective. simpl. rewrite H1.
destruct (f (exist (fun n : nat => n < 1) x l)). simpl. inversion l0. reflexivity.
inversion H2. inversion H1.
rewrite H0. rewrite HL. simpl. intuition.
rewrite HL. simpl. constructor. constructor.
left. simpl ; intuition.
rewrite HL in Hpertinent.
destruct Hpertinent as [Hpertinent|permuted1] ; [|assumption].
(* End of new hypohesis *)


(* Proofs that Permutation is true *)
assert (H5 : (map (@snd _ A) (positions l)) = l).
unfold positions. apply map_snd_combine. rewrite inSigLength.
apply seq_length.
assert (H55 : (map (@snd _ A) sortinglist) = (map (@snd _ A) (positions l))).
unfold sortinglist.  
clear HL sorte permut L H5.
induction (positions l). 
 reflexivity.
 simpl. rewrite IHl0. destruct a ; reflexivity.
assert (H2 : Permutation (map (@snd _ A) sortinglist) (map (@snd _ A) (positions l))).
apply permutationEq. apply H55.
assert (H66 : Permutation (map (@snd _ A) sortinglist) (map (@snd _ A) L)).
apply Permutation_map. apply permut.
assert (H3 : Permutation (map (@snd _ A) L) (map (@snd _ A) (positions l))).
eapply Permutation_trans. apply Permutation_sym. apply H66. apply H2.
assert (H0 : Permutation l (map (@snd _ A) L)).
apply Permutation_trans with (map (@snd _ A) (positions l)).
apply Permutation_sym. apply permutationEq.
apply H5. apply Permutation_sym. apply H3.
rewrite <- HL.
split ; [ | apply H0].
(* end of Proofs that Permutation is true *)

intros a n.
assert (H1 : map (fun x : boundedNat (length l) * A => projT1 (fst x)) L =
map (fun x : boundedNat (length l) * A => projT1 (fst x)) (positions l)).
(*Sorted (positions l) *)
assert (Sorted (@lenA A (length l)) (positions l)).
apply sortedLeLenA.
apply seqSorted. 
(* End Sorted (positions l) *)

(* assert GB : Permutation de la proj de L et de la proj de positions *)
assert (GB : Permutation (map (fun x => projT1 (sigT_of_sig (fst x))) L) 
(map (fun x => projT1 (sigT_of_sig (fst x))) (positions l))).
(* Demo Permutation (fst L) (positions l) *) 
apply Permutation_trans with (map (fun x => projT1 (sigT_of_sig (fst x))) sortinglist). 
 apply Permutation_map. apply Permutation_sym. apply permut.
 apply Permutation_trans with (seq 0 (length l)).
  unfold sortinglist. rewrite map_map.
  unfold positions.
  assert (forall x, ((fun x : boundedNat (length l) * A =>
      projT1 (sigT_of_sig (fst (let (p, x0) := x in (f p, x0)))))) x = 
      (fun x : boundedNat (length l) * A =>
      projT1 (sigT_of_sig (f (fst x)))) x). intuition.
  erewrite map_ext ; [ | apply H4].
  do 3 rewrite <- map_map. rewrite map_fst_combine.
  pose (in_sig := (inSig (fun i : nat => i < length l) (seq 0 (length l))
              (inSeq0Lt (length l)))). fold in_sig.
  assert (Permutation in_sig (map f in_sig)). apply permutationBij. 
   assumption.

 apply nodupInSigSeq. intros. destruct a0. apply inInSig1. apply inSeq. intuition.
  rewrite map_map. rewrite <- inSigSeq with (length l) 0 (fun i : nat => i < length l) (inSeq0Lt (length l)).
  apply Permutation_map. fold in_sig. apply Permutation_sym. apply H6.
  rewrite inSigLength. rewrite seq_length. constructor.
  
  unfold positions. do 2 rewrite <- map_map. rewrite map_fst_combine.
  rewrite map_map. rewrite inSigMapProjT1. apply permutationEq. reflexivity.
  rewrite inSigLength. rewrite seq_length. intuition.
(* Fin assert GB *)

symmetry. apply (sortedPermutationEq le_trans eq_nat_dec le_refl1).
 apply sortedLenALe. assumption.
 apply sortedLenALe. assumption.
 apply Permutation_sym. assumption.

(* equalities of nth l and L with the action of f *)
assert (H15 :forall c, (@eqnA A (length l) (nth (projT1 n) sortinglist c) (nth (projT1 (f n)) L c))).
intros c. apply permute_lemma1 ; intuition. destruct a0. rewrite H4 in l0. inversion l0.
(* end*)

unfold positions in H1. rewrite <- map_map in H1.
rewrite <- map_map in H1. rewrite <- map_map in H1.
rewrite <- map_map in H1. rewrite map_fst_combine in H1 ; [| rewrite inSigLength ; rewrite seq_length ; constructor].
rewrite map_map in H1. rewrite map_map in H1. rewrite map_map in H1.
rewrite inSigSeq in H1.

assert (H4 : NoDup (seq 0 (length l))) by auto with noDup. 
assert (H6 : NoDup (map (fun x : boundedNat (length l) * A => projT1 (fst x)) L)).
rewrite H1. assumption.
assert (l = map (@snd _ A) sortinglist). eapply eq_trans. symmetry. apply H5.
symmetry. apply H55.
intros.

assert (map (@snd _ A) L = map (@snd _ A) L) by reflexivity.
assert (forall c2, 
nth (projT1 n) (map (@fst _ A) sortinglist) c2 = nth (projT1 (f n)) (map (@fst _ A) L) c2).
unfold eqnA in H15. 
intros c2. generalize (H15 (c2, a)). intros H16.
apply projT1BoundedNatInjective in H16.
repeat rewrite <- map_nth.
repeat rewrite <- map_nth in H16.
apply H16.
assert (NoDup (map (@fst _ A) L)).

apply inSigNoDup. rewrite <- map_map. do 2 rewrite <- map_map in H1.
unfold boundedNat in *. rewrite H1. apply noDupSeq.
assert (forall a0:(boundedNat (length l)*A), exists b, fst a0 <> @fst _ A b). intros. destruct a0.
destruct b. destruct x.

assert (1 < length l). apply lt_le_trans with 2 ; intuition.


exists ((exist (fun n => n < length l) 1 H11), a). simpl.
intro H90. 

apply projT1BoundedNat in H90. inversion H90.
assert (0 < length l) by intuition.
exists ((exist (fun n => n < length l) 0 H11), a). intro H90 ; simpl in *.
apply projT1BoundedNat in H90.
inversion H90.
assert (projT1 n < length sortinglist). destruct n. simpl. apply lt_le_trans with (length l).
intuition. assert (length l = length sortinglist). rewrite <- (map_length (@snd _ A) sortinglist).
apply Permutation_length. apply Permutation_trans with (map (@snd _ A) L) ; intuition. intuition.
apply (@projNthCompatibility _ _ l (map (@snd _ A) L) (projT1 n) (projT1 (f n)) sortinglist L H7 H8 H9 permut H0 H10 H11 H12 (n,a)) ; intuition.
Qed.


End Permute.

(* Section Example. *)

(*  Definition f n := *)
(*  match n with *)
(*    | 0 => 1 *)
(*    | 1 => 2 *)
(*    | 2 => 0 *)
(*    | n => n *)
(*  end. *)

(* Lemma cases (n:boundedNat 4) : *)
(*  (projT1 n=0) \/ *)
(*  (projT1 n=1) \/ *)
(*  (projT1 n=2) \/ *)
(*  (projT1 n=3). *)
(* Proof. *)
(*  intros n. *)
(*  destruct n as [ n Hn] ; simpl. *)
(*  repeat(destruct n;intuition). *)
(* Qed. *)

(* Definition f' (n:boundedNat 4) : boundedNat 4. *)
(*  intro n. *)
(*  destruct n. *)
(*  exists (f x). *)
(*  repeat(destruct x ; intuition). *)
(* Defined. *)

(* Lemma bf : bijective f'. *)
(* Proof. *)
(*  unfold bijective, injective, surjective. *)
(*  split. *)
(*    intros x y H. *)
(*    set(Hx:=cases x). *)
(*    set(Hy:=cases y). *)
(*    destruct x ; destruct y ; *)
(*      repeat(destruct Hx);repeat(destruct Hy); intuition; *)
(*        simpl in * ; subst ; ( (rewrite (leUniquenessProof _ _ l0 l); auto) || inversion H). *)

(*    intro y; set(Hy:=cases y); repeat(destruct Hy) ; intuition ; set(P n:=n<4) ; *)
(*    assert(P 0) as H1 ; unfold P ; intuition ; *)
(*    assert(P 1) as H2 ; unfold P ; intuition ; *)
(*    assert(P 2) as H3 ; unfold P ; intuition ; *)
(*    assert(P 3) as H4 ; unfold P ; intuition. *)
(*    destruct y as (y, P1). simpl in H. subst.  *)
(*    exists (exist (fun n : nat => n < 4) 2 H3). simpl. *)
(*    replace P1 with (lt_0_Sn 3) by (apply leUniquenessProof). reflexivity. *)
(*    destruct y as (y, P1). simpl in H0. subst.  *)
(*    exists (exist (fun n : nat => n < 4) 0 H1). simpl. *)
(*    replace P1 with (lt_S 1 3 (lt_S 1 2 (lt_n_Sn 1))) by (apply leUniquenessProof). reflexivity. *)
(*    destruct y as (y, P1). simpl in H. subst.  *)
(*    exists (exist (fun n : nat => n < 4) 1 H2). simpl. *)
(*    replace P1 with (lt_S 2 3 (lt_n_Sn 2)) by (apply leUniquenessProof). reflexivity. *)
(*    destruct y as (y, P1). simpl in H. subst.  *)
(*    exists (exist (fun n : nat => n < 4) 3 H4). simpl. *)
(*    replace P1 with (H4) by (apply leUniquenessProof). reflexivity. *)
(* Qed. *)


(* Require Import String. *)
(* Open Scope string_scope. *)

(* (* *)
(* Eval compute in let (a, b) := (permute ("Bonjour"::"tout"::"le"::"monde"::nil) bf) in a.  *)
(* *) *)
(* End Example. *)

*)
