Require Import List.
Require Import LIFO.List.Notations.
Require Import LIFO.Algebra.AlgebraClasses.
Require Import LIFO.List.App.
Require Import LIFO.List.SeqNat.

Definition  prefix A  (ls : list A) :=
  map (fun n => firstn n ls) (seq 1 (length ls)).
Implicit Arguments prefix [A].

Example prefix_informal : prefix [1;2;3] =  [[1]; [1; 2]; [1; 2; 3]]. reflexivity. Qed.

Lemma prefix_app (A:Type) :
  forall (l l':list A), 
    prefix (l ++ l') = prefix l ++ map (app l) (prefix l').
Proof.
  intros l l'; revert l.
  induction l'.
  intros l'; simpl; repeat rewrite app_nil_r; reflexivity. 
  intros l.
  change ( a :: l') with ([a]++l').
  rewrite <-  (associative (eqA:=eq)) .
  rewrite IHl'.
  rewrite IHl'.
  rewrite map_app.
  rewrite <-  (associative (eqA:=eq)) .
  rewrite map_map.
  f_equal.
  simpl.      
  unfold prefix. 
  autorewrite with length.
  replace (seq 1 (length l + length [a])) with  (seq 1 (S(length l))).
  rewrite seq_S_app.
  rewrite map_app.
  f_equal.
  apply Map.map_assumption.
  intros x H. apply SeqNat.inSeqLt in H. apply   Firstn_skipn.firstn_app1. omega. 
  unfold map. f_equal. rewrite <-   Firstn_skipn.firstn_length_l. f_equal.
  autorewrite with length; simpl; omega.
  f_equal; autorewrite with length; simpl; omega.
  apply map_ext.
  intros a0.        
  apply associative.
Qed.


Lemma prefix_length (A:Type) :  forall (l:list A), length (prefix l) = length l.
Proof.
  intros l.
  unfold prefix. 
  autorewrite with length.
  reflexivity. 
Qed.

Definition  suffix A  (ls : list A) :=
  map (fun n => skipn n ls) (seq 0 (length ls)).
Implicit Arguments suffix [A].

Example suffix_informal : suffix [1;2;3] =   [[1; 2; 3];[2;3];[3]]. reflexivity. Qed.


Lemma suffix_length (A:Type) :  forall (l:list A), length (suffix l) = length l.
Proof.
  intros l.
  unfold suffix. 
  autorewrite with length.
  reflexivity. 
Qed.

(* *)