 Unset Automatic Introduction.    (* coq 8.3_rc1 *)
Require Import LIFO.Nat.Min.
Require Import Arith.

Require Import List.
Require Import LIFO.List.Length.
Require Import LIFO.List.Rev.
 Require Import LIFO.List.Nth.
 Require Import LIFO.List.Map.
 Require Import Tactics.
 
 Open Scope list_scope.

Generalizable All Variables.

Set Implicit Arguments.

Ltac lengthNil :=
  match goal with 
    | [lenl : 0 = length ?l|-_]=>
      symmetry in lenl; rewrite (lengthNil l lenl) in *
    | [lenr : length ?l=0|-_]=>
      rewrite (lengthNil l lenr) in *
  end. 

Section ListTriple.
  Variable  (L C R :Type).

(** decomposition of a triple of list which have the same length*)
Lemma decompListTriple :  
  forall (l1:list L) (c1 : list C) (r1:list R),
  length l1=length c1 ->
  length c1=length r1->
  {l1=nil/\c1=nil/\r1=nil}+{exists x, exists y, exists z,exists l2, exists c2, exists r2, (l1=x::l2/\c1=y::c2/\r1=z::r2)}.
Proof.
  intros  l1 c1 r1 H H0.
  destruct l1 ; destruct c1 ;destruct r1 ;  try(simpl in *; discriminate).
  left; tauto.
  right. exists l.
  exists c. 
  exists r. exists l1; exists c1; exists r1. split;[reflexivity|split;reflexivity].
Qed.


(** combine and append*)
Lemma combine_app :  forall (l1 l2:list L) (r1 r2: list R),
  length l1
=length r1 ->
  combine (l1++l2) (r1++r2) =   (combine (l1) (r1)) ++ (combine l2 r2).
Proof.
  intros l1 l2 r1 r2 len.
  generalize r1 len. clear len;induction l1.
  intros r3 lenr; simpl in lenr.
  lengthNil.
  simpl; reflexivity.
  intros r0 len0.
  destruct r0. 
    simpl in len0; discriminate.
    simpl. 
    rewrite (IHl1). reflexivity. simpl in len0.
    injection len0; intro;assumption. 
Qed.
   

(** Induction principle for resonning over three lists of same length*)  
Lemma list_triple_ind : forall (P : list L->list C->list R -> Prop),
  P nil nil nil ->
  (forall l c r ,
    (length l=length c ->
    length c=length r->
    P l c r ->
    forall x y z ,
    P (x::l) (y::c) (z::r) )
  )->
  (forall l c r ,
    length l=length c ->
    length c=length r->
    P l c r).
Proof.
  intros P P0 IH l c r lenlc lencr.
  generalize l c r lenlc lencr.
  clear l c r lencr lenlc.
  induction l; intros c0 r0 lenlc lencr.
    simpl in lenlc.
    lengthNil.
    simpl in lencr. 
    lengthNil.
    assumption.

(*     generalize a l c0 r0 (IHl c0 r0)  lenlc lencr; clear c0 r0 lenlc lencr  IHl l a.  *)
    destruct c0. (* ; intros r0 IHl lenlc lencr. *)
      simpl in lencr; lengthNil ; simpl in lenlc. discriminate.
(*       generalize r0 IHl (IHc0 r0) lenlc lencr; clear r0  lenlc lencr  IHc0 IHl.  *)
      destruct r0. (* ; intros IHl IHc0 lenlc lencr. *)
        simpl in lencr. discriminate.
        apply IH.
          simpl in lenlc;injection lenlc;intro;assumption.
          simpl in lencr;injection lencr;intro;assumption.
          apply (IHl c0 r0).
          simpl in lenlc;injection lenlc;intro;assumption.
          simpl in lencr;injection lencr;intro;assumption.
Qed.
End ListTriple.
Implicit Arguments decompListTriple [L C R l1 c1 r1].

Ltac listTripleDecomp lenlc lencr :=
  destruct (decompListTriple lenlc lencr) as [nil_lists | cons_lists].

(** *** Zipping 3 lists in one list of triple
   zip3 zip three lists in one list of triple which has the length of the smallest list
   -example :
   [zip3 [1,2,3,4] [f,g] [x,y,z] = [(1,f,x),(2,g,y)]]
 *)
Fixpoint zip3  (L C R :Type) (l : list L) (c:list C) (r : list R) {struct l} : list (L*C*R) :=
  match l,c,r with
    |a::l',b::c',c::r' => (a,b,c):: (zip3 l' c' r')
    |_,_,_=> nil
  end.

(** decomposition of [zip3] when splitting the 3 lists with first parts of same length*)
Lemma decompZip3  : forall L C R (l1 l2:list L) (c1 c2 : list C) (r1 r2:list R), 
  length l1=length c1 ->
  length c1=length r1->
  zip3 (l1++l2) (c1++c2) (r1++r2) = (zip3 l1 c1 r1)++(zip3 l2 c2 r2).
Proof.
  intros L0 C R0 l1 l2 c1 c2 r1 r2 lengthlc lengthcr.
  generalize  l1 c1 r1 lengthlc lengthcr.
  apply list_triple_ind.
  simpl; reflexivity.
  intros l c r lenlc lenrc IH x y z.
  simpl. rewrite IH. reflexivity.
Qed.
Implicit Arguments decompZip3 [ L C R l1 c1 r1].

Ltac  zip3Rewrite :=
  match goal with 
    [|- context [ zip3 (?l1++?l2)  (?c1++?c2) (?r1++?r2)]]
    =>
    cut ( length l1= length c1);[ cut (length c1= length r1);[
      let lencr := fresh "lencr" in let  lenlc:= fresh  "lenlc" in intros lencr lenlc; 
        rewrite (decompZip3 l2 c2 r2 lenlc lencr)|auto]|auto]
  end.

  (** length of [zip3 l l' l''] written with length of [l] [l'] and [l''] *)
Lemma zip3_length :  forall L C R (l:list L) (c:list C) (r:list R), 
  length (zip3 l c r) = min (length l) (min (length c) (length r)).
Proof.
  induction l. 
    simpl. auto.
    destruct c; destruct r;auto.
    simpl; auto.
Qed.
Hint Rewrite zip3_length : length.

(** if l or l' or l'' is nil then zip3 l l' l'' is nil*)
Lemma zip3_nil :  forall L C R (l:list L) (c:list C) (r:list R), 
  l=nil \/ c=nil \/r=nil -> 
  zip3 l c r =  nil.
Proof.
  intros L C R l c r H. destruct H ; subst; auto.
  destruct H.
  assert (length (zip3 l c r) = 0). rewrite zip3_length. subst. simpl. firstorder. 
  lengthNil. reflexivity.
  assert (length (zip3 l c r) = 0). rewrite zip3_length. subst. simpl. 
  rewrite (min_comm (length c)).  rewrite (min_comm (length l)). simpl. reflexivity.
  lengthNil. reflexivity.
Qed.


(** [n]th element of [zip3 l l' l''] written with [n]th elements of [l] [l'] and [l''] *)
Lemma zip3_nth :  forall L C R (l:list L) (c:list C) (r:list R) n d dl dc dr, 
  n< length (zip3 l c r) -> nth n (zip3 l c r) d = ((nth n l dl),(nth n c dc),(nth n r dr)).
Proof.
  induction l.
    simpl. intros c r n d dl dc dr bot. contradict bot;omega.
    
    destruct c; destruct r;auto. 
    simpl; auto. intros n d dl dc dr bot. contradict bot;omega.
    simpl; auto. intros n d dl dc dr bot. contradict bot;omega.
    simpl; auto. intros n d dl dc dr bot. contradict bot;omega.
    intros n d dl dc dr len.
    simpl. induction n. reflexivity.
(*     simpl in len. *)
(*     destruct n0. contradict len;omega. *)
    apply IHl.
    simpl in len. omega.
Qed.

Lemma zip3_nth' :  forall L C R (l:list L) (c:list C) (r:list R) n  (H:n< length (zip3 l c r)) (H':n< length l)  (H'': n < length c)  (H''':n< length  r), 
    nth' n (zip3 l c r) H = ((nth' n l H'),(nth' n c H''),(nth' n r H''')).
Proof.
  induction l.
    simpl. 
    intros c r n H H' H'' H'''.  contradict H;omega.
    
    destruct c; destruct r;auto. 
    simpl; auto. intros n  bot. contradict bot;omega.
    simpl; intros n H H' bot. contradict bot;omega.
    simpl; auto. intros n bot. contradict bot;omega.
    intros n H H' H'' H'''.    
    simpl. induction n. reflexivity.
(*     simpl in len. *)
(*     destruct n0. contradict len;omega. *)
    simpl in *.
    copy H  as H0.
    copy H'  as H'0.
    copy H''  as H''0.
    copy H'''  as H'''0.
    apply lt_S_n  in H0.
    apply lt_S_n  in H'0.
    apply lt_S_n  in H''0.
    apply lt_S_n  in H'''0.
   
    unfold nth'. simpl. 
    rewrite <-  (nth'_nth _ H0).
    rewrite <-  (nth'_nth _ H'0). 
    rewrite <-  (nth'_nth _ H''0). 
    rewrite <-  (nth'_nth _ H'''0).
    apply IHl.
Qed.

Ltac zip3_nth_rewrite :=
  match goal with 
    [ len : ?n< length (zip3 ?l ?c ?r)
      |-context [nth ?n (zip3 ?l ?c ?r) (?la, ?a, ?lb)]
    ] => rewrite (zip3_nth l c r (la, a, lb) la a lb len)
  end.

(**  if zipping lead to an empty list and all lists have same length, first parameter was an empy list *)
Lemma zip3_nil2: 
  forall (A B C : Type)(l1:list A)(l2:list B)(l3:list C), 
    length l1 = length l2 ->
    length l2 = length l3 ->
    zip3 l1 l2 l3 = nil -> l1=nil. 
Proof.
  intros A B C l1 l2 l3 H H0 H1.
  destruct l1. reflexivity.
  rewrite <-H in H0.
  apply lengthCons in H.
  apply lengthCons in H0.
  destruct H as [b2 Htmp]. destruct Htmp as [l2' Hl2].
  destruct H0 as [b3 Htmp]. destruct Htmp as [l3' Hl3].
  rewrite Hl2 in H1. rewrite Hl3 in H1.
  simpl in H1; discriminate H1.
Qed.

(** If zip3 lead to a non-empty list, all parameter are list with a decomposabe as h::t  *)
Lemma zip3_cons:
  forall (A B C : Type)(la:list A)(lb:list B)(lc:list C)
    (t:A*B*C)(l:list(A*B*C)),
    zip3 la lb lc = t::l -> 
    (exists a:A, exists la':list A, la=a::la')\/
    (exists b:B, exists lb':list B, lb=b::lb')\/
    (exists c:C, exists lc':list C, lc=c::lc').
Proof.
  intros A B C la lb lc t l H.
  destruct la.
  destruct lb.
  destruct lc.
  simpl in H; discriminate H.
  right. right. exists c. exists lc. reflexivity.
  right. left. exists b. exists lb. reflexivity.
  left. exists a. exists la. reflexivity.
Qed.    

(** if all input list have the same length, last element of zipping is the triple of lats element of each lists   *)
Lemma last_zip3: 
  forall (A B C:Type) (a:A) (b:B) (c:C) (l1:list A)(l2:list B)(l3:list C),
    length l1 = length l2 ->
    length l2 = length l3 ->
    last (zip3 l1 l2 l3) (a,b,c) = (last l1 a,last l2 b, last l3 c).
Proof.
  induction l1.
  intros l2 l3 H H0. 
  simpl in H. rewrite <-H in H0. 
  symmetry in H. symmetry in H0.
  apply lengthNil in H.
  apply lengthNil in H0. subst. reflexivity. 
  intros l2 l3 H H0.
  rewrite <-H in H0.
  set(H':=H).
  set(H'0:=H0).
  apply lengthCons in H'.
  apply lengthCons in H'0.
  destruct H' as [b2 Htmp]. destruct Htmp as [l2' Hl2].
  destruct H'0 as [b3 Htmp]. destruct Htmp as [l3' Hl3].
  rewrite Hl2. rewrite Hl3.
  simpl.
  rewrite IHl1.
  case_eq(zip3 l1 l2' l3') ; intros.
  apply zip3_nil2 in H1.
  subst.
  simpl in H0. inversion H0 as [H'0]. 
  symmetry in H'0. apply lengthNil in H'0.
  simpl in H. inversion H as [H']. 
  symmetry in H'. apply lengthNil in H'.
  subst. reflexivity.
  subst. simpl in H. inversion H. reflexivity.
  subst. simpl in H. inversion H. 
  subst. simpl in H0. inversion H0. reflexivity.
  apply zip3_cons in H1.

  destruct H1. destruct H1 as [a' H1].
  destruct H1 as [la' Hl1].
  subst.
  inversion H0. simpl in H.
  inversion H. apply lengthS in H2.
  apply lengthS in H3.
  destruct H2 as [b'2 Htmp]. destruct Htmp as [l2'' Hl2].
  destruct H3 as [b'3 Htmp]. destruct Htmp as [l3'' Hl3].
  inversion Hl2. inversion Hl3.
  reflexivity.

  destruct H1. destruct H1 as [b' H1].
  destruct H1 as [lb' Hl2'].
  subst. simpl in H. inversion H.
  inversion H2. symmetry in H2. 
  apply lengthS in H2.
  destruct H2 as [b'2 Htmp]. destruct Htmp as [l2'' Hl2].
  inversion H0.  simpl in H0.
  rewrite H3 in H2. 
  apply lengthS in H2.
  symmetry in H3. 
  apply lengthS in H3.
  destruct H2 as [b'1 Htmp]. destruct Htmp as [l1'' Hl1].
  destruct H3 as [b'3 Htmp]. destruct Htmp as [l3'' Hl3].
  inversion Hl2. inversion Hl1. subst.
  reflexivity.

  destruct H1. destruct H1.
  subst. 
  inversion H0.
  inversion H.
  rewrite H2 in H3. symmetry in H2.
  apply lengthS in H2.
  apply lengthS in H3.
  destruct H2 as [b'1 Htmp]. destruct Htmp as [l1'' Hl1].
  destruct H3 as [b'3 Htmp]. destruct Htmp as [l3'' Hl3].
  subst. reflexivity.
  subst. inversion H. reflexivity.
  inversion H0. inversion H. rewrite H2 in H3. subst.
  inversion H3. reflexivity.
Qed.

Definition fst3 `(t:A*B*C) : A :=
  match t with 
    (a,b,c)=>a
  end.
Definition snd3 `(t:A*B*C) : B :=
  match t with 
    (a,b,c)=>b
  end.
Definition trd3  `(t:A*B*C) : C :=
  match t with 
    (a,b,c)=>c
  end.

Definition map_zip3_indep_1  (A L C R :Type) (l_lst : list L) (l_lst' : list L) (c_lst:list C) (r_lst : list R) (f : L*C*R -> A) :
  (forall l l' c r, f (l,c,r) = f (l', c, r)) 
  -> 
  length l_lst = length l_lst' -> 
  map f (zip3 l_lst c_lst r_lst) =  map f (zip3 l_lst' c_lst r_lst).
Proof.
  intros A L C R l_lst l_lst' c_lst r_lst f H H0.
  apply map_eq_nth.    
  autorewrite with length.
  f_equal; auto.
  intros n H1 H1'.
  copy H1 as H2.
  copy H1' as H2'.
  autorewrite with length in H2.
  autorewrite with length in H2'.
  apply lt_min in H2. destruct H2 as [Hl Hcr].  apply lt_min in Hcr; destruct Hcr as [Hc Hr].
  apply lt_min in H2'. destruct H2' as [Hl' _]. 
  rewrite (zip3_nth' _ _ _ _ Hl Hc Hr)  .
  rewrite (zip3_nth' _ _ _ _ Hl' Hc Hr)  .
  apply H.
Qed.


Definition map_zip3_indep_2  (A L C R :Type) (l_lst : list L)  (c_lst:list C) (c_lst' :list C) (r_lst : list R) (f : L*C*R -> A) :
  (forall l c c' r, f (l,c,r) = f (l, c', r)) 
  -> 
  length c_lst = length c_lst' -> 
  map f (zip3 l_lst c_lst r_lst) =  map f (zip3 l_lst c_lst' r_lst).
Proof.
  intros A L C R l_lst c_lst c_lst' r_lst f H H0.
  apply map_eq_nth.    
  autorewrite with length.
  f_equal; auto.
  intros n H1 H1'.
  copy H1 as H2.
  copy H1' as H2'.
  autorewrite with length in H2.
  autorewrite with length in H2'.
  apply lt_min in H2; destruct H2 as [Hl Hcr];  apply lt_min in Hcr; destruct Hcr as [Hc Hr].
  apply lt_min in H2'; destruct H2' as  [_ Hcr]; apply lt_min in Hcr; destruct Hcr as [Hc' _] . 
  rewrite (zip3_nth' _ _ _ _ Hl Hc Hr)  .
  rewrite (zip3_nth' _ _ _ _ Hl Hc' Hr)  .
  apply H.
Qed.


Definition map_zip3_indep_3  (A L C R :Type) (l_lst : list L)  (c_lst:list C) (r_lst : list R) (r_lst' : list R) (f : L*C*R -> A) :
  (forall l c  r r', f (l,c,r) = f (l, c, r')) 
  -> 
  length r_lst = length r_lst' -> 
  map f (zip3 l_lst c_lst r_lst) =  map f (zip3 l_lst c_lst r_lst').
Proof.
  intros A L C R l_lst c_lst r_lst r_lst' f H H0.
  apply map_eq_nth.    
  autorewrite with length.
  f_equal; auto.
  intros n H1 H1'.
  copy H1 as H2.
  copy H1' as H2'.
  autorewrite with length in H2.
  autorewrite with length in H2'.
  apply lt_min in H2; destruct H2 as [Hl Hcr];  apply lt_min in Hcr; destruct Hcr as [Hc Hr].
  apply lt_min in H2'; destruct H2' as  [_ Hcr]; apply lt_min in Hcr; destruct Hcr as [ _ Hr'] . 
  rewrite (zip3_nth' _ _ _ _ Hl Hc Hr)  .
  rewrite (zip3_nth' _ _ _ _ Hl Hc Hr')  .
  apply H.
Qed.

(* Implicit Arguments fst3  [A B C].  *)
(* Implicit Arguments snd3  [A B C].  *)
(* Implicit Arguments trd3  [A B C].  *)

Lemma map_fst3_zip3 :
  forall {A B C : Type} 
    (la : list A) (lb : list B) (lc : list C), 
    length la <= length lb -> 
    length la <= length lc ->
    map (@fst3 A B C ) (zip3 la lb lc) = la .
Proof.
  induction la.
  simpl. intros ; reflexivity.
  intros lb lc H H0.
  simpl in *.
  cut (lb<>nil). intro. destruct lb. elim H1. reflexivity.
  cut (lc<>nil). intro. destruct lc. elim H2. reflexivity.
  simpl. 
  rewrite (IHla lb lc).
  reflexivity.
  simpl in H. omega.
  simpl in H0. omega.
  assert (0 < length lc) by omega.
  destruct lc. simpl in H2.
  absurd (0<0). apply lt_irrefl. assumption. discriminate.
  assert (0 < length lb) by omega.
  destruct lb. simpl in H1.
  absurd (0<0). apply lt_irrefl. assumption. discriminate.
Qed.  

Lemma map_f_fst3_zip3 :
  forall (A B C D : Type) (la : list A) (lb : list B) (lc : list C) (f : A ->D) , 
    length la <= length lb -> length la <= length lc 
    ->
    map (fun t => f (fst3 t)) (zip3 la lb lc) =  map f la .
Proof.
  induction la.
  simpl. intros ; reflexivity.
  intros lb lc f H H0.
  simpl in *.
  cut (lb<>nil). intro. destruct lb. elim H1. reflexivity.
  cut (lc<>nil). intro. destruct lc. elim H2. reflexivity.
  simpl. 
  rewrite (IHla lb lc).
  reflexivity.
  simpl in H. omega.
  simpl in H0. omega.
  assert (0 < length lc) by omega.
  destruct lc. simpl in H2.
  absurd (0<0). apply lt_irrefl. assumption. discriminate.
  assert (0 < length lb) by omega.
  destruct lb. simpl in H1.
  absurd (0<0). apply lt_irrefl. assumption. discriminate.
Qed.  

Lemma map_snd3_zip3 :
  forall (A B C : Type) (la : list A) (lb : list B) (lc : list C), 
    length lb <= length la -> length lb <= length lc 
    ->
    map (fun t => snd3 t) (zip3 la lb lc) = lb .
Proof.
  intros A B C la lb; generalize A B C lb la. induction lb0.
  simpl. intros. rewrite zip3_nil.  simpl. reflexivity. right. left; reflexivity.
  intros la0 lc H H0.
  simpl in *.
  cut (la0<>nil). intro. destruct la0. elim H1. reflexivity.
  cut (lc<>nil). intro. destruct lc. elim H2. reflexivity.
  simpl. 
  rewrite (IHlb0 la0 lc).
  reflexivity.
  simpl in H. omega.
  simpl in H0. omega.
  assert (0 < length lc) by omega.
  destruct lc. simpl in H2.
  absurd (0<0). apply lt_irrefl. assumption. discriminate.
  assert (0 < length la0) by omega.
  destruct la0. simpl in H1.
  absurd (0<0). apply lt_irrefl. assumption. discriminate.
Qed.  


Lemma map_trd3_zip3 :
  forall (A B C : Type) (la : list A) (lb : list B) (lc : list C), 
    length lc <= length la -> length lc <= length lb 
    ->
    map (fun t => trd3 t) (zip3 la lb lc) = lc .
Proof.
  intros A B C la lb lc.  generalize A B C lc la lb. induction lc0.
  simpl. intros. rewrite zip3_nil.  simpl. reflexivity. right. right; reflexivity.
  clear la lb lc;intros la lb H H0.
  simpl in *.
  cut (la<>nil). intro. destruct la. elim H1. reflexivity.
  cut (lb<>nil). intro. destruct lb. elim H2. reflexivity.
  simpl. 
  rewrite (IHlc0 la lb).
  reflexivity.
  simpl in H. omega.
  simpl in H0. omega.
  assert (0 < length lb) by omega.
  destruct lb. simpl in H2.
  absurd (0<0). apply lt_irrefl. assumption. discriminate.
  assert (0 < length la) by omega.
  destruct la. simpl in H1.
  absurd (0<0). apply lt_irrefl. assumption. discriminate.
Qed.  

Section Map3.

  Variables (A B C D : Type ).
  Fixpoint  map3_aux  (f: A ->B ->C->D) (la: list A) (lb : list B) (lc: list C)  (acc : list D) {struct la} :=
    match la with 
      | nil => acc
      | ha::ta => 
      match lb with 
        | nil => acc
        | hb::tb => 
        match lc with 
          | nil => acc
          | hc::tc => map3_aux f ta tb tc ((f ha hb hc)::acc)
        end
      end
    end.      
    
  Definition map3   (f: A ->B ->C->D) (la: list A) (lb : list B) (lc: list C)  : list D :=
    rev' (map3_aux f la lb lc (@nil D)).

  Definition map3_nil:
    forall (f: A ->B ->C->D) (la: list A) (lb : list B) (lc: list C),
    la =nil \/ lb=nil \/ lc=nil  -> map3 f la lb lc = nil.
  Proof.
  intros f la lb lc nihilism.
  destruct nihilism as[lanil | [lbnil | lcnil]].
  subst; firstorder.
  unfold map3; destruct la; subst; simpl; reflexivity.
  unfold map3; destruct la;destruct lb; subst; simpl; reflexivity.
Qed.


(** Induction principle for resonning over three lists of same length*)  
Lemma triple_list_ind : forall (P : list A->list B->list C -> Prop),
  (forall lb lc , P nil lb lc) ->
  (forall la  lc ,P la nil lc )->
  (forall la lb ,P la lb nil )->
  (forall la lb lc ,
    P la lb lc ->
    forall a b  c ,
    P (a::la) (b::lb) (c::lc)) 
  ->
  (forall la lb lc ,   P la lb lc).
Proof.
  intros P BH1  BH2  BH3 IH.
  induction la; induction lb; induction lc; auto.
Qed.
Implicit Arguments triple_list_ind [P].

Functional Scheme map3_aux_ind := Induction for map3_aux  Sort Prop.

Lemma map3_aux_head: 
  forall (f: A ->B ->C->D) (a: A) (b : B) (c: C)  
  (la: list A) (lb : list B) (lc: list C) (acc1 acc2 : list D),
  map3_aux f la lb lc (acc1++acc2) = map3_aux f la lb lc acc1 ++ acc2.
Proof.
  intros f a b c la lb lc acc1 acc2.
  revert acc1 acc2.
  apply (triple_list_ind (P:=(fun la lb lc =>(forall acc1 acc2 , map3_aux f la lb lc (acc1 ++ acc2) = map3_aux f la lb lc acc1 ++ acc2)) ));
    try (intros l1 l2 ;case l1;case l2;  reflexivity).
  intros la0 lb0 lc0 IH a0 b0 c0.
  simpl.
  setoid_rewrite <- IH.
  reflexivity.
Qed.

Definition map3_head:
  forall
  (f: A ->B ->C->D) (a: A) (b : B) (c: C)  
  (la: list A) (lb : list B) (lc: list C),
  map3 f (a::la) (b::lb) (c::lc) = (f a b c)::(map3 f la lb lc).
Proof.
  intros f a b c la lb lc.
  unfold map3.
  rev'_to_rev.
  rewrite  <- rev_unit.
  rewrite <- map3_aux_head; auto. 
Qed.

Definition map3_length: forall   
  (f: A ->B ->C->D) (la: list A) (lb : list B) (lc: list C),
  length (map3 f la lb lc) = min (length la) (min (length lb) (length lc)).
Proof.
  intros f.
  apply (triple_list_ind );intros; try(rewrite map3_nil ;[ 
    (simpl; repeat (rewrite min_0_r||rewrite min_0_l); reflexivity)
    |
      firstorder] ).
  simpl.
  rewrite map3_head. simpl. rewrite H. reflexivity.
Qed.

Definition map3_nth:
  forall(f: A ->B ->C->D) 
    (la: list A) (lb : list B) (lc: list C) 
    (n:nat) (nla : n< length la)(nlb : n< length lb)(nlc : n< length lc)  nmap,
    nth' n (map3 f la lb lc) nmap  = f (nth' n la nla)  (nth' n lb  nlb) (nth' n lc nlc).
Proof.
  intros f la lb lc. 
  apply (triple_list_ind (
    P:=(fun la lb lc =>  (forall n nla nlb nlc nmap, nth' n (map3 f la lb lc) nmap  =
      f (nth' n la  nla) (nth' n lb nlb) (nth' n lc nlc)))));
  try (simpl; intros ; match goal with [ abs : ?n<0 |-_] => contradict abs; auto with arith end).
  intros la0 lb0 lc0 IH a b c n. 
  case n. intros nla nlb nlc nmap.
  rewrite (nth'_nth    _ _ (f a b c) ); 
    rewrite (nth'_nth   _ _ ( a) );
      rewrite (nth'_nth   _ _ (b) );
        rewrite (nth'_nth   _ _ (c) ).
  rewrite map3_head.
  reflexivity.
  intros n0  nla nlb nlc nmap. 
  rewrite (nth'_nth   _ _ (f a b c) ) ;
    rewrite (nth'_nth   _ _ ( a) ) ;
      rewrite (nth'_nth   _ _ (b) ) ;
        rewrite (nth'_nth   _ _ (c) ) .
  rewrite map3_head in *. 
  simpl in *. 
  repeat (match goal with [ H: S _< S _ |-_] =>  generalize (lt_S_n _ _ H ); clear H; intro H  end).
  rewrite <- (nth'_nth  _ nmap);
    rewrite <- (nth'_nth  _ nla);
      rewrite <- (nth'_nth  _ nlb);
        rewrite <- (nth'_nth  _ nlc).
  apply IH.
Qed.

  Lemma map_zip3: 
    forall ( A' B' C':Type) (l1 : list A) (l2 : list B) (l3 : list C)
      (f : A * B * C  -> D) (f1 : A->A') (f2 : B->B') (f3 : C->C') 
      (combine: A' * B' * C' -> D),
      (forall x, f x = combine ((f1 (fst3 x)),(f2 (snd3 x)),(f3 (trd3 x)))) ->
      map f (zip3 l1 l2 l3) = 
      map combine (zip3 (map f1 l1) (map f2 l2) (map f3 l3)).
  Proof.
    intros A' B' C' l1 l2 l3 f f1 f2 f3 combine H.
    generalize dependent l2. 
    generalize dependent l3.
    induction l1. auto.
    intros l3 l2.
    destruct l2. auto. 
    destruct l3. auto.
    simpl. rewrite IHl1. rewrite H. reflexivity.
  Qed.

End Map3.
Hint Rewrite map3_length : length.
Implicit Arguments map3 [A B C D].