Require Import List.
Require Import LIFO.List.Notations.

(** *Library Zip
   provides the zip function, which take two lists of same size  and give a list of pair 
   
   zip [a,b,c] [a,b,c] = [(a,a),(b,b),(c,c)]
   *)

(** Type of elements is implicit in the whole library  *)
Section  implicit_A.
Set Implicit Arguments.
Variable A B:Type.

(** The zip function 
   [l1 l2] are the list to be zipped
   they have the same length   
   *)
Program Fixpoint zip  (l1 : list A) (l2 : list B) (same_length :  length l1 = length l2) :=
  match l1 with 
    | [] => 
      match l2 with 
            | [] => []
            | h::l2'=> _
          end
    | h1::l1' =>
      match l2 with 
                  | [] => _
                  | h2::l2'=>  (h1,h2):: zip l1' l2' _
                end
  end.


(** **zip length is the length of its input lists  *)
Lemma zipLength :forall  (l1 : list A) (l2 : list B) (same_length :  length l1 = length l2) ,
  length (zip l1 l2 same_length) = length l1.
Proof.
  induction l1.  
  intros l2 same_length.
  simpl.
  destruct l2. 
      reflexivity.
      simpl in *.
      contradict same_length; simpl; discriminate.
 
  intros l2 same_length.
  destruct l2.
       simpl in *.
       contradict same_length; simpl; discriminate.
       simpl in *.
       rewrite IHl1.       reflexivity.
Qed.      
Hint Rewrite zipLength : length.

(** ** nth element of zip
   is the pair of nth elements of l1 and l2
   *)
Lemma zip_nth :   forall  l1 l2 same_length,
  forall n, n < length l1 -> 
    forall da db, nth n (zip l1 l2 same_length) (da,db) =(nth n l1 da, nth n l2 db).
Proof.
  induction l1; destruct l2 ; try (intros; contradict same_length; simpl; discriminate).
  intros same_length n H d.
  contradict H; simpl; auto with arith.

  intros same_length n H d.
  simpl in *. 
  destruct n.
  reflexivity.
  apply IHl1.
  auto with arith.
Qed.


End implicit_A.

(*  *)