(* $Id: calculationalFormProof.v,v 1.4 2009-05-19 17:23:26 julien Exp $ *)

(** * Definition of tactics for the step by step refinement of
   specification The purpose here is to have an interactive refinement
   which remains readable and for which the proof of equivalence
   beetween each step is certified by coq.  This proof should be as
   automatic as possible so that the user can focus on the derivation
   itself.  *)
Require Import String.
Require Export Setoid.
Require Export Relation_Definitions.
(* Require Import List. *)


(* Module Type AbstractEquivalence. *)
(* Parameter A : Type. *)
(* Parameter R : A->A->Prop. *)
(* Instance  equiv_R : Equivalence  R. *)

(* End AbstractEquivalence. *)


(* Module CalcProof ( S : AbstractEquivalence). *)

(* Export S. *)

(* Add Parametric Relation  : (A) (R) *)
(*  reflexivity proved by (bisim_refl) *)
(*  symmetry proved by (bisim_sym) *)
(*  transitivity proved by (bisim_trans) *)
(*  as bisim_rel. *)

(* Parameter A : Type.  *)
(* Parameter R : A -> A-> Prop.  *)



Ltac whileevar a  eve evf neve nevf  := 
  match a with 
    context e [?a] => is_evar a; let A:=fresh "EVAR" in eve A a; let x := evf A a in let  y := context e[x]  in whileevar y  ltac:eve ltac:evf ltac:neve ltac:nevf
    | _  => let x:=nevf a in  neve a
  end.

Ltac id a := a.

Ltac fst a b := a.

Ltac introEvar A a  := 
  match type of a with 
    ?Ta =>   evar (A: Ta)
  end.

Ltac  print a := idtac a.

Ltac  assert a := assert a.

Ltac  name_evar t :=  whileevar t ltac:introEvar ltac:fst ltac:assert ltac:id .

Ltac  name_evar_goal := match goal with |-?G => name_evar G end.



Inductive state : Prop :=
  RHS : state
| LHS : state
| GOAL : state
| ADHOC : forall s:string, state.  


Inductive memo (s: state) : Prop :=
  mem  :  memo s
.

Tactic Notation "memorize" constr(s) := 
  pose ( mem _ : memo s)
.
Tactic Notation "setRHS" := 
  first [(match goal with 
    | [ H  : memo RHS |- _]=> idtac "already RHS"
    | [ H  : memo LHS |- _]=> clear H; memorize RHS
    | [ H  : memo GOAL |- _]=> clear H; memorize RHS
    | [ H  : memo (ADHOC ?s) |- _]=> memorize RHS
  end
  ) | (memorize RHS)]
.


Tactic Notation "setLHS" := 
  first [(match goal with 
    | [ H  : memo RHS |- _]=> clear H ; memorize LHS
    | [ H  : memo LHS |- _]=> idtac "already LHS"
    | [ H  : memo GOAL |- _]=> clear H ; memorize LHS
    | [ H  : memo (ADHOC ?s) |- _]=> memorize LHS
  end
  ) | (memorize LHS)]
.

Tactic Notation "setGOAL" := 
  first [(match goal with 
    | [ H  : memo RHS |- _]=> clear H ; memorize GOAL
    | [ H  : memo LHS |- _]=> clear H ; memorize GOAL
    | [ H  : memo GOAL |- _]=> idtac "already GOAL"
    | [ H  : memo (ADHOC ?s) |- _]=> memorize GOAL
  end
  ) | (memorize GOAL)]
.


Tactic Notation (at level 0) "if_RHS_LHS_BOTH_NOTH"  tactic(trhs) tactic(tlhs) tactic(tboth) tactic(tnoth) :=
  (lazymatch goal with
    | [ H : memo RHS |- _]=> trhs
    | [ H : memo LHS |- _]=>  tlhs
    | [ H : memo GOAL |- _]=> tboth
    | _ => tnoth
  end
)
  .

Tactic Notation  (at level 0)  "if_RHS_LHS_BOTH"  tactic(trhs) tactic(tlhs) tactic(tboth)  :=
  if_RHS_LHS_BOTH_NOTH  (trhs) (tlhs) (tboth) (fail "no side indication").


Tactic Notation "is_equivalence" constr(R) :=
  let H:= fresh in assert (Equivalence R) by (typeclasses eauto); clear H.

Tactic Notation "is_refinement" constr(R) :=
  let H:= fresh in assert (Equivalence R) by (typeclasses eauto); clear H.

(* Tactic Notation "is_refinement" constr(R) := *)
(*   let H:= fresh in assert (Refinement R) by (typeclasses eauto); clear H. *)

(** Selection of the side to rewrite in the equation *)
(** Right Side  *)
Tactic Notation  (at level 1)  "RHS" tactic(t) :=
  (if_RHS_LHS_BOTH_NOTH
    (idtac) 
    (idtac;
      match goal with 
        | [|- (?R ?lhs  ?rhs)] =>  subst rhs
      end
    ) 
    (idtac)
    (idtac)
  ) ;
  setRHS;
  match goal with 
    | [|-?R ?lhs ?rhs] =>  is_equivalence R;let LHS := fresh "LHS" in set (LHS:=lhs);
      assert(R LHS   rhs);[|assumption]
  end
  ;
  t.

(** Left Side  *)
Tactic Notation (at level 1) "LHS" tactic(t) :=
  (if_RHS_LHS_BOTH_NOTH
  (idtac;
    match goal with 
      | [|-?R ?lhs ?rhs] =>  subst lhs
    end
  ) 
  (idtac
  ) 
  (idtac)
  (idtac)
  );
  setLHS;
  match goal with 
    | [|-?R ?lhs ?rhs] =>  is_equivalence R; let RHS := fresh "RHS" in set (RHS:=rhs);
      assert(R lhs   RHS);[|assumption]
  end;
  t
  .

(** Both Side  *)
Tactic Notation  (at level 1)  "GOAL" tactic(t) :=
  (if_RHS_LHS_BOTH_NOTH
  (idtac;
    match goal with 
      | [|-?R ?lhs ?rhs] =>  subst lhs
    end
  ) 
  (idtac;
    match goal with 
      | [|-?R ?lhs ?rhs] =>  subst rhs
    end
  ) 
  (idtac)
  (idtac)
  );
  setGOAL;
  t
  .
  

(* debbuging tactics *)
Ltac print_goal :=
  match goal with
    [|-?G]=> idtac G
end.


(** *Derivation tactics  *)

(**Elimination of existentially quantified variable *)
Tactic Notation "elim_ex" constr(e):=
  repeat (
    match (type of e) with 
      | exists x,_ => econstructor 
      | sig _ => econstructor 
      | _ => fail
    end
    )
.
Tactic Notation "eexistss" := 
  repeat( match goal with 
            | [|- exists A :?T,_] => 
              eexists
            | [|-sig _] => eexists
          end
 ).


(** Start an interactive derivation session*)
Ltac Begin := 
  intros;
  match goal with 
    | [|-?R ?lhs ?rhs] => is_equivalence R; idtac
    | [|- exists A ,_] =>
      eexists;Begin
    | [|- sig _] =>
      eexists;Begin
    | _ => eexists;intros
    | _ => idtac "no equivalence"
  end.


Ltac BeginEvar := name_evar_goal.

Ltac endEvar :=     
repeat  
  match reverse goal with
    | [var : ?A |- _] => (lazymatch type of var with memo (ADHOC "stop") => fail  | ?a => idtac end);subst var
end;
match  goal with
    | [var : ?A |- _] => apply var
end.


(** Start an interactive derivation session*)
Ltac BeginDebug := 
  intros;
  match goal with 
    | [|-?R ?LHS ?rhs] =>  is_equivalence R;idtac "Goal";print_goal
    | [|- exists A :?T,?B] =>  
      idtac "Existential goal ";print_goal; eexists; BeginDebug
  end.


(** End of an interactive derivation session *)
Tactic Notation "[]" := 
  GOAL (reflexivity || let Stop := fresh "Stop" in 
  (pose (Stop := mem (ADHOC "stop"));
    repeat  
      match reverse goal with
        | [var : ?A |- _] => (lazymatch type of var with memo (ADHOC "stop") => fail  | ?a => idtac end);subst var
  end
  ; clear Stop
);reflexivity).

(** this tactic adds e=e2 in hypothesis if it can be proved by 1 rewritting step (using e1)
   then e is rewritted in e2 in the goal  *)
Tactic Notation "prove" constr(e) "=" "{" constr(e1) "}" constr(e2) :=
  let h := fresh in assert (h : e = e2) by (rewrite e1;reflexivity) ;rewrite h.

Tactic Notation "prove" constr(e) "=" "{" tactic(t1) "}" constr(e2) :=
  let h := fresh in assert (h : e = e2) by (t1;reflexivity) ;rewrite h.

Tactic Notation "prove" constr(e) "=" "{" constr(e1) "}" "using" constr(R) constr(e2) :=
  let h := fresh in assert (h : R e e2) by (rewrite e1;reflexivity) ;rewrite h.
Tactic Notation "prove" constr(e) "=" "{" tactic(t1) "}" "using" constr(R) constr(e2) :=
  let h := fresh in assert (h : R e e2) by (t1;reflexivity) ;rewrite h.


(* Vérification du but courant *)

(** check that e1 is the LHS*)
Tactic Notation "check" constr(e1) "is" "LHS" (* ident(e2) *) :=
match goal with 
  [|- ?R e1?RHS] =>  is_equivalence R;idtac "OK"
  | [|- ?R e1?RHS] =>  is_refinement R;idtac "OK"
  | _ => idtac "failed";fail
end.

(** check that e1 is the LHS*)
Tactic Notation "check" constr(e1) "is" "RHS" (* ident(e2) *) :=
match goal with 
  [|- ?R ?LHS e1] =>  is_equivalence R;idtac "OK"
  | [|- ?R ?LHS e1] =>  is_refinement R;idtac "OK"
  | _ => idtac "failed";fail
end.

(** check that e1 is the GOAL*)
Tactic Notation "check" constr(e1) "is" "GOAL" :=
match goal with 
  [|-e1] =>  idtac "OK"
  | _ => idtac "failed";fail
end.


(** Tactique de transformation de programmes 
   Cette tactique assure l'équivalence entre deux programmes
*)
Tactic Notation  (at level 2)    "=" "{"tactic(t1) "}" constr(e2) :=
  if_RHS_LHS_BOTH
  (idtac;
    ((match goal with
      | [|-?R ?lhs ?rhs] =>
        is_equivalence R;
        ((let h := fresh "rewriting" in 
          assert(h : R rhs e2) by (t1;try reflexivity) ; 
            rewrite h
        )
        || (change (rhs) with (e2))
        )
      end)
    ||
      (t1; (match goal with
              | [|-?R ?lhs e2] => is_equivalence R
            end)
    )
    )
  )
  (idtac;
    ((match goal with
     [|-?R ?lhs ?rhs] =>
            is_equivalence R;
            ((let h := fresh "rewriting" in 
              assert(h : R lhs e2) by (t1; try reflexivity) ;(rewrite h || change (lhs) with (e2))
            )
            || (change (lhs) with (e2))
            )
      end
    )
    || (t1; (match goal with
                   | [|-?R e2 ?rhs] => is_equivalence R
                 end)
    )
    )
  )
  ( (t1; check e2 is GOAL)
    ||
    (idtac;     
    match goal with
     [|-?R ?lhs ?rhs] =>
     is_equivalence R;
     match e2 with
       | ?R ?lhs_e2  ?rhs_e2 =>
         is_equivalence R;
         let h := fresh "rewriting" in 
           assert(h : R lhs lhs_e2) by (t1;try reflexivity) ;
           rewrite h;
             let h2 := fresh "rewriting" in 
               assert(h : R rhs rhs_e2) by (t1;try reflexivity) ;
                 rewrite h2
     end
    end
    )
  )
    
  .

Tactic Notation (at level 2 )   "=" "{" tactic(t1) "}d" constr(e2) :=
 info(  = { t1 } e2).


(** Tactique de raffinement de programmes
   Cette tactique assure la correction du rafinement (non nécéssairement symmetrique)
   
 *)
Tactic Notation  (at level 2)    ">" "{"tactic(t1) "}" constr(e2) :=
  if_RHS_LHS_BOTH
  (idtac;
    match goal with
      | [|-?R ?lhs ?rhs] =>
        is_refinement R;
        let h := fresh "rewriting" in 
          assert(h : R rhs e2) by (t1;reflexivity) ;rewrite h
    end)
  (idtac;
    match goal with
     [|-?R ?lhs ?rhs] =>
     is_refinement R;
     let h := fresh "rewriting" in 
       assert(h : R lhs e2) by (t1;reflexivity) ;rewrite h
   end
  )
  (
    (t1; check e2 is GOAL)
    ||
      (idtac;
    match goal with
     [|-?R ?lhs ?rhs] =>
     is_refinement R;
     match e2 with
       | ?R ?lhs_e2  ?rhs_e2 =>
         is_equivalence R;
         let h := fresh "rewriting" in 
           assert(h : R lhs lhs_e2) by (t1;reflexivity) ;
           rewrite h;
             let h2 := fresh "rewriting" in 
               assert(h : R rhs rhs_e2) by (t1;reflexivity) ;
               rewrite h2
     end
   end
    )
  )
  .

Tactic Notation (at level 2 )   ">" "{" tactic(t1) "}d" constr(e2) :=
 info(  > { t1 } e2).


(** Starting form a goal with a LHS=RHS form, this tactic adds LHS=e2 in hypothesis if it can be proved by 1 rewritting step at place given by the function f using the equality e2
   then LHS is rewritted in e2 in the goal 
example with a goal x+x+x+x+x we want rewrite it as 2x+x+2x using ax: forall x, x+x=2x : 
    LHS 
    ={ pat (fun l=> l+x+l) (ax x)}
    2*x+x+2*x

*)
Tactic Notation "pat" constr(f) constr(e1)  :=
  refine (eq_ind_r f _ e1) || fail "refining failed"
.



(** derivations true by definition of a given term*)
Tactic Notation "by" "def" reference(e1) :=
  unfold e1.



(** derivation by tactics t   *)
Tactic Notation "!<"tactic(t) ">" (* ident(e2) *) :=
  t.

Tactic Notation "!<" "apply" constr(thm) ">":=
    apply thm;intros.



(* (** check that e1 is the LHS*) *)
(* Tactic Notation "case" constr(e1) := *)
(*   if_RHS_LHS_BOTH *)
(*   (idtac; *)
(*     match goal with *)
(*       | [|-?lhs=?rhs] => let h:=fresh in *)
(*           (assert(h :rhs=-= e1) by (reflexivity)||fail);clear h *)
(*     end) *)
(*   (idtac; *)
(*     match goal with *)
(*      [|-?lhs=?rhs] =>let h:=fresh in *)
(*           (assert(h :lhs=-= e1) by (reflexivity)||fail);clear h *)
(*    end *)
(*   ) *)
(*   (idtac; *)
(*     match goal with *)
(*      [|-?lhs=?rhs] => *)
(*      match e1 with *)
(*        | ?lhs_e2 = ?rhs_e2 => *)
(*          let h := fresh "rewriting" in  *)
(*            (assert(h : R lhs lhs_e2) by (reflexivity)||fail) ; *)
(*            clear h; *)
(*            let h2 := fresh "rewriting" in  *)
(*              (assert(h : R rhs rhs_e2) by (reflexivity)||fail) ; *)
(*                clear h2 *)
(*      end *)
(*    end *)
(*   ) *)
(*   . *)


(** check that e1 is the GOAL*)
Tactic Notation "check" constr(e1) "is" "GOAL" :=
match goal with 
  [|-e1] =>  idtac "OK"
  | _ => idtac "failed";fail
end.

(* End CalcProof. *)
(** *Use case :

Lemma map_map_fusion:
  forall (A B C : Type) (f : B->C) (g : A->B),
  (map f :o: map g) = map (f :o: g).
Proof.
  Begin.
  !< apply fun_ext>.
  induction x.
(*   step x = nil *) 
  LHS (*  ((map f :o: map g) nil) *)
  ={ by def comp}
  (map f (map g nil)).
  ={ by def map}
  (map f nil).
  ={ by def map}
  ( nil (A:=C)).

  RHS (*  map (f :o: g) nil *) 
  ={ by def map}
  ( nil (A:=C)).

  [].
(*   step x = a::x *) 
  LHS  (*  ((map f :o: map g) a::x) *)
  ={ by def comp}
  ( map f (map g (a :: x))).
  ={ by def map}
  ( map f ((g a)::(map g x))).
  ={ by def map}
  ( f (g a):: map f (map g x)).

  RHS   (*  (map (f:o:g) a::x) *)
  ={ by def map}
  ( ((f:o:g) a):: map (f:o:g) x).
  ={ by def comp}
  ( f(g a):: map (f:o:g) x).
  ={IHx}
  ( f(g a):: (map f :o: map g) x).
  ={ by def comp}
  ( f(g a):: (map f( map g x))).
  [].
Qed.

 *)
