(** *Tower Building example  *)

Unset Automatic Introduction.   (* coq 8.3 *)
Require Import List.
Require Import Bool.

Require Import SDPP.Misc.total_order.
Require Import SDPP.Misc.bounded_set.
Require Export SDPP.Skel.map_around.

Module Type TOWER_PARAMETERS.
  Variable A : Type.
  Variable R : relation A.
  Variable preoR : PreOrder R.
  Variable totR : TotalOrder (preo:=preoR) R.
  Variable eq_A_dec : forall a a0 : A, {a=a0}+{a<>a0}.
  Variable R_dec : forall a a0 : A, {R a a0}+{~ R a a0}.
  Variable AngleL : A->A->A->A->A.
  Variable AngleR : A->A->A->A->A.
  Variable preprocess_L_x preprocess_L_h preprocess_R_x preprocess_R_h : A -> A.
End TOWER_PARAMETERS.  

Module Tower  (Bs : PRIMITIVES) 
                     (Bsml_prop : Properties.TYPE Bs) 
                     (ParSig : ParSig.TYPE Bs Bsml_prop) 
                     (Bsml_base : Base.TYPE Bs Bsml_prop ParSig)  
                     (Parallel : Parallelisation )
                     (Partition : Lists.TYPE Bs Bsml_prop ParSig Bsml_base Parallel)
                     (Bsml_bh : bh_bsml.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition) 
                     (Bh_Classes : bh_classes.TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh) 
                     (Ma : map_around.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh Bh_Classes)
                     (tp: TOWER_PARAMETERS).

  Import Ma.  
  Import AlgebraClasses.
  
  
  Module BoundedSigInst <: BoundedSig.
    Definition A:=tp.A.
    Definition R :=tp.R.
    Definition preoR := tp.preoR.
    Definition totR := tp.totR.
  End BoundedSigInst.
  
  Module BoundedA := BoundedSets (BoundedSigInst).

  Import BoundedA.

  Notation "a ?< b" := (BSOrder_dec (R_dec:=tp.R_dec)  a b) (at level 80, right associativity).


  Definition Max (a b: BoundedSet) :=
    if a ?< b then b else a.

  Program Instance max_min_is_id : Neutral Max Min.
  Next Obligation.
    unfold Max. 
    constructor. intro a.
    destruct (Min ?<  a) as [Minb | notMinb];[reflexivity | assert ( BSOrder Min a) by (unfold BSOrder; unfold makeTotalOrderBoundedSet; tauto);contradiction].  
  Qed.
  Next Obligation.
    unfold Max. 
    constructor. intro a.
    destruct (a ?<  Min) as [bMin | notbMin] ;[ 
      assert ( BSOrder Min a) by (unfold BSOrder; unfold makeTotalOrderBoundedSet; tauto);
        apply (antisymmetry (R:=BSOrder)); tauto 
      | reflexivity].  
  Qed.

  Lemma max_commute : forall a b, Max a b = Max b a.
  Proof.
    unfold Max. 
    intros a b.
    destruct (BSOrder_dec (R_dec:=tp.R_dec) a b);
      destruct (BSOrder_dec (R_dec:=tp.R_dec) b a); auto.
    apply partial_order_antisym;auto; apply TotalOrder_PartialOrder; auto.
    generalize (TotalOrder_totality (R:=BSOrder)). unfold TotalRelation. intro H. generalize (H a b); clear H; intro H.
    elim H; contradiction.
  Qed.

  Program Instance max_assoc : Associative  Max.
  Next Obligation.
    unfold Max. intros x y z.
    destruct (x?<y) as [xy | yx] ; destruct (y?<z) as [yz | zy].
    destruct (x?<z) as [xz | zx ]. reflexivity.
    assert (BSOxz : BSOrder x z) by apply (transitivity xy yz).
    contradiction.

    destruct (x?<y) as [ | yx] . reflexivity. contradiction.

    destruct (x?<z) as [xz | zx]. reflexivity. reflexivity.
    destruct (x?<y) as [xy | ];destruct (x?<z). contradiction. contradiction.
(*   apply (partial_order_antisym(R:=BSOrder));auto;[ *)
(*   apply (TotalOrder_PartialOrder (R:=BSOrder))|]. *)
    assert (BSOyx : BSOrder y x).
    generalize (TotalOrder_totality (R:=BSOrder)). unfold TotalRelation. 
    intros H. generalize (H x y); clear H; intro H.
    elim H; intro H0 ; [ contradiction |  tauto].

    assert (BSOzy : BSOrder z y).
    generalize (TotalOrder_totality (R:=BSOrder)). unfold TotalRelation. 
    intros H. generalize (H y z); clear H; intro H.
    elim H; intro H0 ; [ contradiction |  tauto].
    generalize (transitivity BSOzy BSOyx ). intro BSOzx.
    apply (partial_order_antisym(R:=BSOrder));auto. apply (TotalOrder_PartialOrder (R:=BSOrder)).
    reflexivity.
  Qed.


  Section parameterized.

    Variable xL hL xR hR : A.

    Fixpoint maxAngleL (xs :list (A*A)) : BoundedSet   := 
      match xs with 
        | nil => Min
        | (x,h)::l => let angle1:= Elmt (tp.AngleL xL hL x h) 
          in let angle2:=(maxAngleL l)  
            in Max angle1 angle2 
      end.

  Fixpoint maxAngleR (xs :list (A*A)) : BoundedSet   := 
    match xs with 
      | nil => Min
      | (x,h)::l => let angle1:= Elmt (tp.AngleR xR hR x h) 
        in let angle2:=(maxAngleR l)  
          in Max angle1 angle2 
    end.


  Definition visibleL (xs :list (A*A)) (xi:A) (hi:A) : bool   :=  
    if  (maxAngleL xs) ?< (Elmt (tp.AngleL  xL hL (tp.preprocess_L_x xi) (tp.preprocess_L_h hi))) then true else false.

  Definition visibleR (xs :list (A*A)) (xi:A) (hi:A) : bool   :=  
    if  (maxAngleR xs) ?< (Elmt (tp.AngleR  xR hR (tp.preprocess_R_x xi) (tp.preprocess_R_h hi))) then true else false.

  Definition visibleLR (l :(list (A*A))*(A*A)*(list (A*A))) : bool :=
    let (lsxh,rs):= l in
      let (ls, xh):=lsxh in 
        let (x,h):=xh in
          andb (visibleL ls x h)  (visibleR rs x h).


  Definition tower  (xs : list (A*A)) : list bool:=
    mapAround  visibleLR xs.




  Program Instance maxangleL_homomorphism : Homomorphism.Homomorphism' Max maxAngleL.
  Next Obligation.
    induction x; induction y ; simpl in *. 
    rewrite left_neutral. reflexivity.
    rewrite left_neutral. rewrite IHy. reflexivity.
    rewrite IHx. destruct a.   rewrite (right_neutral).   rewrite (right_neutral ). reflexivity.
    destruct a. destruct a0. 
    rewrite associative.
    rewrite IHx.
    reflexivity.
  Qed.

  Program Instance maxangleR_homomorphism : Homomorphism.Homomorphism' Max maxAngleR.

  Next Obligation.
    induction x; induction y ; simpl in *. 
    rewrite left_neutral. reflexivity.
    rewrite left_neutral. rewrite IHy. reflexivity.
    rewrite IHx.   
    destruct a.  rewrite right_neutral.   rewrite (right_neutral (op:=Max)). reflexivity.
    destruct a. destruct a0. 
    rewrite associative.
    rewrite IHx.
    reflexivity.
  Qed.


Require Import Program.
  Import tp.
 
  Program Instance tower_bhable : MapAround_BHable tower
    (L:=BoundedSet)
    (R:=BoundedSet)
    (k:= (fun maxl (xh:A*A) maxr =>
      let (xi,hi):= xh in
        andb 
        (if maxl ?< (Elmt (AngleL  xL hL (preprocess_L_x xi) (preprocess_L_h hi))) then true else false)
        (if maxr ?< (Elmt (AngleR  xR hR (preprocess_R_x xi) (preprocess_R_h hi))) then true else false)
    ))
    (gl:=maxAngleL)
    (opl:=Max)
    (idl:=Min)
    (gr:=maxAngleR)
    (opr:=Max)
    (idr:=Min)
    :={MA_BH_f':=visibleLR}.
  

Definition  Parallel_tower_aux : {f|f=Parallel.parallel (f:=tower)}.
    eexists.
    simpl.
    autounfold with Parallel.
    simpl.
    reflexivity.
    Defined.

    (** The Final parallel tower building implementation  *)
    Definition Parallel_tower := Eval simpl in proj1_sig Parallel_tower_aux.
  
  End parameterized.

End Tower.


