(** *Distribution of lists over Processor, should include partition.v*)
Require Export CertifiedBSML.QuickStart.

Require Export Datatypes.Lists.Lists.
Require Export Datatypes.Distribution.

Require Import Bool.
Require Import List.
Require Import Arith.MinMax.
Require Import Arith.

Require Import LIFO.List.Notations.
Require Import LIFO.ProofUnicity.proof_unicity.
Require Import LIFO.Tactics.
Require Import LIFO.List.Map.
Require Import LIFO.List.Length.

Module Type  ListDistributionByList(Bs : PRIMITIVES) 
  (Bsml_Prop : Properties.TYPE Bs) 
  (ParSig : ParSig.TYPE Bs Bsml_Prop) 
  (Bsml_Base : Base.TYPE Bs Bsml_Prop ParSig)  
  (Parallel : Parallelisation )
  (Partition : Lists.TYPE Bs Bsml_Prop ParSig Bsml_Base Parallel) 
  <: DISTRIBUTION_WITH_LOCAL_INDEX Bs
  <: DISTRIBUTION_WITH_COST Bs.
  
Import Bs.
 Import Bsml_Prop.
 Import Bsml_Base.
 Import Partition.

(** lists is a  polymorph type  *)
Definition T:=Type -> Type.
Definition t:=list.


(** The index is the position in the list  *)
Definition index :=nat.

Implicit Arguments has_length [A]. 

(** A distribution is a list of size bsp_p of  size of the local list. *)
Definition distribution := {d :list nat |  has_length d (S max_pid)}.

(** Number of elements distributed (size of the distributed list)  *)
Definition nb_elem (d:distribution) : nat :=
  fold_left plus (proj1_sig d) 0.

  (** Property ensuring that the index denotes a element in the distribution  *)
Definition correct_index  ind  distrib :=
  ind < nb_elem distrib.

Definition strongIndex d :={n:index | correct_index n d}.


(** equality over distributions  *)
Lemma  distribution_eq_length_irrel (d1 d2 : distribution):
  proj1_sig d1 = proj1_sig d2 -> d1 = d2.
Proof.
  intros list_eq; destruct d1,d2.
  simpl in list_eq.
  revert h; rewrite list_eq.
  intro.
  rewrite (has_length_uniqueness_proof nat eq_nat_dec x0 _ h h0).
  reflexivity.
Qed.

(** local index: position in the local list *)
Definition local_index := nat.

 Infix   "?<=" := Arith.Compare_dec.leb (at level 70).


 (** Adaptation of nb_elem to partial distribution  *)
Definition nb_elem'  n (d: {d : list  nat | has_length d n }) : nat :=
  fold_left plus (proj1_sig d) 0.


 (** Adaptation of correct_index to partial distribution  *)
Definition correct_index'  ind  n  distrib :=
  ind < nb_elem' n distrib.

(** creaating partial distribution  *)
Program Definition distribUpTo (p:processor) (d:distribution) : {d : list  nat | has_length d (S (proj1_sig p)) }:=
  firstn (S p) d.
Next Obligation.
 destruct p as [p p_proc] ;  destruct d as [d d_spec]; simpl; 
    rewrite has_length_length in *.
 destruct d. contradict d_spec. auto with bsml arith.
      simpl;autorewrite with length in *.
      simpl in d_spec.
    rewrite max_pid_S_bsp_p in *; injection d_spec as d_spec';rewrite d_spec'. 
    f_equal. 
    auto with arith.
  Qed.

(* a transferer lifo coqlib*)
Notation firstn_skipn_decomp := Firstn_skipn.cutInThree.
  (** tail decomposition of distribUpTo  *)
  Lemma distribUpTo_decomp: 
    forall (p p':processor)  (d:distribution) (H: natOfProcessor p = S (natOfProcessor p')),
      proj1_sig (distribUpTo p d)
      =
      proj1_sig (distribUpTo p' d) ++[nth  (natOfProcessor p)  (proj1_sig d) 0].
  Proof.        
    intros p p' d H.
    destruct p as [p p_proc];destruct  p' as  [p' p'_proc].
    compute [proj1_sig natOfProcessor] in *.
    revert  p_proc; rewrite H.
    intros p_proc.
    unfold distribUpTo.
    compute [proj1_sig].
    destruct d as [d d_spec].
    rewrite has_length_length in d_spec. 
    rewrite max_pid_S_bsp_p in *.
    rewrite <- d_spec in p_proc.
    rewrite  (firstn_skipn_decomp    d p_proc 0) at 1.
    change (nth (S p') d 0 :: skipn (S (S p')) d)  with ([nth (S p') d 0]++ skipn (S (S p')) d). 
    rewrite <- (AlgebraClasses.associative (eqA:=eq) (op:=@app _) ).
    assert ( len : (S (S p')) = length (firstn (S p') d ++ [nth (S p') d 0])) by( (* . Focus 2. *)
       autorewrite with length; rewrite d_spec in *; simpl length;  replace (MinMax.min (S p')  (S max_pid)) with (S p') by( symmetry; auto with arith); omega).

     rewrite (Firstn_skipn.firstn_app1 ((firstn (S p') d ++ [nth (S p') d 0])));[| rewrite len; auto].
     rewrite len.
     rewrite Firstn_skipn.firstn_length_l.
     reflexivity.
   Qed.

  (** tail decomposition of nb_elem'  *)
  Lemma nb_elem'_decomp: 
    forall (p p':processor)  (d:distribution) (H: natOfProcessor p = S (natOfProcessor p')),
      nb_elem' (S (natOfProcessor p)) 
      (distribUpTo p d)
      =
      nb_elem'
      (S  (natOfProcessor p'))
      (distribUpTo p' d)
      +
      nth  (natOfProcessor p)  (proj1_sig d) 0.
  Proof.   
    intros p p' d H.
    unfold nb_elem'.
    setoid_rewrite (distribUpTo_decomp p p' d H).
    rewrite fold_left_app. 
     reflexivity.
   Qed.

 (** nb_elem' on bsp_p  *)
Lemma nb_elem'_nb_elem  :
  forall d, 
    nb_elem'  (S max_pid) d = nb_elem d.
Proof.  
  intros d. 
  reflexivity. 
Qed.

(* Lemma pred_s_bsp_p_bsp_p :  (S (natOfProcessor Bsml_Prop'.lastProcessor)) = bsp_p. Proof.  simpl. unfold bsp_p;omega . Qed.
Lemma eq_partial_distrib : 
{d : list nat | has_length d (S (natOfProcessor Bsml_Prop'.lastProcessor))} = distribution.
  unfold distribution.
  f_equal.
Qed.
  Hint Resolve eq_partial_distrib.

Definition distribution' := {d : list nat | has_length d (S (natOfProcessor Bsml_Prop'.lastProcessor))} .


Definition to_distribution (d: distribution'): distribution. 
unfold distribution' in *.
rewrite eq_partial_distrib in d.
exact d.
Defined.

*)
  (* Coercion  to_distribution':  distribution >-> distribution' . *)

  (** distribUpTo lastProcessor  *)
Program Definition maxProc : processor := max_pid.
Next Obligation. 
  rewrite max_pid_S_bsp_p.
  auto with arith  bsml.
Qed.

Lemma maxProc_lastProcessor :  maxProc = lastProcessor. 
Proof.
  unfold maxProc, lastProcessor.
  apply processor_eq.
  simpl.
  rewrite max_pid_S_bsp_p.
  auto with arith  bsml.
  Qed.
  
  Lemma distribUpToLastProcessor: 
    forall d,
      distribUpTo maxProc d
      =
      d.
  Proof.        
    intros  d.
    destruct d as [d d_spec].
    copy d_spec as d_spec'.
    rewrite has_length_length in d_spec'. 
    unfold distribUpTo.
    set (H:=distribUpTo_obligation_1  maxProc  (exist (fun d0 : list nat => has_length d0  (S max_pid)) d d_spec)).
    generalize  H.
    compute [proj1_sig].
        
    unfold maxProc in *.
    revert dependent d_spec.
    compute [proj1_sig].
    rewrite <- d_spec'.
    rewrite (Firstn_skipn.firstn_length_l d) .
    intros d_spec H.
    rewrite (has_length_uniqueness_proof   _ eq_nat_dec  _  _ H d_spec).
    reflexivity.
   Qed.

Require Import Program.

 (** auxiliary function to build get_position and its spec  *)
Program Fixpoint get_position_aux n (d: {d : list  nat | has_length d n })  (i:index)  (H: correct_index' i n d)(p:processor)  (n_ok: n+p<=bsp_p) {struct n}: 
  {pos : (local_index  * processor) | 
    let  (l,p'):=pos  in p'>=p  /\ i = fold_left (fun a b => a+b) (firstn ( p'-p) d)  0 + l /\  l < nth (p'-p) d 0 
    }:=
    match n with
      | 0 => _
      | S n =>
        match d with
          [] => _
          | h::d' =>
              match (S i ?<= h) as Si_leb_h with
                | true => (i, p)
                | false =>
                  match  (max_pid ?<= p) as leb_pmax_pid_p with
                    |true => _ 
                    |false => get_position_aux n d' (i-h)   _ (S p) _
                  end
              end
        end
    end
    .
(* ===============proof obligations===================== *)
Next Obligation.
(* case d is nil  *)
  contradict H.
  unfold correct_index'.
  unfold nb_elem'.
  copy H0 as H1.
  rewrite has_length_length in H1.
  revert dependent H0.
  rewrite (Length.lengthNil  d H1).
  simpl.
  auto with arith.
Qed.

Next Obligation.
  copy H0 as H1; rewrite has_length_length in H1.
  contradict H1.
  simpl;
  discriminate.
Qed.

Next Obligation.
destruct h.  
discriminate.
destruct p; simpl. 
split ; [auto with arith|]. 
rewrite Minus.minus_diag.
split;[reflexivity | symmetry in Heq_Si_leb_h; apply Compare_dec.leb_complete in Heq_Si_leb_h; omega] .
Qed.

Next Obligation.
  rewrite max_pid_S_bsp_p in n_ok.
  symmetry in Heq_leb_pmax_pid_p.
  apply Compare_dec.leb_complete in Heq_leb_pmax_pid_p.
  (*  h is lesser than S i (thus i is not valid at current processor) and 
     p is greater or equal to max_pid, thus next S p is not a valid processor  *)
  destruct p; destruct n0;  simpl in *.
  (* n0 =0,  absurd because i cannot be a correct index for the given h *)
  unfold correct_index' in H;unfold nb_elem' in H;
    simpl in H. 
  destruct d';[
  contradict H; auto with arith|
  copy H0 as H1; rewrite has_length_length in H1; simpl in H1; contradict H1 ; auto with arith].
  simpl.
  destruct h. auto with arith. 
  symmetry in Heq_Si_leb_h ; apply leb_complete_conv in Heq_Si_leb_h.
  omega.  

  contradict n_ok. 
  intro.
  clear H.
  rewrite has_length_length in H0.
  simpl in H0; omega.
Qed.

Next Obligation.
  inversion H0.
  injection H2.
  intros H6 H7. injection H3; intros H8.
  rewrite H8  ,H6;
  assumption.
Qed.


Ltac leb_to_le :=
  try
    match goal with 
      | [ H :  _ = (?i ?<= ?h) |- _ ] => symmetry in H; leb_to_le
      | [ H :  (?i ?<= ?h)= false |- _ ] => apply  Compare_dec.leb_complete_conv in H;leb_to_le
        |[ H :  (?i ?<= ?h)= true |- _ ] => apply  Compare_dec.leb_complete in H;leb_to_le
    end.

Next Obligation.
(*  case h < i and p < max_pid  *)
  destruct p. simpl in *.
  set (H' := (get_position_aux_obligation_5 (S n0)
           (exist (fun d : list nat => has_length d (S n0)) (h :: d') H0) i H
           (exist (fun pid : nat => pid < bsp_p) x l) n_ok n0 eq_refl h d'
           eq_refl Heq_Si_leb_h Heq_leb_pmax_pid_p)).
  unfold correct_index' in *.
  assert ( nb_elem_eq : nb_elem' n0 (exist (fun d : list nat => has_length d n0) d' H') =
    nb_elem' (S n0)
    (exist (fun d : list nat => has_length d (S n0)) (h :: d') H0) - h).
  unfold nb_elem'.
  simpl.  
  rewrite  (Fold.fold_left_assoc_with_id_prop (e:=0)  _ h). 
  auto with arith.
  rewrite nb_elem_eq. 
  set (len:=nb_elem' (S n0)  (exist (fun d : list nat => has_length d (S n0)) (h :: d') H0)) in *.
  assert (len_h : len >= h).
  subst len; simpl. 
  unfold nb_elem'.
  simpl.  
  rewrite  (Fold.fold_left_assoc_with_id_prop (e:=0)  _ h). 
  omega.
  destruct h.
  omega.

  copy Heq_Si_leb_h as h_i_order.
  leb_to_le.
  omega. 
Qed.

Next Obligation.
  destruct p. simpl in *.
  leb_to_le.
  rewrite max_pid_S_bsp_p.
  omega.
Qed.

Next Obligation.
destruct p; simpl in *; omega.
Qed.

Next Obligation.

  destruct p as [p p_proc].
  destruct p0 as [pi pi_proc].
  simpl in *.
  
  replace (fold_left (fun a b : nat => a + b) (firstn (pi - p) (h :: d')) 0 ) with (fold_left (fun a b : nat => a + b) (firstn (pi - S p) (d')) 0 +h) .
  rewrite AlgebraClasses.associative.
  replace   (h + l) with  (l + h)  by (  rewrite AlgebraClasses.commutative; reflexivity). 
  rewrite <- AlgebraClasses.associative.
  rewrite <- e.
  assert (i >= h ) by(  destruct h;[ omega | leb_to_le;omega ]).

  split;[omega| split;[unfold index in *; omega|]].
  
  case_eq  (pi - p).
  intros pi_eq_p.
  contradict pi_eq_p; omega.
  
  intros n H'.
  assert (pi_Sp_n :pi - S p = n) by omega.
  rewrite  pi_Sp_n in l0. 
  assumption.

  case_eq  (pi - p).
  intros pi_eq_p.
  contradict pi_eq_p; omega.
  intros n H'.
  simpl.
  assert (pi_Sp_n :pi - S p = n) by omega.
  rewrite pi_Sp_n .
  rewrite <-  Fold.fold_left_simpl.
  reflexivity.
Defined.


(* =============== end proof obligations===================== *)

 Obligation Tactic := (try setoid_rewrite max_pid_S_bsp_p); program_simpl;auto with arith.
(** get_position with its specification  *)
 Program Definition get_position_w_spec (d : distribution) (i:index) (i_ok :correct_index i d)  :{
   pos : (local_index  * processor) | 
     let (l,p'):= pos in
       i = fold_left (fun a b => a+b) (firstn  p' d)  0 + l /\  l < nth p' d 0 
 }
 :=
   get_position_aux bsp_p d  i _ 0 _. 
(* =============== proof obligations===================== *)

 Next Obligation.
    destruct d.
    simpl; assumption. 
  Qed.

 Next Obligation.
 simpl; omega.
 Qed.


Obligation Tactic := idtac.
Next Obligation.
  intros d i i_ok.
  set (pos:= (get_position_aux bsp_p
           (exist (fun d0 : list nat => has_length d0 bsp_p) 
              (`d) (get_position_w_spec_obligation_1 d i i_ok)) i
           (get_position_w_spec_obligation_2 d i i_ok)
           (exist (fun pid : nat => pid < bsp_p) 0
              (get_position_w_spec_obligation_3 d i i_ok))
           (get_position_w_spec_obligation_4 d i i_ok))). 
destruct pos as [pos pos_spec].
revert pos_spec.
destruct pos as [li [pi pi_proc]].
simpl.
intros pos_spec.
destruct pos_spec as [pi_pos [i_val li_lt_nth_d]] .
split. rewrite i_val. repeat f_equal. auto with arith.

replace  (pi - 0) with pi in li_lt_nth_d by auto with arith.
auto.
Qed.
(* =============== end proof obligations===================== *)

(** get_position matching the Distribution signature: it returns the   (local_index  * processor)  
   corresponding to the element of index [i] in the sequential view  of the data structure *)
Definition get_position (d : distribution) (i:index) (i_ok : correct_index i d) : (local_index  * processor) := proj1_sig (get_position_w_spec d i i_ok).

(** spec of get_position  *)
Lemma get_position_specification (d : distribution) (i:index) (i_ok : correct_index i d): 
  let (l,p') := get_position d i i_ok in 
    i = fold_left (fun a b => a+b) (firstn  (proj1_sig p') (proj1_sig d))  0 + l
    /\  l < nth   (proj1_sig p') (proj1_sig d) 0 
.
Proof.
  unfold get_position .
  set  (pos :=get_position_w_spec d i i_ok).
  destruct pos.
  simpl in * .
  destruct x. 
      destruct p. destruct d.  simpl in * . assumption. 
Qed.

(** get_proc returns the processor at wich the element indexed by [i] is stored (if available)  *)
Definition get_proc (d : distribution) (i:index) (i_ok : correct_index i d) : processor := 
   snd ( get_position d i i_ok).

 
(** get_proc retrurs the local index at wich the element indexed by [i] is stored (if available)  *)
 Definition get_local_index  (d : distribution) (i:index) (i_ok : correct_index i d) : local_index:=
   fst( get_position d i i_ok).

 (** relation between get_position, get_proc et get_local_index   *)
 Lemma position_is_proc_and_local : forall d i i_ok, 
  get_position d i i_ok =  (get_local_index d i i_ok , get_proc d i i_ok).
 Proof.
   intros d i i_ok.
   unfold get_proc, get_local_index.
   destruct (get_position d i i_ok );intros.  
   reflexivity.
 Qed.

(** tactic for automating the use of the get_position specification  *)
Ltac get_position_simpl := 
  match goal with
    [|- context t [get_position ?d ?i ?i_ok] ]   =>   generalize (get_position_specification d i i_ok); 
      case_eq (get_position d i i_ok); 
        let li := fresh "li" in let pi := fresh "pi" in let pos_dec :=fresh "pos_dec" in  
          let pos_spec := fresh "pos_spec" in
            intros li pi pos_dec pos_spec;
              let i_val := fresh "i_val" in let li_lt_nth_d := fresh "li_lt_nth_d" in 
                destruct pos_spec as [i_val li_lt_nth_d]
            
|    [H: context t [get_position ?d ?i  ?i_ok] |- _]   =>   
  generalize (get_position_specification d i i_ok); 
    case_eq (get_position d i i_ok); 
        let li := fresh "li" in let pi := fresh "pi" in let pos_dec :=fresh "pos_dec" in  
          let pos_spec := fresh "pos_spec" in
            intros li pi pos_dec pos_spec;
              let i_val := fresh "i_val" in let li_lt_nth_d := fresh "li_lt_nth_d" in 
                destruct pos_spec as [i_val li_lt_nth_d]
  end.



Import Bsml_Prop.

Obligation Tactic :=program_simpl.

(** *Computing the actual distribution of elements of parallel list  *)
  Program Definition  compute_distribution (A:Type)  (plst : par (list A))  : {
    d :distribution |     
      forall i i_ok ,
        forall  default:A ,
          let (li,pi):= get_position d i i_ok in
            nth li (get plst pi) default = nth i (listOfParList plst) default
            /\
            fold_left (fun a b => a + b ) d 0 = length (listOfParList plst)
  }
  :=
  map 
  (proj (parfun (@length A) plst))
  (proj1_sig processors)
  . 

(* =============== proof obligations===================== *)

Opaque processors.
Next Obligation. 
  apply has_length_length.
  autorewrite with length.
  rewrite max_pid_S_bsp_p.
  reflexivity.
Qed.

Next Obligation.
  get_position_simpl.
  simpl in *.
  rewrite i_val.

  decomp_lopl pi.
  split.
  replace ( (fold_left (fun a b : nat => a + b)
        (map (proj1_sig (proj (proj1_sig (Bs.apply (proj1_sig (replicate (@length _))) plst)))) lg) 0 )) with
    (length  (fold_left (fun acc y : list A => acc ++ y)
      (map (fun x : processor => get plst x) lg) 
      [])) in *  by ( 
        autorewrite with length;
          f_equal; rewrite map_map ; apply map_ext ; 
            intros a; autorewrite with bsml ; reflexivity ).  
  assert   (H':
    li =
    length
    (fold_left (fun acc y : list A => acc ++ y)
      (map (fun x : processor => get plst x) lg) 
      []) 
    + li 
    - length
     (fold_left (fun acc y : list A => acc ++ y)
        (map (fun x : processor => get plst x) lg) 
        [])
  )by (
  set (n:= length
     (fold_left (fun acc y : list A => acc ++ y)
        (map (fun x : processor => get plst x) lg) 
        []));
  unfold local_index in *; 
    omega).

  rewrite app_nth2.
  rewrite app_nth1.

  f_equal; assumption.

  rewrite <- H'.
  rewrite (nth_indep _ _ (proj1_sig(proj (proj1_sig(Bs.apply (proj1_sig (replicate (@length _))) plst))) firstProcessor )  ) in li_lt_nth_d.
  rewrite map_nth in li_lt_nth_d.
  rewrite nthProcessors in li_lt_nth_d.
 autorewrite with bsml in li_lt_nth_d.
 assumption.
autorewrite with length.
destruct pi as [p p_proc];  auto.
auto with arith.

  change (pi :: ld) with ([pi]++ld) .
  rewrite map_app. 
  
  repeat rewrite fold_left_app.
   rewrite Fold.fold_left_simpl.
   rewrite (Fold.fold_left_simpl(map (proj1_sig (proj (proj1_sig (parfun (@length _) plst)))) [pi]))  .
  autorewrite with length .
  rewrite <- AlgebraClasses.associative.
  rewrite AlgebraClasses.commutative.
  rewrite( AlgebraClasses.commutative (fold_left (fun x y : nat => x + y) (map _ ld) 0)).
  f_equal; simpl.
  f_equal.
  rewrite  map_map.  apply map_ext.
  intros a;  autorewrite with bsml; reflexivity.

  f_equal; simpl.   autorewrite with bsml; reflexivity.

  f_equal.
  rewrite  map_map.  apply map_ext.
  intros a;  autorewrite with bsml; reflexivity.

Qed.
(* =============== end proof obligations===================== *)

Implicit Arguments compute_distribution [A].


(** get_global_position give a global position from a local one   *)
Program Definition get_global_position (d:distribution) (p:processor) (i:{i:local_index| i <nth p d 0}) : {j: index | exists H : (correct_index j d), get_position d j H = (i, p)}  :=
  (fold_left (fun a b => a + b) (firstn p d ) i).

(* =============== proof obligations===================== *)
Next Obligation.
assert (H' : correct_index
            (fold_left (fun a b : nat => a + b)
               (firstn (proj1_sig p) (proj1_sig d)) i) d).
unfold correct_index. unfold nb_elem.
destruct d as [d d_length]; simpl in *.
copy d_length as d_length0;rewrite has_length_length in d_length0.
destruct p as [ p p_proc]; simpl.
copy p_proc as p_proc0.
rewrite max_pid_S_bsp_p in p_proc0.
rewrite <- d_length0 in p_proc0.
rewrite  (firstn_skipn_decomp  d p_proc0 0) at 2.
change (nth p d 0 :: skipn (S p) d) with ([nth p d 0] ++ skipn (S p) d).

rewrite Fold.fold_left_simpl .
rewrite fold_left_app.
rewrite fold_left_app.
simpl in *.
rewrite (Fold.fold_left_assoc_with_id_prop (f:=plus) _ (e:=0)).
change (fold_left plus (firstn p d) 0) with (fold_left (fun x y : nat => x + y) (firstn p d) 0) .
omega.

exists H'.

get_position_simpl.
rewrite Fold.fold_left_simpl in i_val.
 replace p with pi in *. replace i with li. reflexivity.

 (* li =i  as sum pi fst elements +i = sum pi fst elements + li *)
set (n := fold_left (fun x y : nat => x + y)
            (firstn (proj1_sig pi) (proj1_sig d)) 0 ) in i_val.
unfold local_index in *;omega.

destruct p as [p p_proc].
destruct pi as [pi pi_proc].
destruct d as [d d_length].

copy d_length as d_length0;rewrite has_length_length in d_length0.
destruct (lt_eq_lt_dec p pi) as [[lt_p_pi | eq_p_pi]| lt_pi_p] .

(*  case p<pi 
   this is proved by contradiction, 

  i <  nth p (proj1_sig d) 0

  is not compatibe with 
  fold_left (fun x y : nat => x + y)  (firstn p (proj1_sig d)) 0 + i 
  =
   fold_left (fun a b : nat => a + b)  (firstn pi (proj1_sig d)) 0 + loc_index

   as the LHS is strictely included in the RHS 
*)
             rewrite <- max_pid_S_bsp_p in d_length0.
             assert (pi_lt_length : pi< length d) by (rewrite d_length0;assumption).
             assert (p_lt_length : p< length d) by (rewrite d_length0;assumption).
             
             assert( fold_left_pi_min_p :
               fold_left (fun x y : nat => x + y) (firstn pi d) 0 - fold_left (fun x y : nat => x + y) (firstn p d) 0 
               =
               fold_left (fun x y : nat => x + y) (firstn (pi  - p) (nth p d 0 :: skipn (S p) d)) 0
             ).
                     rewrite (Firstn_skipn.cutInThree (n:=p) d p_lt_length 0 ) at 1.
                     rewrite  (Firstn_skipn.firstn_app2 (firstn p d)(nth p d 0 :: skipn (S p) d) ).
                     rewrite fold_left_app. 
                     rewrite (Fold.fold_left_simpl).
                     rewrite Firstn_skipn.firstn_length1 ;[|omega].
                     replace  ( fold_left (fun x y : nat => x + y)
                       (firstn (pi - p) (nth p d 0 :: skipn (S p) d)) 0 +
                       fold_left (fun x y : nat => x + y) (firstn p d) 0 -
                       fold_left (fun x y : nat => x + y) (firstn p d) 0) 
                       with 
                         ( fold_left (fun x y : nat => x + y)  (firstn (pi - p) (nth p d 0 :: skipn (S p) d)) 0)
                         by omega.
                     reflexivity.
                     rewrite Firstn_skipn.firstn_length1 ;omega.
      
                assert ( fold_fst_p_lt_fold_fst_pi:
                  fold_left (fun x y : nat => x + y)  (firstn p d) 0
                  < 
                  fold_left (fun a b : nat => a + b) (firstn pi  d) 0 ).
                        rewrite (Firstn_skipn.cutInThree (n:=p) d p_lt_length  0 ).
                        rewrite  (Firstn_skipn.firstn_app_length (firstn p d)(nth p d 0 :: skipn (S p) d) ).
                        rewrite  (Firstn_skipn.firstn_app2 (firstn p d)(nth p d 0 :: skipn (S p) d) ).
                        rewrite fold_left_app. 
                        rewrite (Fold.fold_left_simpl (firstn (pi - length (firstn p d)) (nth p d 0 :: skipn (S p) d)) ).
                        rewrite Firstn_skipn.firstn_length1;[|omega].
                        case_eq  (pi - p) ; intro p_min_pi;[contradict p_min_pi; omega|].
                        simpl.
                        rewrite (Fold.fold_left_simpl _  (nth p d 0) ) .
                        simpl in *.
                        assert (nth p d 0 >0)by omega.
                        intros H2.
                        set (p_sum:=  fold_left (fun x y : nat => x + y) (firstn p d) 0).
                        omega.
                        rewrite Firstn_skipn.firstn_length1;omega.
                        rewrite Firstn_skipn.firstn_length1;omega.

                 revert   i_val fold_fst_p_lt_fold_fst_pi fold_left_pi_min_p.
                 simpl in*.
                 set (fold_fst_p:= fold_left (fun x y : nat => x + y) (firstn p d) 0 ).
                 set (fold_fst_pi:= fold_left (fun x y : nat => x + y) (firstn pi  d) 0 ).
                 intros i_val fold_fst_p_lt_fold_fst_pi fold_left_pi_min_p.
                 rewrite AlgebraClasses.commutative in i_val.
                 assert ( equation :i+ fold_fst_p - fold_fst_p = fold_fst_pi + li - fold_fst_p) by omega.
                 assert ( equation' :i+ fold_fst_p - fold_fst_p=i) by omega.
                 rewrite equation' in equation; clear equation'.
                 rewrite AlgebraClasses.commutative in equation.
                 assert ( equation' :li + fold_fst_pi - fold_fst_p= li + (fold_fst_pi - fold_fst_p)) by omega.
                 rewrite equation' in equation; clear equation'.
                 
                 rewrite fold_left_pi_min_p in equation.

                 
                 revert equation. case_eq (pi-p);[intros p_min_pi equation; contradict p_min_pi; omega|].
                 intros n H1 equation.
                 simpl in equation. rewrite Fold.fold_left_simpl in equation.
                 rewrite equation in H; contradict H; omega.                 
                 
    (* case p=pi 
       this is the good case !
       *)
                 apply processor_eq; simpl; auto.

    (* case p> p' 
   this is proved by contradiction, 

 loc_index < nth p' (proj1_sig d) 0

  is not compatibe with 
  fold_left (fun x y : nat => x + y)  (firstn p (proj1_sig d)) 0 + i 
  =
   fold_left (fun a b : nat => a + b)  (firstn p' (proj1_sig d)) 0 + loc_index

   as the RHS is strictely included in the LHS 
*)
             rewrite <- max_pid_S_bsp_p in d_length0.
             assert (pi< length d) by (rewrite d_length0;assumption).
             
             assert( fold_left_p_min_pi :
               fold_left (fun x y : nat => x + y) (firstn p d) 0 - fold_left (fun x y : nat => x + y) (firstn pi d) 0 
               =
               fold_left (fun x y : nat => x + y) (firstn (p  - pi) (nth pi d 0 :: skipn (S pi) d)) 0
             ).
                     rewrite (Firstn_skipn.cutInThree (n:=pi) d H0 0 ) at 1.
                     rewrite  (Firstn_skipn.firstn_app2 (firstn pi d)(nth pi d 0 :: skipn (S pi) d) ).
                     rewrite fold_left_app. 
                     rewrite (Fold.fold_left_simpl).
                     rewrite Firstn_skipn.firstn_length1 ;[|omega].
                     replace  ( fold_left (fun x y : nat => x + y)
                       (firstn (p - pi) (nth pi d 0 :: skipn (S pi) d)) 0 +
                       fold_left (fun x y : nat => x + y) (firstn pi d) 0 -
                       fold_left (fun x y : nat => x + y) (firstn pi d) 0) 
                       with 
                         ( fold_left (fun x y : nat => x + y)  (firstn (p - pi) (nth pi d 0 :: skipn (S pi) d)) 0)
                         by omega.
                     reflexivity.
                     rewrite Firstn_skipn.firstn_length1 ;omega.
      
                assert ( fold_fst_pi_lt_fold_fst_p:
                  fold_left (fun x y : nat => x + y)  (firstn pi d) 0
                  < 
                  fold_left (fun a b : nat => a + b) (firstn p  d) 0 ).
                        rewrite (Firstn_skipn.cutInThree (n:=pi) d H0  0 ).
                        rewrite  (Firstn_skipn.firstn_app_length (firstn pi d)(nth pi d 0 :: skipn (S pi) d) ).
                        rewrite  (Firstn_skipn.firstn_app2 (firstn pi d)(nth pi d 0 :: skipn (S pi) d) ).
                        rewrite fold_left_app. 
                        rewrite (Fold.fold_left_simpl (firstn (p - length (firstn pi d)) (nth pi d 0 :: skipn (S pi) d)) ).
                        rewrite Firstn_skipn.firstn_length1;[|omega].
                        case_eq  (p - pi) ; intro pi_min_p;[contradict pi_min_p; omega|].
                        simpl in *.
                        rewrite (Fold.fold_left_simpl _  (nth pi d 0) ) .
                        assert (nth pi d 0 >0) by omega.
                        intros H2.
                        set (pi_sum:=  fold_left (fun x y : nat => x + y) (firstn pi d) 0).
                        omega.
                        rewrite Firstn_skipn.firstn_length1;omega.
                        rewrite Firstn_skipn.firstn_length1;omega.

                 revert  i_val fold_fst_pi_lt_fold_fst_p fold_left_p_min_pi.
                simpl. 
                 set (fold_fst_pi:= fold_left (fun x y : nat => x + y) (firstn pi d) 0 ).
                 set (fold_fst_p:= fold_left (fun x y : nat => x + y) (firstn p  d) 0 ).
                 intros i_val fold_fst_pi_lt_fold_fst_p fold_left_p_min_pi.
                 rewrite AlgebraClasses.commutative in i_val.
                 rewrite (AlgebraClasses.commutative _ li) in i_val.
                 assert ( equation :i+ fold_fst_p - fold_fst_pi = li+fold_fst_pi  - fold_fst_pi) by omega.
                 assert ( equation' :li + fold_fst_pi - fold_fst_pi= li) by omega.
                 rewrite equation' in equation; clear equation'.
                 assert ( equation' :i + fold_fst_p - fold_fst_pi= i + (fold_fst_p - fold_fst_pi)) by omega.
                 rewrite equation' in equation; clear equation'.
                 
                 rewrite fold_left_p_min_pi in equation.

                 
                 revert equation. case_eq (p-pi);[intros pi_min_p equation; contradict pi_min_p; omega|].
                 intros n H1 equation.
                 simpl in equation. rewrite Fold.fold_left_simpl in equation.
                 rewrite <-  equation in li_lt_nth_d ; contradict li_lt_nth_d; simpl in * ; omega.                 
                Qed.
(* =============== end proof obligations===================== *)

Program Definition get_global_position' (d:distribution) (p:processor) (i:local_index)(i_ok: i <nth p d 0) : {j: index | exists H : (correct_index j d), get_position d j H =  (i, p)}  :=
  get_global_position d p i .


  
Definition cost := nat.

Program Definition get_cost (d : distribution) (p : processor) : cost := 
  StrongNth.strongNth  d p.

Next Obligation.
autorewrite with length.
destruct p;   rewrite <- max_pid_S_bsp_p ; assumption.
Qed.


Program Definition  get_balance (d:distribution) :=
  let (min, max):= fold_left (fun res c => (min (fst res) c, max (snd res) c)) d (Matrix.strongHead d ,0) in
    max - min.
Next Obligation.
destruct d.
simpl.
rewrite has_length_length in h.
 lengthPos h.
discriminate.
Qed.

(** ** Parallel List of Indexes *)

(** *** computing parallel list of indexes *)
Program Definition index_par_list(d:distribution) : par (list index) 
    :=
  mkpar (fun p => 
    let len:=(nth p d 0) in
      match len with 
        0=> nil
        |S n=>  seq (get_global_position'  d p 0 _) len 
          end).

Next  Obligation.
auto with arith.
Defined.

(* Ltac prepare_proc_induction :=  *)
(*     match goal with  *)
(*        [|- *)
(*          context e[  listOfParList (?plst) ] *)
(*        ] => *)
(*        let p:=fresh "p" in *)
(*          let X:=  *)
(*            context e[ *)
(*              fold_left (fun x y : list index => x ++ y) (map (fun i : processor => get (plst) i) ( Bsml_Prop'.processorsUpTo p)) (@nil nat)] in  *)
(*            let X':= *)
(*              match X with  *)
(*                context f [nb_elem ?d] =>  context f [nb_elem' ( S (natOfProcessor p)) (distribUpTo p d)] *)
(*                | _ => X *)
(*              end  *)
(*              in *)
(*              let X'':=fun p =>X' *)
(*                in idtac X'' *)
(*              (* rewrite listOfParList_fold_left_app_map_get ;assert (forall p, X' p ) *) *)
(*     end. *)


(** *** sequentialisation of index_par_list is the sequence starting at 0 of length [nb_elem d]   *)
Lemma  index_par_list_spec  (d:distribution)  :
            listOfParList (index_par_list d) = seq 0 (nb_elem d).
  Proof.
    unfold distribution in *.
    (* generalize Bsml_Prop'.lastProcessor ;intro p; *)
    (* match goal with  *)
    (*    [|- *)
    (*      context e[  listOfParList (?plst) ] *)
    (*    ] => *)
    (*      let X:=  *)
    (*        context e[ *)
    (*          fold_left (fun x y : list index => x ++ y) (map (fun i : processor => get (plst) i) ( Bsml_Prop'.processorsUpTo p)) (@nil nat)] in  *)
    (*        let X':= *)
    (*          match X with  *)
    (*            context f [nb_elem ?d] =>  context f [nb_elem' ( S (natOfProcessor p)) (distribUpTo p d)] *)
    (*            | _ => X *)
    (*          end  *)
    (*          in *)
    (*          rewrite listOfParList_fold_left_app_map_get ;assert (X') *)
    (* end. *)
    assert (forall p,    
      fold_left (fun x y : list index => x ++ y)
      (map (fun i : processor => get (index_par_list d) i)
        (processorsUpTo p)) [] =
      seq 0 (nb_elem' (S (natOfProcessor p)) (distribUpTo p d))).
    intros p.
    destruct p as [p p_proc].
    induction p.
    simpl.
    unfold index_par_list; autorewrite with bsml.
    unfold nb_elem'.
    destruct d.
    simpl.
    rewrite has_length_length in *.
    rewrite max_pid_S_bsp_p in *.
    lengthPos h.
    simpl.
    destruct a;
    reflexivity.

    rewrite  (processorsUpToApp p_proc).
    rewrite map_app.
    rewrite Fold.fold_left_app_general. 
    rewrite  IHp.
       rewrite (nb_elem'_decomp 
     (exist (fun pid : nat => pid < bsp_p) (S p) p_proc)  (* p *)
     (exist (fun pid : nat => pid < bsp_p) p  (lt_trans p (S p) bsp_p (lt_n_Sn p) p_proc))  (*S p*)
     ).
   rewrite SeqNat.seqApp0.
   simpl app.
   f_equal.

   destruct d as [d d_spec];   compute [proj1_sig]. 
   case_eq (nth (S p) d 0).
   intros H.
   simpl.

   unfold index_par_list;
   autorewrite with bsml.
   simpl;   rewrite H.
   reflexivity.

   intros n H.
   unfold index_par_list;
   autorewrite with bsml.
   simpl;   rewrite H.
   unfold nb_elem'.
   unfold distribUpTo.
   reflexivity.
   reflexivity.

   generalize (H lastProcessor); clear H; intro H. 
   rewrite processorsUpToLastProcessorIsProcessors in H.
   rewrite <- listOfParList_fold_left_app_map_get in H.
   rewrite <- maxProc_lastProcessor in H.
   rewrite distribUpToLastProcessor in H.
   rewrite nb_elem'_nb_elem in H. 
   assumption.
Qed.

(** ***[index_par_list d] distribution is [d]  *)
Lemma  index_par_list_preserve_distrib  (d:distribution)  :
  proj1_sig (compute_distribution  (index_par_list d)) = d.
  Proof.
    unfold compute_distribution. 
    simpl.
    apply distribution_eq_length_irrel.
    destruct d as [d  d_spec].
    copy d_spec as d_spec' ;rewrite has_length_length in d_spec'. 
    simpl.
    apply Nth.nthEqual.
    exists 0; exists 1; discriminate.
    
    intros n d0.
    case_eq (le_lt_dec bsp_p  n ) .

    intros l n_not_proc.
    rewrite nth_overflow.
    rewrite nth_overflow.
    reflexivity.
    rewrite has_length_length in d_spec. rewrite d_spec. 
    rewrite <- max_pid_S_bsp_p.
    assumption.
    autorewrite with length. assumption.

    intros l n_proc.    
    rewrite (nth_indep  _  _ ((proj1_sig(proj
             (proj1_sig(Bs.apply (proj1_sig(replicate (@length _)))
                  (index_par_list
                     (exist (fun d1 : list nat => has_length d1 (S max_pid))
                        d d_spec)))))) firstProcessor) ).
    rewrite map_nth .
    unfold index_par_list.
    autorewrite with bsml.
    simpl.
    change n  with (natOfProcessor (processorOfNat n l)).
    case_eq (nth n d 0).
    rewrite nthProcessors.
    simpl.
    rewrite (nth_indep d d0  0) .
    intros H.
    rewrite H.
    reflexivity.    
    rewrite d_spec'.    
    rewrite <- max_pid_S_bsp_p.
    assumption.
    
    intros n0 H.
    change n  with (natOfProcessor (processorOfNat n l)).
    rewrite nthProcessors.
    simpl.
    rewrite H.
    autorewrite with length.
    rewrite (nth_indep d d0  0) .
    symmetry; assumption.
    rewrite d_spec'; rewrite <- max_pid_S_bsp_p ; assumption.
    autorewrite with length.
    assumption.
  Qed.

Lemma same_distrib_same_local_length :
  forall (A B :Type) (plst: par (list A)) (plst': par (list B)), 
    proj1_sig (compute_distribution plst) =  proj1_sig (compute_distribution plst') ->
    forall p,
      length (get plst p) = length (get plst' p).
Proof.
  intros A B plst plst' H p.
  unfold compute_distribution in *. 
  simpl in *.
  injection H.
  intros H0.  
  generalize (map_inversion  H0 p).
  intros H1.
  autorewrite with bsml in H1.  
  apply H1.
  apply inProcessors.
Qed.

Implicit Arguments same_distrib_same_local_length [A B plst plst' ].
(* (** *** computing parallel list of indexes *) *)
(* Program Definition strong_index_par_list(d:distribution) : par (list (strongIndex d))  *)
(*     := apply (mkpar   (fun  l => InSig.inSig  (fun i => correct_index i d) l _))  (index_par_list d). *)
(*   (* mkpar (fun p =>  *) *)
(*   (*   let len:=(nth p d 0) in *) *)
(*   (*     match len with  *) *)
(*   (*       0=> nil *) *)
(*   (*       |S n=>  InSig.inSig _ (seq (get_global_position'  d p 0 _) len) _ *) *)
(*   (*         end). *) *)

(* Next  Obligation. *)
(* auto with arith. *)
(* Qed. *)

(* Next Obligation. *)
(*   unfold correct_index. *)
(*   apply SeqNat.inSeqLt in H. *)
(*   unfold nb_elem. *)
(*   destruct p as [p p_proc]; *)
(*   destruct d as [d d_spec]; *)
(*   simpl in *. *)
(*   rewrite has_length_length in d_spec. *)
(*   rewrite <- d_spec in p_proc. *)
(*   rewrite (Firstn_skipn.cutInThree d p_proc 0).  *)
(*   replace (fold_left plus (firstn p d ++ nth p d 0 :: skipn (S p) d) 0) *)
(*     with *)
(*        (fold_left (fun a b : nat => a + b) ((firstn p d ++ [nth p d 0])++ skipn (S p) d) 0).  *)
(*   rewrite <- Fold.fold_fold. *)
(*   rewrite <- Fold.fold_fold. *)
(*   rewrite  Fold.fold_left_simpl. *)
(*   rewrite  (Fold.fold_left_simpl ([nth p d 0]) ) . *)
(*   simpl. *)
(*   omega. *)
(*   change   (nth p d 0 :: skipn (S p) d ) with ( [nth p d 0] ++ skipn (S p) d ). *)
(*   rewrite AlgebraClasses.associative. *)
(*   apply Fold.fold_left_ext. *)
(*   reflexivity. *)
(* Qed.   *)

(* Lemma strong_index_index : *)
(*   forall d,  *)
(*     parfun (map  (@proj1_sig _ _)) (strong_index_par_list d)  = index_par_list d. *)
(* Proof. *)
(*   intros d.    *)
(*   apply par_eq. *)
(*   intros i. *)
(*   apply (Nth.nthSameLengthEqual _ _ 0). *)
(*   autorewrite with bsml length.  admit. *)

  (* =================================== *)

End  ListDistributionByList.

Module Make(Bs : PRIMITIVES) 
  (Bsml_Prop : Properties.TYPE Bs) 
  (ParSig : ParSig.TYPE Bs Bsml_Prop) 
  (Bsml_Base : Base.TYPE Bs Bsml_Prop ParSig)  
  (Parallel : Parallelisation )
  (Partition : Lists.TYPE Bs Bsml_Prop ParSig Bsml_Base Parallel) 
  <: DISTRIBUTION_WITH_LOCAL_INDEX Bs
  <: DISTRIBUTION_WITH_COST Bs.

  Include ListDistributionByList Bs Bsml_Prop ParSig Bsml_Base Parallel Partition.

End Make.