Unset Automatic Introduction.    (* coq 8.3 *)
Require Import Arith.
Require Import Bool.
Require Import Lists.List.
Require Import Omega.
Require Import Zdiv.
Require Import Euclid.

Require Import Program.
Require Import ProgramCalculation.Tactics.
(* Require Import bmf_classes. *)

Require Import LIFO.List.Main.
Require Export  CertifiedBSML.QuickStart.
Require Export SDPP.CorrectParallelisation.Structure.
Require Export SDPP.Datatypes.Lists.Lists.

Module Type  TYPE(Bs : PRIMITIVES) 
  (Bsml_Prop : Properties.TYPE Bs) 
  (ParSig : ParSig.TYPE Bs Bsml_Prop) 
(Bsml_Base : Base.TYPE Bs Bsml_Prop ParSig)  
(Parallel : Parallelisation )
(PartList : SDPP.Datatypes.Lists.Lists.TYPE Bs Bsml_Prop  ParSig Bsml_Base Parallel ).

 (* Module Bsml_prop := Bsml_prop Bs. *)
  (* Module Bsml_base := Bsml_base Bs. *)

  Import Bs.
  Import Bsml_Base.
  Import Bsml_Prop.

Set Implicit Arguments.


Definition SeqListGeBspp  A:Type := {l: list A | length l >= bsp_p}.
  
Definition nonEmptyParList  A:Type :={pl: par (list A) | forall p, get pl p <> nil }.

Import Bs Bsml_Prop  ParSig Bsml_Base Parallel PartList.

Program Definition listOfNonEmptyParList (A:Type) (plst : nonEmptyParList A) : SeqListGeBspp A :=
  listOfParList plst.
Next Obligation.
  intros A plst.
  rewrite bsppLastProcessor.
  listOfParListInduction.
  simpl.
  destruct plst.
  simpl.
  generalize (n firstProcessor); intros.
  auto with list.

  rewrite app_length.
  set (len1 :=  length (flatten (map (get (`plst)) (processorsUpTo p'))) ) in *.
  destruct plst.  
  simpl. 
  simpl in IHp'.
  generalize (n S_p).
  intros.
  apply lengthNotNil' in H.
  omega.
Qed.

Axiom partition_list_ok :  forall A (l: SeqListGeBspp A) p, get (partition_list (proj1_sig l)) p <>nil. 
Axiom SeqListGeBspp_proofIrrel :  forall A (l1 l2: SeqListGeBspp A),  proj1_sig l1 = proj1_sig l2  -> l1 = l2.

Program Definition partition_SeqListGeBspp (A:Type) ( lst : SeqListGeBspp A) : nonEmptyParList  A :=
  partition_list (proj1_sig lst).
Next Obligation.
  intros A plst p.
  apply partition_list_ok.
Defined.

Program Instance  nonEmptyListPartitionnable  A :  Partitionnable (SeqListGeBspp A):={
  parallel_type:=nonEmptyParList A;
  join:= (fun pl => @listOfNonEmptyParList A pl);
  partition:= (fun l =>  @partition_SeqListGeBspp A l)}.
Next Obligation.
  intros A a.
  unfold listOfNonEmptyParList.
  apply SeqListGeBspp_proofIrrel.
  destruct a; simpl.
  generalize (join_part_match (A:= list A)).
  simpl.
  auto.
Defined.
End  TYPE.

Module Make (Bs : PRIMITIVES) 
  (Bsml_Prop : Properties.TYPE Bs) 
  (ParSig : ParSig.TYPE Bs Bsml_Prop) 
(Bsml_Base : Base.TYPE Bs Bsml_Prop ParSig)  
(Parallel : Parallelisation )
(PartList : SDPP.Datatypes.Lists.Lists.TYPE Bs Bsml_Prop   ParSig Bsml_Base Parallel ).
  Include  TYPE Bs Bsml_Prop  ParSig Bsml_Base Parallel PartList.
End Make.