Require Import CertifiedBSML.QuickStart.

Require Import SDPP.Applications.TowerBuilding.tower_building.
Require Import SDPP.Applications.MaximumPrefixSum.maximum_prefix_sum.
Require Import SDPP.Applications.HeatEquation.Heat1D.
Require Import SDPP.Applications.HeatEquation.HeatBH.
Require Import SDPP.CorrectParallelisation.Structure.

Require Export ExtractionDirectives.


(** Instantiation of the tower building example *)
Module TowerInstance (Bs : PRIMITIVES)(Tp : TOWER_PARAMETERS).
  
  Module BsmlProp:=(Properties.Make Bs).
  Module ParSig := (ParSig.Make Bs BsmlProp ).
  Module Bsml_Base := (Base.Make Bs BsmlProp ParSig).
  Module Parallel := Structure.Make.
  Module Partition :=(Lists.MakePartition Bs BsmlProp ParSig Bsml_Base Parallel).
Module Bsml_bh := bh_bsml.Make  Bs BsmlProp ParSig Bsml_Base Parallel Partition.
Module Bh_Classes := bh_classes.Make Bs BsmlProp ParSig Bsml_Base Parallel Partition Bsml_bh.
Module Ma :=  map_around.MAKE  Bs BsmlProp ParSig Bsml_Base Parallel Partition Bsml_bh Bh_Classes.
Include  (Tower Bs BsmlProp ParSig Bsml_Base  Parallel Partition Bsml_bh Bh_Classes Ma Tp).
End TowerInstance.
  

(** Instantiation of the maximum example *)
Module MaximumPrefixSumInstance (Bs : PRIMITIVES)(mpsp : MAXIMUM_PREFIX_SUM_PARAM).
  
  Module BsmlProp:=(Properties.Make Bs).
  Module ParSig := (ParSig.Make Bs BsmlProp ).
  Module Bsml_Base := (Base.Make Bs BsmlProp ParSig).
  Module Parallel := Structure.Make.
  Module Partition :=(Lists.MakePartition Bs BsmlProp ParSig Bsml_Base Parallel).
  Module Bsml_bh := bh_bsml.Make  Bs BsmlProp ParSig Bsml_Base Parallel Partition.
  Module Bh_Classes := bh_classes.Make Bs BsmlProp ParSig Bsml_Base Parallel Partition Bsml_bh.
  Module Ma :=  map_around.MAKE  Bs BsmlProp ParSig Bsml_Base Parallel Partition Bsml_bh Bh_Classes.
  Module LastParallel := LastPar.Make Bs BsmlProp  ParSig Bsml_Base Parallel Partition.
  Module Scan := scan.Make  Bs BsmlProp  ParSig Bsml_Base Parallel Partition Bsml_bh Bh_Classes Ma.
  Module Homomorphism := homomorphism.Make Bs BsmlProp  ParSig Bsml_Base Parallel Partition Bsml_bh Bh_Classes Ma LastParallel.
  Include  (Maximum_prefix_sum Bs BsmlProp ParSig Bsml_Base Parallel Partition Bsml_bh Bh_Classes Ma LastParallel Scan Homomorphism mpsp).
End MaximumPrefixSumInstance.


(** Instantiation of the maximum example *)
Module HeatEquationBHInstance (Bs : PRIMITIVES)  (Number : NUMBER).  
  Module BsmlProp:=(Properties.Make Bs).
  Module ParSig := (ParSig.Make Bs BsmlProp ).
  Module Bsml_Base := (Base.Make Bs BsmlProp ParSig).
  Module Parallel := Structure.Make.
  Module Partition :=(Lists.MakePartition Bs BsmlProp ParSig Bsml_Base Parallel).
  Module Bsml_bh := bh_bsml.Make  Bs BsmlProp ParSig Bsml_Base Parallel Partition.
  Module Bh_Classes := bh_classes.Make Bs BsmlProp ParSig Bsml_Base Parallel Partition Bsml_bh.
Include HeatBH.Make Number Bs BsmlProp  ParSig Bsml_Base Parallel Partition Bsml_bh Bh_Classes.
End HeatEquationBHInstance.

Module HeatEquationNonNil (Bs : PRIMITIVES)  (Number : HeatEquation.Heat1D.NUMBER).  
  Module BsmlProp:=(Properties.Make Bs).
  Module ParSig := (ParSig.Make Bs BsmlProp ).
  Module Bsml_Base := (Base.Make Bs BsmlProp ParSig).
  Module Parallel := Structure.Make.
  Module Partition :=(Lists.MakePartition Bs BsmlProp ParSig Bsml_Base Parallel).
  Module PartitionNonEmpty :=(NonEmptyLists.Make Bs BsmlProp ParSig Bsml_Base Parallel Partition).
  Include HeatEquation.Heat1D.Make Bs Number BsmlProp  ParSig Bsml_Base Parallel Partition PartitionNonEmpty.
End HeatEquationNonNil. 
 