Unset Automatic Introduction.    (* coq 8.3 *)
Require Import LIFO.Algebra.AlgebraClasses.
Require Import SDPP.Misc.total_order.

Module Type GENERIC_MAX_PARAMETERS.
  Variable A : Type.
  Variable R : relation A.
  Variable preoR : PreOrder R.
  Variable totR : TotalOrder (preo:=preoR) R.
End GENERIC_MAX_PARAMETERS.  

Module Generic_Max(gmp  : GENERIC_MAX_PARAMETERS).
  Export gmp.
  (* re-instantiate Order classes *)
   Instance PreOrder_R : PreOrder R. exact  preoR. Qed.
   Instance TotalOrder_R : TotalOrder R (preo:=preoR). exact  totR. Qed.

(** Definition of Max on the relation *)
  Definition Max (a b: A) :=
    if (a ?<=  b) then b else a.

(**  Max is commutative *)
  Lemma max_commute : forall a b, Max a b = Max b a.
  Proof.
    unfold Max. 
    intros a b. 
    generalize (  a ?<= b) ; intro s;  destruct s;
          generalize (  b ?<= a) ; intro s;  destruct s; auto.
    apply (partial_order_antisym  (preo:=preoR))  ;auto; apply (TotalOrder_PartialOrder  (TotalOrder:=totR)); auto.
    apply (partial_order_antisym  (preo:=preoR)) ;auto ; apply (TotalOrder_PartialOrder  (TotalOrder:=totR)); auto.
  Qed.

(**  Max is commutative *)
Global   Instance max_commute_instance : Commutative Max. 
  constructor.     
  apply max_commute.
Qed.

Ltac destructAdHoc t :=
generalize t; intro s; destruct s.

(** Max is associative  *)
Global  Program Instance max_assoc : Associative  Max.
  Next Obligation.
    unfold Max.
    destructAdHoc  (x?<=y);    destructAdHoc  (y?<=z).
    destructAdHoc (x?<=z) . reflexivity.
    assert (xz : R x z) by (apply (transitivity r r0 )).
    apply (antisymmetry r1 xz) ;auto. 

    destructAdHoc (x?<=y)  ; try reflexivity.    apply (antisymmetry  r1 r). 

    destructAdHoc (x?<=z) ; reflexivity. 
    destructAdHoc (x?<=y) ;destructAdHoc (x?<=z) . rewrite  (antisymmetry  r r1) in r0. rewrite  (antisymmetry r r1  ).  apply (antisymmetry r0 r2).  apply   (antisymmetry  r1 r). 
    assert (zx : R z x) by (apply (transitivity   r0 r1)). apply (antisymmetry zx r2). reflexivity.
  Qed.
Generalizable All Variables.

(** Defining the neutral element for Max  *)
Global Program Instance min_unit  {min:A}  {min_smallest : forall a:A, min <= a} : Neutral Max min.
Next Obligation.
  unfold Max. constructor. intros a.
destruct (min ?<= a) as [r|r]. reflexivity. 
apply (antisymmetry (min_smallest a) r).
Qed.

Next Obligation.
constructor; unfold Max. intro a.
destruct (a ?<= min) as [r|r].  apply (antisymmetry (min_smallest a) r). reflexivity.
Qed.
End Generic_Max.