(** *Total Order Type Classes  *)
Unset Automatic Introduction.    (* coq 8.3 *)
Require Export Coq.Classes.RelationClasses.
Require Import Coq.Program.Basics.
Require Export Coq.Relations.Relation_Definitions.
Require Import Arith.Le.
Require Import Omega.

Global Generalizable All Variables.
(** *Type classes related to total orders  *)

(** a total relation [R] over [A] is a relation such that forall x, y in A, R x y hold or R y x hold *)
Class TotalRelation  {A} (R: relation A):=
totality : forall x y : A, {R x y}+{R y x}.


(**  a total order is a total relation which has the properties of a Partial order*)
Class TotalOrder  `{equ : Equivalence A eqA} (R: relation A)  `{preo : PreOrder A R} :={
  TotalOrder_totality :> TotalRelation R ;
  TotalOrder_PartialOrder :>  PartialOrder  eqA  R 
}.

(** the [le] relation over nat is a preorder  *)
Program Instance le_preOrder : PreOrder le .
Next Obligation.
  unfold Transitive. intros x y z.
  apply (le_trans x y z);auto.
Qed.

(** the [le] relation over nat is antisymmetric  *)
Program Instance le_antisymmetric: Antisymmetric nat (@eq nat) le.
Next Obligation.
  omega.
Qed.

(** the [le] relation over nat is a partial order  *)
Program Instance le_PartialOrder : PartialOrder (@eq nat) le.
Next Obligation.
  unfold flip; unfold relation_equivalence; unfold relation_conjunction.
unfold predicate_intersection; unfold predicate_equivalence. simpl.
intros x x0.
omega.
Qed.

(** the [le] relation over nat is a total order  *)
Program Instance le_totalOrder : TotalOrder le. 
Next Obligation.
  unfold TotalRelation.  intros x y.
  destruct (le_ge_dec x y). left ;assumption.
  right; auto.
Qed.

Next Obligation.
  apply le_PartialOrder.
Qed.

(**  a total order is a total relation which has the properties of a Partial order*)
Class TotalyOrdered (A:Type)  `{equ : Equivalence A eqA}:={
   total_order : relation A;
   total_order_preo :> PreOrder total_order;
   total_order_prop :> TotalOrder  total_order  (eqA:=eqA) (equ:=equ)
}.

Program Instance TotalOrder_TotalyOrdered  `(TotalOrder A ) : TotalyOrdered A.
  
Notation "a <= b" := (total_order a b ) (at  level 70).
Notation "a ?<= b" := (totality a b ) (at  level 70).

Transparent totality.
