 Unset Automatic Introduction.    (* coq 8.3 *)
Require Import myTactics.
Require Import List.
Require Import Arith.
Require Import Omega.

Require Import LIFO.List.Notations.
Require Import LIFO.List.Nth.
Require Import LIFO.List.Homomorphism.
Require Import LIFO.List.SeqNat.
Require Import LIFO.List.Firstn_skipn.
Require Import LIFO.List.Map.
(*************************************************)
(** *   BSP Homomorphism definition              *)
(*************************************************)

(** ** Definitions *)
Open Scope nat_scope.

Definition associativity (A: Type) (op: A -> A -> A) :=
  forall (x y z: A), op (op x y) z = op x (op y z).

Definition is_homomorphism (A B: Type) (h: list A -> B) (op: B -> B -> B) :=
  associativity B op
  /\ (forall (b:B), op b (h nil) = b)
  /\ (forall (b:B), op (h nil) b = b)
  /\ (forall (x y: list A), h (x++y) = op (h x) (h y)).


(** ** specification of bh *)

Section BH_SPEC.
  Variables (L A R B:Type).

  Variable (k: L -> A -> R -> B).
  Variable (gl: list A -> L).
  Variable (opl: L -> L -> L).
  Variable (gr: list A -> R).
  Variable (opr: R -> R -> R).

  (*Definition bh_spec := (list A -> L) * (L -> A -> R -> B) * (list A -> R).*)

  (** Definition of the nth element of bh *)
  Definition bh_nth (za: A) (l:L) (lst: list A) (r:R) (n:nat) :=
    k (opl l (gl (firstn n lst))) (nth n lst za) (opr (gr (skipn (S n) lst)) r).

  (** Definition bh *)
  Definition bh_comp (l:L) (lst: list A) (r:R) : list B :=
    match lst with
      | nil => []
      | a :: lst0 => map (bh_nth a l (a :: lst0) r) (seq 0 (length (a :: lst0)))
    end.

  (** ** BH properties *)

  (** BH preserve length of lists*)
  Lemma bh_length (l:L) (lst: list A) (r:R) : length (bh_comp l lst r) = length lst.
  Proof.
    intros.
    destruct lst. simpl. reflexivity.
    simpl.
    rewrite map_length.
    rewrite seq_length.
    reflexivity.
  Qed.

  Lemma lt_len_bh : forall (l:L) (lst:list A) (r:R) (n:nat), 
    n < length lst -> n < length (bh_comp l lst r).
    intros. rewrite bh_length. assumption.
  Defined.

  (** nth element of bh_comp*)
  Lemma bh_bh (l:L) (lst: list A) (r:R) (n:nat) (pre: n < length lst) :
    nth' n (bh_comp l lst r) (lt_len_bh l lst r n pre) = 
    k (opl l (gl (firstn n lst))) (nth' n lst pre) (opr (gr (skipn (S n) lst)) r).
  Proof.
    intros.
    destruct lst. simpl in pre; absurd (n<0). apply lt_n_O. assumption.
    rewrite nth'_nth with (d:=bh_nth a l (a::lst) r 0).
    rewrite nth'_nth with (d:=a).
    unfold bh_comp.
    rewrite map_nth.
    unfold bh_nth.
    rewrite seq_nth.
    auto.
    assumption.
  Qed.

  (** nth element of bh_comp*)
  Theorem nth_bh : 
    forall  (za :A)  (zb:B) (l:L) (r:R) (lst :list A) (n : nat),
      length lst > n ->
      nth n (bh_comp  l lst r) zb =  
      k (opl l (gl (firstn n lst))) (nth n lst za) (opr (gr (skipn (S n) lst)) r). 
  Proof.
    intros za zb l r lst n lstsupn.
    (* destructing list following the homomorphism cases (nil, singleton, list concatenation*)
    destructList lst as H.
    (* case lst =nil*)
    subst. simpl in *.
    absurd (0>n). 
    case n.
    apply (gt_irrefl); assumption.
    intros n0. 
    apply (gt_asym).
    apply (gt_Sn_O).
    assumption.

    (* case lst = [a]*)          
    subst. 
    simpl in *.
    induction n.
    reflexivity.
    absurd  (1 > S n).
    omega.
    assumption.
    (* case lst = a0::la ++ b::lb]*)          
    elim_ex H.
    subst.
    unfold bh_comp.
    induction n.
    auto.
    simpl in *.
    match goal with [|- context e [(bh_nth ?x0 ?l (?x0 :: ?x1 ++ ?x :: ?x2) ?r)]] =>
      rewrite (map_nth' (bh_nth x0 l (x0 :: x1 ++ x :: x2) r) (seq 1 (length (x1 ++ x :: x2))) zb n)
    end.
    rewrite seq_nth;
      unfold bh_nth;
        simpl.
    assert(H: n < length (la ++ b :: lb));[omega| rewrite (nth_indep ( la++ b :: lb) za a H)].
    auto.
    omega.
    rewrite seq_length. omega.
  Qed.
  Hint Resolve nth_bh.

  (** bh_comp follow BH definition*)
  Theorem bh_hom (l:L) (x y: list A) (r:R):
    (is_homomorphism A L gl opl) -> (is_homomorphism A R gr opr) ->
    bh_comp l (x++y) r = bh_comp l x (opr (gr y) r) ++ bh_comp (opl l (gl x)) y r.
  Proof.
    intros l x y r Homl Homr.
    unfold is_homomorphism in Homl,Homr.
    decompose [and] Homl. unfold associativity in H.
    decompose [and] Homr. unfold associativity in H2.
                          (* case study on x *)
    destruct x. 
    (* x is nil *)
    simpl. rewrite H1. reflexivity.
     (* x is not nil => a::x   Case study on y *)
    destruct y. 
    (* y is nil *)
         simpl. rewrite H4. repeat (rewrite <- app_nil_end). reflexivity.
     (* y is not nil => a0::y*)
         unfold bh_comp. 
         rewrite <- app_comm_cons.
         change (a::x++a0::y) with ((a::x)++(a0::y)).
         rewrite app_length.
         rewrite seqApp0.
         rewrite map_app.
         assert (map (bh_nth a l ((a :: x) ++ (a0 :: y)) r) (seq 0 (length (a :: x))) = 
           map (bh_nth a l (a :: x) (opr (gr (a0 :: y)) r)) (seq 0 (length (a :: x)))).
             eapply nthSameLengthEqual. repeat (rewrite map_length). reflexivity.
         
         intros.
         rewrite map_length in H6. rewrite seq_length in H6.
         instantiate (1:=(bh_nth a l ((a :: x) ++ (a0 :: y)) r) 0).
         rewrite map_nth.
         replace ((bh_nth a l ((a :: x) ++ (a0 :: y)) r) 0) with 
           ((bh_nth a l (a :: x) (opr (gr (a0 :: y)) r)) 0).
         rewrite map_nth.
         unfold bh_nth.
         rewrite seq_nth. change (0+n) with n.
         f_equal. f_equal. f_equal. 
         apply firstn_app1. apply lt_le_weak. assumption.
         
         apply app_nth1. assumption.
         
         rewrite skipn_app1.
         rewrite H7. unfold associativity in H2. rewrite H2. reflexivity.
         simpl. firstorder. assumption.

         unfold bh_nth. f_equal. simpl. rewrite H7. rewrite H2. reflexivity.
         
         rewrite H6. f_equal.
         eapply nthSameLengthEqual.
         repeat (rewrite map_length). repeat (rewrite seq_length). reflexivity.
         
         intros.
         rewrite map_length in H8. rewrite seq_length in H8.
         instantiate (1:=(bh_nth a l ((a :: x) ++ (a0 :: y)) r) (length (a::x))).
         rewrite map_nth.
         replace ((bh_nth a l ((a :: x) ++ (a0 :: y)) r) (length (a::x))) with 
           ((bh_nth a0 (opl l (gl (a :: x))) (a0 :: y) r) 0).
         rewrite map_nth.
         unfold bh_nth.
         rewrite seq_nth.
         f_equal.
         rewrite firstn_app2.
         rewrite H3. rewrite H. rewrite seq_nth. rewrite minus_plus. simpl. reflexivity.

         assumption. omega.

         rewrite app_nth2. rewrite seq_nth. rewrite minus_plus.
         apply nth_indep. auto.
         assumption.
         omega.

         rewrite seq_nth. rewrite skipn_app2. rewrite <- plus_Sn_m.
         rewrite plus_Snm_nSm. rewrite minus_plus. auto.

         omega. assumption. assumption.
         
         unfold bh_nth. f_equal.
         rewrite firstn_app2. rewrite H3. rewrite H. rewrite <- minus_n_n. reflexivity.
         
         unfold ge. auto.

         rewrite app_nth2. rewrite <- minus_n_n. reflexivity.
         auto.

         rewrite skipn_app2. rewrite <- minus_Sn_m. rewrite <- minus_n_n. reflexivity.
         
         auto.
         auto.
 Qed.
  (* hide*)
End BH_SPEC.

Implicit Arguments bh_nth [L A R B].
Implicit Arguments nth_bh [L A R B ].
Implicit Arguments bh_comp [L A R B].
Hint Unfold bh_nth.
Hint Rewrite bh_length : length.
