 Unset Automatic Introduction.    (* coq 8.3 *)

Require Export CertifiedBSML.QuickStart.

Require Export CorrectParallelisation.Structure.
Require Export Datatypes.Lists.Lists.
Require Import Skel.BH.bh.
Require Import myTactics.

Require Import List.
Require Import Arith.
Require Import Arith.Min.
Require Import Omega.
Require Import EqNat.
Require Import FunctionalExtensionality.
Require Import LIFO.List.Triple_lists LIFO.List.Firstn_skipn LIFO.List.Fold LIFO.List.Inits_tails.
Require Import LIFO.List.Map LIFO.List.Notations LIFO.List.Nth LIFO.List.Rev LIFO.List.SeqNat. 
Require Import LIFO.List.Homomorphism.
Require Import LIFO.Algebra.AlgebraClasses.

Require Import ProgramCalculation.Tactics.
(* Require Import bmf_classes. *)
        
Global Generalizable All Variables.

Module Type TYPE(Bs : PRIMITIVES) 
  (Bsml_prop : Properties.TYPE Bs) 
  (ParSig : ParSig.TYPE Bs Bsml_prop) 
  (Bsml_base : Base.TYPE Bs Bsml_prop ParSig)  
  (Parallel : Parallelisation )
  (Partition : Lists.TYPE Bs Bsml_prop ParSig Bsml_base Parallel).

  Export Bs.
  Export Parallel.
  Export Partition.
  Export Bsml_base.
  Export Bsml_prop.


  Section BSML_BH.

  (** All bsml function ar defined with a weak return type, 
     removing the strong specification greatly improve compilation time 
     because Strong type are Heavy for the Type checker
     *)

    Variables (L A R B:Type).
    
    (** BH variables *)
    Variable (k: L -> A -> R -> B).
    Variable (gl: list A -> L).
    Variable (opl: L -> L -> L).
    Variable (gr: list A -> R).
    Variable (opr: R -> R -> R).

    (** applying k to a triple *)
    Definition k_tripl l r (trpl: list A * (A * list A)) :=
      match trpl with (ini,(a,tls)) => 
        k (opl l (gl ini)) a (opr (gr tls) r) end.
    Hint Unfold k_tripl.


    (** A definition of BH different from [bh_comp]
        This writing is actually closest to our definition, if slightly
        more complicated than the one on the nth element seen in bh.v *)
    Definition bh_seq (l:L) (lst: list A) (r:R) := 
      map (k_tripl l r) (inits_id_tails lst).
    (** Equivalence of these two writings *)
    Lemma bh_seq_bh : forall l lst r,
      bh_seq l lst r = bh_comp k gl opl gr opr l lst r.
    Proof.
      intros.
      destruct lst.
      simpl. unfold bh_seq. simpl. reflexivity.
      
      set (ls := a::lst).
      apply nthSameLengthEqual with (d:=k_tripl l r (ls,(a,nil))).
      transitivity (length ls).
      unfold bh_seq.
      rewrite map_length. rewrite inits_id_tails_length. reflexivity.
      rewrite bh_length. reflexivity.

      intros.
      unfold bh_seq.
      unfold bh_seq in H; rewrite map_length in H.
      rewrite inits_id_tails_length in H.
      rewrite map_nth.
      unfold inits_id_tails. rewrite combine_nth. rewrite combine_nth. (* ;[|assumption|]. *)
      rewrite inits_nth. rewrite tails_nth.
      unfold bh_comp.
      unfold ls.
      rewrite nth_indep with
        (d':=(bh_nth k gl opl gr opr a l (a::lst) r) 0%nat).
      rewrite map_nth.
      unfold bh_nth.
      rewrite seq_nth.
      rewrite plus_0_l. unfold k_tripl.
      reflexivity.
      
      assumption.
      autorewrite with length. assumption.
      assumption.
      autorewrite with length. assumption.
      autorewrite with length. assumption.
      autorewrite with length. 
      rewrite min_l; auto with arith.
    Qed.


    (** *** Preliminary functions to define our parallel definition of BH *)

    (** Data structure for data to be communicated  *)
    Inductive t_comm : Type :=
    | lcx : L -> t_comm
    | rcx : R -> t_comm
    | ncx : t_comm.

    (** function to decide what sender has to communicate to receiver  *)
    Program Definition comm_fun (sender:processor) l r (receiver:processor) :=
      match nat_compare sender receiver with
        | Lt => lcx l
        | Eq => ncx
        | Gt => rcx r
      end.

    (** List Parallelisation using list partition  *)
    (* Definition scatter (B: Type) (l: list B) := mkpar_nat (list_partition l). *)
    (* Implicit Arguments scatter [B]. *)

    (* Definition precompose (A B C : Type) (f : A -> B) (g : B -> C) (x : A) := g (f x) . *)

    (* Program Definition put_nat_processor (A : Type) (vf : par (nat -> A)): par (processor -> A) :=  *)
    (*   put *)
    (*   (apply *)
    (*     (mkpar *)
    (*       (fun _ : processor => precompose processor nat A natOfProcessor)) *)
    (*     vf) *)
    (*  . *)

    (** Computation of left and right "digest" and perform  communications *)
    Program Definition bsml_comp_bh_comm (lst: par (list A))  : par (processor -> t_comm) := 
      put (apply (apply (mkpar comm_fun) 
        (apply (replicate gl) lst)) (apply (replicate gr) lst)).

    (** function to compute digest of digests received from left  *)
    Definition fold_range_left_procs (f: processor -> L) (l:L) (i: nat) :=
      fold_left (fun (acc:L) x => opl acc (f x)) (firstn i (proj1_sig processors)) l.

    (** function to compute digest of digests received from right *)
    Definition fold_range_right_procs (f: processor -> R) (r:R) (i: nat) :=
      fold_right (fun x acc => opr (f x) acc) r (skipn i (proj1_sig processors)).

    (** Kind of no_some with default value for  data received from the left  *)
      Definition rem_lcx_comm (comm: processor -> t_comm) (default: L) (i: processor) :=
      match comm i with
        | lcx l => l
        | _ => default
      end.

    (** Kind of no_some with default value for  data received from the right *)
    Definition rem_rcx_comm (comm: processor -> t_comm) (default: R) (i: processor) :=
      match comm i with
        | rcx r => r
        | _ => default
      end.
     
    (** Parallel vector of folded values communicated from the left processors *)
    (** /!\ The  default value of rem_lcx_comm is never used, as fold_range_left iterate over left processors  *)
    Program Definition loc_lefts  (l:L) (comms: par(processor->t_comm)) :par L :=
      apply 
        (mkpar (fun i com => fold_range_left_procs (rem_lcx_comm com l) l (natOfProcessor i)))
        comms.
    (* (bsml_comp_bh_comm l lst r) *)

    (** Parallel vector of folded values communicated from the right processors *) 
    Program Definition loc_rights (r:R)  (comms: par(processor->t_comm)) : par R :=
      apply 
        (mkpar (fun i com => fold_range_right_procs (rem_rcx_comm com r) r ((natOfProcessor i)+1)))
        comms.
    (* (bsml_comp_bh_comm l lst r) *)

    (** ** Computation of BH using BSML *)
    Program Definition bh_bsml_comp (l:L) (vl: par (list A)) (r:R) : par (list B) :=
      let comms := bsml_comp_bh_comm vl  in
        apply (apply (apply 
          (replicate bh_seq) 
          (loc_lefts l  comms)) vl)
          (loc_rights  r comms).

    (** ** Now the hypotheses and lemmas / theorems of equivalence of
     these different BH computations *)
    Hypothesis gl_hom: is_homomorphism _ _ gl opl.
    Hypothesis gr_hom: is_homomorphism _ _ gr opr.

    (** lemmas extracting the homomorphism properties from the hypothesis *)
    Lemma opl_ass: bh.associativity _ opl. 
    Proof. 
      destruct gl_hom. assumption. 
    Qed.


    Instance app_ass (A:Type) :  LIFO.Algebra.AlgebraClasses.Associative (eqA:=eq)  (@app A).
    Proof. constructor. apply app_assoc_reverse. Qed.

    Instance opl_ass_c :  LIFO.Algebra.AlgebraClasses.Associative opl.
    Proof. constructor; auto using opl_ass. Qed.
    
    Lemma gl_neutral_r: forall b, opl b (gl nil) = b. 
    Proof. 
      destruct gl_hom. intuition. 
    Qed.
    Lemma gl_neutral_l: forall b, opl (gl nil) b = b. 
    Proof.
      destruct gl_hom. intuition. 
    Qed.

    Program Instance gl_id : LIFO.Algebra.AlgebraClasses.Neutral opl (gl nil).
    Solve Obligations using (constructor; auto using gl_neutral_l, gl_neutral_r).

    
    Lemma gl_opl_hom: forall x y, gl (x++y) = opl (gl x) (gl y). 
    Proof. 
      destruct gl_hom. intuition. 
    Qed.

    Lemma opr_ass: bh.associativity _ opr. 
    Proof. 
      destruct gr_hom. assumption. 
    Qed.

    Instance opr_ass_c :  LIFO.Algebra.AlgebraClasses.Associative opr.
    Proof. constructor; auto using opr_ass. Qed.
    
    Lemma gr_neutral_r: forall b, opr b (gr nil) = b. 
    Proof. 
      destruct gr_hom. intuition. 
    Qed.
    
    Lemma gr_neutral_l: forall b, opr (gr nil) b = b. 
    Proof. 
      destruct gr_hom. intuition. 
    Qed.

    Program Instance gr_id : LIFO.Algebra.AlgebraClasses.Neutral opr (gr nil).
    Solve Obligations using (constructor; auto using gr_neutral_l, gr_neutral_r).
    
    Lemma gr_opr_hom: forall x y, gr (x++y) = opr (gr x) (gr y). 
    Proof. 
      destruct gr_hom. intuition. 
    Qed.


    (** *** Now the lemmas and theorems about the properties of our
       computation *)
                    
    (** No communication if the sender is the receiver *)
    Lemma comm_fun_eq i l r: comm_fun i l r i = ncx.
    Proof.
      intros.
      unfold comm_fun.
      destruct i as [i i_proc]; simpl.
      replace (nat_compare i i) with Eq.
      reflexivity. 
      symmetry; apply  nat_compare_eq_iff.
    reflexivity.
    Qed.

    (** Communication to the left if the sender is greater than the receiver *)
    Lemma comm_fun_lt i l r j: 
      proj1_sig i< proj1_sig j -> 
      comm_fun i l r j = lcx l.
    Proof.
      intros.
      unfold comm_fun.
      rewrite nat_compare_lt in H.
      rewrite H.
      reflexivity.
    Qed.

    (** Communication to the right if the sender is lesser than the receiver *)
    Lemma comm_fun_gt i l r j: 
      proj1_sig i> proj1_sig j
      -> comm_fun i l r j = rcx r.
    Proof.
      intros.
      unfold comm_fun.
      rewrite nat_compare_gt in H.
      rewrite H.
      reflexivity.
    Qed.

    (** *** First results about the information gathered from the left and
    from the right *)

    (* (** *** First results about the information gathered from the left and *)
    (* from the right *) *)


    

    (* if i is in seq x y, i>=x  *)
    Lemma in_seq_ge i x y: In i (seq x y) -> i >= x.
      induction y.
      simpl.
      tauto.
      intro.
      replace (S y) with (y+1) in H by omega.
      rewrite seqAppN in H.
      apply in_app_or in H.
      destruct H.
      apply IHy. assumption.
      simpl in H.
      destruct H.
      omega.
      tauto.
    Qed.

    Instance opl_gen_assoc : AlgebraClasses.Associative (fun acc y =>opl acc y).
    Proof.
      constructor. destruct gl_hom.
      auto.
    Qed.

    Instance opl_l_id : AlgebraClasses.NeutralRight (fun acc y =>opl acc y) (gl nil).
    Proof.
      constructor. destruct gl_hom.
      destruct H0 as [H' H''].
      auto.
    Qed.
         

    (*  Proof that for any distributed list, digest of left parts computed at proc p is correct*)
    Lemma loc_lefts_def'  l  (plst: par (list A)) (p:processor) :
      get 
        (loc_lefts  l  
          (bsml_comp_bh_comm  plst)) p =  
        opl l
      (gl
        (fold_left  (fun x i => x++(get plst i))  (firstn (proj1_sig  p) (proj1_sig processors)) nil )).
    Proof.
      intros  l lst p.
      unfold loc_lefts.
      autorewrite with bsml.
      destruct p as (i,In).
      
      (*  induction over proc indice (nat) *)
      revert l; induction i; intros l.
      (* proc 0 *)
           unfold fold_range_left_procs.  simpl.
           destruct gl_hom. destruct H0. rewrite H0.
           reflexivity.

       (* proc i should receive  l + "sum" of all left procs summaries  *)
           simpl natOfProcessor in *.
           generalize dependent IHi.
           assert (iIn : i < bsp_p) by omega.
             destruct  (splitProcessors (processorOfNat  i iIn)) as [ l1 H]; destruct  H as [ l2 H]. 
             rewrite H.
           intros IHi.
           generalize(splitedProcessorsLengthLeft l1 (processorOfNat i iIn) l2 H).
           intros length_l1. simpl in length_l1.
           assert ( fst_Si :  firstn (S i) (l1 ++ processorOfNat i iIn :: l2) =  l1 ++ [processorOfNat i iIn]) .
               revert iIn H IHi.  rewrite length_l1. intros iIn H IHi. 
               rewrite firstn_app2.  f_equal.
               assert (S (length l1) - length l1 =1) by (omega).
               rewrite H0.
               reflexivity.
                   omega. 
           assert ( fst_i :  firstn  i (l1 ++ processorOfNat i iIn :: l2) =  l1) .
               revert iIn H IHi fst_Si.  rewrite length_l1. intros iIn H IHi fst_SI. 
               rewrite firstn_app2.  f_equal.
               assert (length l1 - length l1 =0) by (omega).
               rewrite H0.
               simpl. rewrite <- app_nil_end.
               reflexivity.
                   omega. 
          simpl proj1_sig in *.
          setoid_rewrite fst_i in IHi.
          setoid_rewrite fst_Si.
           rewrite (fold_left_f_inv _ (@app A) ).
           rewrite map_app.
           rewrite fold_left_app.
           simpl.
           destruct gl_hom as [opl_asso H0]; destruct H0 as [right_neutral  H0]; destruct H0 as [ left_neutral  hom_app].
           rewrite hom_app.
           rewrite <- opl_asso.
           rewrite <- (fold_left_f_inv _ (@app A) ).
           rewrite <- (IHi iIn l ).

           unfold fold_range_left_procs.
           Opaque firstn.
           simpl .
           Transparent firstn.
           rewrite H. 
           setoid_rewrite fst_i. 
           setoid_rewrite fst_Si.
           rewrite fold_left_f_inv.
           rewrite map_app.
           rewrite List.fold_left_app.
           simpl.
           replace (map
           (rem_lcx_comm
              (get (bsml_comp_bh_comm lst)
                 (exist (fun pid : nat => pid < bsp_p) (S i) In)) l) l1)
             with 
               (map
                 (rem_lcx_comm
                   (get (bsml_comp_bh_comm lst)
                     (exist (fun pid : nat => pid < bsp_p)  i iIn)) l) l1)
           .
           unfold rem_lcx_comm. unfold bsml_comp_bh_comm.
           autorewrite with bsml.
           rewrite comm_fun_lt.
           rewrite <- fold_left_f_inv.
           reflexivity.
           simpl. omega.
                  (* as pid in l1 are lesser than i, communicated value is the same, wether we ask it from i or S i *)
         apply map_assumption.
         intros x H0.
         destruct x.
         (*  proving that x is lesser than i *)
         assert (x<i).
            destruct (in_split  _ l1 H0) as [l1' H0']; destruct H0' as [l2' H0''] .
            assert ( (proj1_sig processors) =  l1' ++ exist (fun pid : nat => pid < bsp_p) x l0::(l2'++processorOfNat i iIn :: l2)).
            replace  (exist (fun pid : nat => pid < bsp_p) x l0::(l2'++processorOfNat i iIn :: l2)) with  ([exist (fun pid : nat => pid < bsp_p) x l0] ++(l2'++processorOfNat i iIn :: l2)) by reflexivity.
            rewrite (app_assoc _ l2' _).
            rewrite (app_assoc _ _ ( processorOfNat i iIn :: l2)).
            simpl.
            setoid_rewrite <- H0''.
            apply H.
            generalize (splitedProcessorsLengthLeft l1' _ _ H1).
            intros value_x. simpl natOfProcessor in value_x.
            rewrite H0'' in length_l1. simpl in *. rewrite  value_x. rewrite length_l1.
            autorewrite with length.
            simpl.
            omega.
         unfold rem_lcx_comm. unfold bsml_comp_bh_comm.
         autorewrite with bsml.
         rewrite 2 comm_fun_lt.
       reflexivity.
       simpl; auto.
       simpl; auto.
     Qed.
     (*  Wouhouuuuuu !!  *)



    (*  Proof that for any distributed list, digest of right parts computed at proc p is correct*)
    Lemma loc_rights_def'   r (plst: par (list A)) (p:processor) :
      get  (loc_rights  r  
        (bsml_comp_bh_comm (plst) )) p =  
        (opr(gr (fold_left  (fun x i => x++(get plst i))  (skipn (proj1_sig p +1) (proj1_sig processors)) nil) ) r) .
    Proof.
      intros   r lst p.
      unfold loc_rights.
      autorewrite with bsml.
      unfold fold_range_right_procs.

      (* Spliting the list of processors *)
      pose (H:=splitProcessors p).
      destruct H as [l1 H]. destruct H as [l2 H].
      pose (H0:=splitedProcessorsLength l1 p l2 H).
      pose (H1:=splitedProcessorsLengthLeft l1 p l2 H).
      rewrite H.
      generalize H H0 H1. generalize r l1 p. clear H H0 H1 l1 p r.
     (* induction over the list of processor at the right of p *)
      induction l2.
      (* case l2 is nil *)
          intros.
          simpl in H0. assert (length l1 = max_pid).
          rewrite max_pid_S_bsp_p in H0. omega.
          rewrite skipn_app2. Focus 2.  rewrite max_pid_S_bsp_p in H0. unfold natOfProcessor. omega.
          rewrite H1. destruct p.  simpl in H1. simpl proj1_sig. revert l H; rewrite H1.
          intros l0 H.
          rewrite minus_plus.
          simpl.
          replace (length l1 + 1) with (length (l1 ++ [exist (fun pid : nat => pid < bsp_p) (length l1) l0]))  by (autorewrite with length; reflexivity).
          rewrite skipn_length_l. 
          simpl.
          rewrite gr_neutral_l.
          reflexivity.

          intros.
          assert ((proj1_sig processors) = (l1++p::nil)++(a::l2)) as H'.
             rewrite <- ass_app.
             rewrite <- app_comm_cons.
             auto.
          pose (H'0 := splitedProcessorsLength (l1 ++ p :: nil) a l2 H').
          pose (H'1 := splitedProcessorsLengthLeft (l1 ++ p :: nil) a l2 H').
          pose (IH:= IHl2 r (l1++p::nil) a H' H'0 H'1).
          generalize IH; clear IH; intro IH; clear IHl2.
          destruct p as [p p_proc].
          (* eliminating skipn in goal *)
          simpl natOfProcessor in *. simpl proj1_sig.
               replace (p + 1) with (length  (l1 ++ [exist (fun pid : nat => pid < bsp_p) p p_proc])) by(
                 simpl in H1;generalize dependent p_proc; rewrite H1; intros; autorewrite with length; reflexivity).
               rewrite  skipn_app2.
               Focus 2.
                autorewrite with length; omega.
               autorewrite with length. simpl. replace (length l1 + 1 - length l1) with 1 by auto with arith.
               simpl.
          (* eliminating skipn in Induction hypothesis *)
               generalize (splitedProcessorsLengthLeft _ a _ H' ).
               intros a_Sp.
               destruct a as [Sp Sp_proc]. 
               simpl natOfProcessor in *. simpl proj1_sig in *.
               replace (Sp + 1) with (length  ((l1 ++ [exist (fun pid : nat => pid < bsp_p) p p_proc]) ++[ exist (fun pid : nat => pid < bsp_p) Sp Sp_proc] ))in IH .
               Focus 2.
                    generalize dependent Sp_proc. rewrite a_Sp; intros; autorewrite with length. simpl. autorewrite with length. reflexivity.
                    rewrite  skipn_app2 in IH.
               Focus 2.
                    autorewrite with length; simpl; omega.  
              autorewrite with length in IH; simpl in IH.
               replace ( length l1 + 1 + 1 - (length l1 + 1)) with 1 in IH by auto with arith.
               simpl in IH.
           (* preparing the right hand side for rewriting with IH *)

               rewrite (fold_left_assoc_with_id_generalized_prop (f:=@app A)).
               let H' := fresh H' in destruct gr_hom as [opr_asso H']; destruct H' as [right_neutral  H']; destruct H' as [ left_neutral  hom_app].
               rewrite hom_app.
               rewrite opr_asso. 
               (*  rewriting IH *)
               rewrite <- IH.
               
               f_equal.
               (* treating the new element *)
               unfold bsml_comp_bh_comm.
               unfold rem_rcx_comm.
               autorewrite with bsml.
               rewrite comm_fun_gt.
               reflexivity.
                         generalize dependent Sp_proc. rewrite  a_Sp ; intros; autorewrite with length ; simpl;autorewrite with length ; simpl; omega.
                (* treating the list euqiv by induction : 
                   need to prove that all proc in l2 are greater than  S p to ensure the comm are the same *)
               rewrite fold_right_f_inv. rewrite (fold_right_f_inv processor _ _ opr). 
               f_equal. 
               apply map_assumption.
               intros x InXl2.
               destruct x as [x x_proc].
               assert (x  > Sp).
               destruct (in_split  _ l2 InXl2) as [l1' H0']; destruct H0' as [l2' l2_split] .
               assert ( (proj1_sig processors) =  
                 ((l1 ++ [exist (fun pid : nat => pid < bsp_p) p p_proc]) ++ 
                 exist (fun pid : nat => pid < bsp_p) Sp Sp_proc :: l1') ++ exist (fun pid : nat => pid < bsp_p) x x_proc :: l2' ).
                       simpl;rewrite H. rewrite l2_split. repeat rewrite <- app_assoc.  reflexivity.
                 
               generalize (splitedProcessorsLengthLeft _ _ l2' H2).
            intros value_x. simpl natOfProcessor in value_x.
            rewrite a_Sp. simpl in value_x; rewrite  value_x. 
            autorewrite with length.
            simpl.
            omega.
            (* unfolding comm *)
         unfold rem_rcx_comm. unfold bsml_comp_bh_comm.
         autorewrite with bsml.
         rewrite 2 comm_fun_gt.
       reflexivity.
       simpl; auto.
       simpl; auto.
       (* still need to prove Sp > p need to  *)
     autorewrite with length in a_Sp. rewrite <- H1 in a_Sp. simpl in a_Sp. rewrite a_Sp in H2. omega.
     Qed.
     (*  Wouhouuuuuu !!  *)





      (*******************************)
      (*******************************)
      (*******************************)
      (** Optimized Sequential Bh versions  *)
      (*******************************)
      (*******************************)
      (*******************************)

    (* (* version fausse de  bh_seq_opt   *) *)
    (* (*       k (opl l (gl nil)) a (opr (gr nil) r) = k (opl l (gl (a :: nil))) a (opr (gr (a :: nil)) r)   *) *)
    (* Definition opfl :=  *)
    (*   fun (macc : (L*list L)) ( p:A)  => *)
    (*     let (m,acc):=macc in *)
    (*       let m := opl m (gl (p::nil)) in  *)
    (*         (m, m::acc). *)

    (* Definition opfr :=  *)
    (*   fun p (macc :R* list R) =>  *)
    (*     let (m,acc):=macc in *)
    (*       let m := opr (gr (p::nil)) m in  *)
    (*         (m, m::acc). *)
    Definition opfl :=
      fun (macc : (L*list L)) ( p:A)  =>
        let (m,acc):=macc in
          let m' := opl m (gl (p::nil)) in
            (m', m::acc).

    Definition opfr :=
      fun p (macc :R* list R) =>
        let (m,acc):=macc in
          let m' := opr (gr (p::nil)) m in
            (m', m::acc).
    
    (** A definition of BH different from [bh_seq] more efficient *)
    Definition left lv l  :=
      rev'(snd(fold_left opfl lv (l,nil))).

    Definition right  lv r :=
      snd(fold_right opfr (r,nil) lv).

    Definition bh_seq_opt (l:L) (lst: list A) (r:R) :=
      let ll:= left  lst l in let lr := right lst r in
        map3 k ll lst lr.
    
    (* Lemma opf_app :  forall l l1 l2 lst'  p, opfl lst (l,l1++l2) *)

    (** Length Properties *)
    Lemma length_left_generalized:
      forall l l',
        length(snd(fold_left opfl l l')) = (length l) + length(snd l').
    Proof.
      intros l. 
      induction l.
      reflexivity.
      simpl.
      setoid_rewrite IHl.
      destruct l'.
      simpl. 
      assert(forall n, n + 1 = S n) as plus1.
        induction n. reflexivity. simpl. rewrite IHn. reflexivity.
      rewrite <- plus1. 
      rewrite <- (plus1 (length l + length l1)). 
      rewrite plus_assoc.
      reflexivity.
    Qed.
    Lemma left_length : forall lv l , length (left lv l) = length lv.
    Proof.
      unfold left.
      intros lv l.
      setoid_rev'_to_rev.
      setoid_rewrite rev_length.
      setoid_rewrite length_left_generalized.
      setoid_rewrite plus_0_r.
      reflexivity.
    Qed.

Hint Rewrite left_length : length.

    Lemma length_opfr: 
      forall p l, 
        length(snd(opfr p l)) = S(length(snd l)).
    Proof.
      intros p l.
      destruct l.
      reflexivity.
    Qed.

    Lemma right_length : forall lv r, length (right lv r) = length lv.
    Proof.
      intros lv r.
      unfold right.
      induction lv. 
      reflexivity.
      simpl. rewrite length_opfr. rewrite IHlv. reflexivity.
    Qed.

Hint Rewrite right_length : length.
    

Lemma  opfl_app  :`(snd(opfl (l ,lst) a) = snd (opfl (l ,nil) a) ++lst).
Proof. 
  intros l lst a.
  induction lst;  reflexivity.
Qed.

Lemma  opfl_app_generalized  :`(snd (fold_left opfl lst (l,lst') ) = snd (fold_left opfl lst (l,nil) ) ++ lst').
Proof. 
  induction lst.  
  reflexivity.
  intros ; simpl.  rewrite (IHlst (opl l (gl (a :: nil))) (l :: nil)) .  rewrite IHlst.  rewrite <- app_assoc.
  reflexivity.
Qed.

Lemma fold_opfl_map_gl : `((@snd L (list L))  (fold_left opfl  lst (l,nil))  =  rev (map (fun lst => opl l (gl lst)) (inits lst))).
Begin.
unfold is_homomorphism in *. destruct gl_hom as [opl_assoc [right_unit [left_unit gl_hom_destr]]].
generalize dependent l.
induction lst.
(* empty list case is trivial *)
[].
(* non empty list *)
intros l. 
rewrite opfl_app_generalized.

check ( snd (fold_left opfl (a :: lst) (l, [])) ++ [] = rev (map (fun lst0 : list A => opl l (gl lst0)) (inits (a :: lst)))) is GOAL.

RHS={simpl}
(rev  (map (fun lst0 : list A => opl l (gl lst0)) (inits ([a] ++ lst)))).
={rewrite decompInits}
 (rev (map (fun lst0 : list A => opl l (gl lst0)) (inits [a] ++ map (fun z => [a]++z) (inits lst)))).
={simpl }
 (rev ( opl l (gl (rev' []))
   :: map (fun lst0 : list A => opl l (gl lst0))
        (map (fun z : list A => a :: z) (inits lst)))).
={simpl;rewrite map_map }
 ( rev (opl l (gl ([]))
   :: map (fun lst0 : list A => opl l (gl (a::lst0)))
   (inits lst))).
={simpl}
 ( rev (opl l (gl ([]))
   :: map (fun lst0 : list A => opl l (gl ([a]++ (lst0))))
   (inits lst))).
assert (H': forall lst : list A, (fun  (lst0 : list A) => opl l (gl ([a] ++ lst0))) lst = (fun  (lst0 : list A) => opl (opl l  (gl [a]))  (gl lst0)) lst).
intros lst0.
rewrite gl_hom_destr.  rewrite  opl_assoc. reflexivity.
={rewrite (map_ext _ _ H')}
 ( rev (opl l (gl ([]))
   :: map (fun lst0 : list A => opl (opl l  (gl [a]) ) (gl lst0))   (inits lst))). 
={auto}
 (rev (map (fun lst0 : list A => opl (opl l  (gl [a]) ) (gl lst0))   (inits lst)) ++([opl l (gl ([]))])) . 
={rewrite (IHlst); rewrite right_unit}
(snd (fold_left opfl lst (opl l (gl [a]), [])) ++[l]) .
LHS 
={rewrite app_nil_r}
(snd (fold_left opfl (a :: lst) (l, [])) ).
 ={simpl} 
( snd (fold_left opfl lst (opl l (gl [a]), [l]))).
 ={rewrite opfl_app_generalized}
(snd (fold_left opfl lst (opl l (gl [a]), [])) ++[l]).
[].
Qed.

Lemma fold_opfl_map_gl_generalized : `((@snd L (list L))  (fold_left opfl  lst (l,l_lst))  =  rev (map (fun lst' => opl l (gl lst')) (inits lst)) ++ l_lst).
Begin.
LHS 
={rewrite opfl_app_generalized}
( snd (fold_left opfl lst (l, [])) ++l_lst).
RHS 
={rewrite fold_opfl_map_gl}
( snd (fold_left opfl lst (l, [])) ++l_lst).
[].
Qed.




    Lemma  opfr_app  :`(snd(opfr  a (r ,lst ))= snd (opfr a (r,nil) ) ++lst).
      Proof. 
      intros a r lst.
      induction lst;  reflexivity.
    Qed.



Lemma  opfr_fst_fold   :   `(forall lst', fst (fold_right opfr (r, lst') lst) =  opr (gr lst) r).
  intros r lst. revert r.
  destruct gr_hom as [opr_assoc [right_unit [left_unit gr_hom_destr]]].
  induction lst.
  simpl.  auto.
Begin.
  RHS 
  ={simpl }
  (opr (gr ([a]++ lst)) r).
  ={rewrite gr_hom_destr }
  (opr (opr (gr [a])  (gr lst)) r).
  ={rewrite opr_assoc }
  (opr  (gr [a])  (opr (gr lst) r) ).
  ={rewrite IHlst}
  (opr (gr [a])  (fst (fold_right opfr (r, lst') lst))).
  LHS
={by def fold_right}
(fst (opfr a (fold_right opfr (r, lst') lst))).
GOAL idtac.
set (H:= fold_right opfr (r, lst') lst).  unfold opfr;  destruct H. simpl. reflexivity.
Qed.
    (* Lemma  opfr_app_generalized  :`(snd (fold_right opfr (r,lst') lst  ) = snd (fold_right opfr  (r,[]) lst ) ++ lst'). *)
    (*   Proof. *)
    (*     intros r lst' lst. revert r lst'. *)
    (*   induction lst. *)
    (*   reflexivity. *)
    (*   intros ; simpl. *)
    (*   assert (H  :forall lst'0,snd (opfr a (fold_right opfr (r, lst'0) lst)) = *)
    (*       snd ((opr (gr [a]) (fst (fold_right opfr (r, lst'0) lst))), (fst (fold_right opfr (r, lst'0) lst))::(snd  (fold_right opfr (r, lst'0) lst)))) by( *)
    (*   intros lst'0 ;set (H:= fold_right opfr (r, lst'0) lst);  unfold opfr;  destruct H; reflexivity). *)
    (*   rewrite H.   rewrite H. simpl.  rewrite <- IHlst. rewrite <- app_assoc. *)
    (* reflexivity. *)
    (* Qed. *)

Lemma opfr_compose :   `(opfr a (opfr a0 (fold_right opfr (r, nil) lst)) = (opr (gr (a::nil))   (opr (gr (a0 :: lst)) r) , (opr (gr (a0 :: lst)) r) :: snd (opfr a0 (fold_right opfr (r, nil) lst)))).
Proof.
  intros a a0 r lst; revert a a0 r. 
  induction lst; simpl. 
    reflexivity.
    setoid_rewrite IHlst.
    unfold opfr.
    intros a0 a1 r.
    rewrite <- ?opr_ass.
    rewrite <- ?gr_opr_hom.
    reflexivity.
  Qed.
Lemma fold_opfr_map_gr : `((@snd R (list R))  (fold_right opfr (r,[])   lst)  =  map (fun lst => opr (gr lst) r) (tails lst)).
Begin.
unfold is_homomorphism in *. destruct gr_hom as [opr_assoc [right_unit [left_unit gr_hom_destr]]].
generalize dependent r.
induction lst.
(* empty list case is trivial *)
[].
(* non empty list *)
intros r. 
RHS={simpl}
 (map (fun lst0 : list A => opr (gr lst0) r ) (tails ([a] ++ lst))).
={rewrite decompTails}
 ( map (fun lst0 : list A => opr (gr lst0) r)
     (map (fun z : list A => z ++ lst) (tails [a]) ++ tails lst)).
={simpl }
 (opr (gr lst) r :: map (fun lst0 : list A => opr (gr lst0) r) (tails lst)).
={rewrite IHlst}
 (opr (gr lst) r ::  snd (fold_right opfr (r, []) lst) ).
LHS 
={by def fold_right}
(snd (opfr a (fold_right opfr (r, [])  lst))). 
assert (H  :forall lst'0,snd (opfr a (fold_right opfr (r, lst'0) lst)) =  fst (fold_right opfr (r, lst'0) lst)
      :: snd (fold_right opfr (r, lst'0) lst)) by( 
    intros lst'0 ;set (H:= fold_right opfr (r, lst'0) lst);  unfold opfr;  destruct H; reflexivity).  
rewrite H. simpl.
LHS
={rewrite (opfr_fst_fold)}
( opr (gr lst) r ::snd (fold_right opfr (r, []) lst)).
[].
Qed.

(* Lemma fold_opfr_map_gr_generalized :  `((@snd R (list R))  (fold_right opfr (r,r_lst)   lst)  =  map (fun lst => opr (gr lst) r) (tails lst)++r_lst). *)
(* Begin. *)
(* rewrite <- fold_opfr_map_gr. *)
(* myadmit bsml_bh  fold_opfr_map_gr_generalized. *)
(* Qed. *)
  
  


    
    (** Equivalence of these two writings *)
    Lemma bh_seq_bh_seq_opt : forall l lst r,
      bh_seq l lst r = bh_seq_opt l lst r.
    Proof.
      intros l lst.
      generalize dependent l.
      unfold bh_seq. unfold bh_seq_opt.
      induction lst.
      (* case lst = Nil *)
      unfold map3. simpl. reflexivity.
      (* case lst = Cons *)
      intros l r.
      simpl.
      apply nthSameLengthEqual with (d:=k_tripl l r ((a::lst),(a,nil))).  
      autorewrite with length.  repeat rewrite min_idempotent; reflexivity.     

      intros n H.
      assert (nlst :n < length (a :: lst) ) by (rewrite map_length in H; rewrite inits_id_tails_length in H; assumption).
      assert (nmap3 : n < length (map3 k (left (a :: lst) l) (a :: lst) (right (a :: lst) r)) ) by
        (rewrite map3_length; rewrite left_length ; rewrite right_length;  rewrite  ?min_idempotent; assumption). 
      assert ( nla : n< length (left (a :: lst) l)) by ( rewrite left_length ;simpl; auto with arith).
      assert ( nlc : n< length (right (a :: lst) r)) by ( rewrite right_length ;simpl; auto with arith).      
      assert ( nlb: n< length  (a :: lst)) by ( simpl; auto with arith).      
      rewrite <- nth'_nth with (pre:=nmap3).
      rewrite  (map3_nth _ _ _ _ nla nlb nlc)  .
      rewrite map_nth. 
      unfold inits_id_tails. unfold inits. simpl. 
      rev'_to_rev. simpl. 
      rewrite combine_nth. 
      Focus 2.         
            ( simpl; autorewrite with length; simpl in nlst;simpl; auto with *).
      Focus 2.
            ( simpl; autorewrite with length; rewrite min_idempotent;  auto with *).
      assert( left_part : (opl l (gl (nth n (rev (map (@rev A) (tails (rev lst ++ a :: nil)))) (a :: lst)))) = nth' n (left (a :: lst) l) nla ).
      revert nla.   unfold left. rewrite rev'_rev.
      rewrite tails_app_end. rewrite map_app.  rewrite map_map. simpl. rewrite ?rev_unit.  rewrite <- map_rev.
      destruct  n.
      simpl. rewrite opfl_app_generalized. autorewrite with list. unfold nth'. simpl. firstorder.
      simpl. 
      intro.       
      revert nla.  rewrite  fold_opfl_map_gl_generalized.
      rewrite rev_unit. rewrite rev_involutive. intro nla. rewrite (nth'_nth  _ nla  (opl l (gl (a :: lst)))). 
      LHS rewrite <- map_nth.  rewrite map_map.
={rewrite (map_ext (fun x : list A => gl (rev (x++[a])))(fun x : list A => gl (a::rev (x)))) ;[|intro; rewrite rev_unit;reflexivity]}
 (opl l
     (nth n
        (map (fun x : list A => gl (a::rev (x))) (rev (tails (rev lst))))
        (gl (a :: lst)))).  
={rewrite  (map_map (@rev A))}
(  opl l
     (nth n (map (fun x : list A => gl (a :: x)) (map (@rev A)  (rev (tails (rev lst)))))
        (gl (a :: lst)))).
={ rewrite map_rev }
( opl l
     (nth n
        (map (fun x : list A => gl (a :: x))
           (rev (map (@rev A)(tails (rev lst))))) (gl (a :: lst)))
).
={rev_to_rev'}
( opl l
     (nth n
        (map (fun x : list A => gl (a :: x))
           (rev' (map (@rev' A)(tails (rev' lst))))) (gl (a :: lst)))
).
={by def inits}
( opl l
     (nth n
        (map (fun x : list A => gl (a :: x))
           (inits lst)) (gl (a :: lst)))
).
rewrite <- map_nth; rewrite map_map.
unfold is_homomorphism in *. destruct gl_hom as [opl_assoc [right_unit [left_unit gl_hom_destr]]].
rewrite  (map_ext (fun x : list A => opl l (gl (a :: x))) (fun x : list A => opl l (opl (gl [a]) (gl x))));[|intros; rewrite <- gl_hom_destr; reflexivity].
RHS simpl.
rewrite (map_ext (fun lst' : list A => opl (opl l (gl [a])) (gl lst')) (fun lst' : list A => opl l (opl  (gl [a])  (gl lst'))));[|intros;rewrite opl_assoc; reflexivity].
[].
unfold k_tripl.

assert( middle_part : fst(nth n ((a, lst) :: combine lst (tails lst)) (a, []) ) = nth' n (a :: lst) nlb).
rewrite <- map_nth.
LHS ={ by def combine}
(
nth n (map (@fst A (list A)) (combine (a::lst) (lst::(tails lst)))) (fst (a, (@nil A)))).
 rewrite map_fst_combine.
rewrite <- (nth'_nth  _ nlst _).
[].
simpl.
autorewrite with length.
auto with arith.

assert( right_part : opr (gr (snd (nth n ((a, lst) :: combine lst (tails lst)) (a, [])))) r = nth' n (right (a :: lst) r) nlc).
LHS ={ rewrite <- map_nth}
( 
opr (gr (nth n (map (@snd A (list A)) ((a, lst) :: combine lst (tails lst))) (@snd A (list A) ( (a, []))))) r
).
 ={ by def combine}
( 
opr (gr (nth n (map (@snd A (list A)) (combine (a::lst) (lst::(tails lst)))) (@snd A (list A) ( (a, []))))) r
).
={ rewrite map_snd_combine;[|simpl;autorewrite with length;auto with arith]}
( 
opr (gr (nth n ( lst::(tails lst)) (@snd A (list A) ( (a, []))))) r
).
RHS revert nlc; unfold right.
rewrite fold_opfr_map_gr.
intro nlc;
rewrite (nth'_nth  _ _ ((fun lst0 : list A => opr (gr lst0) r) [])   ).
rewrite (map_nth (fun lst0 : list A => opr (gr lst0) r)). 
[].
rewrite <- left_part.
rewrite <- middle_part.
rewrite <- right_part.
set (H' :=  nth n ((a, lst) :: combine lst (tails lst)) (a, [])).
destruct H'.
reflexivity.
Qed.
      
(** Computation of BH using BSML *)
Program Definition bh_bsml_opt_comp (l:L) (vl: par (list A)) (r:R)  : par (list B) :=
  let comms := bsml_comp_bh_comm  vl  in
    apply (apply (apply 
      (replicate bh_seq_opt) 
      (loc_lefts l  comms)) vl)
    (loc_rights  r comms).


Lemma bh_bsml_opt_comp_bh_bsml_comp (l : L) (vl : par (list A)) (r : R) :
bh_bsml_opt_comp l vl r  = bh_bsml_comp l vl r.
Proof.
  unfold bh_bsml_opt_comp ; unfold bh_bsml_comp.
  assert (fun_ext_eq : bh_seq_opt = bh_seq). 
  repeat (apply functional_extensionality;intros). symmetry; apply bh_seq_bh_seq_opt.
  rewrite fun_ext_eq.
  reflexivity.
Qed.

(* Theorem  bh_bsml_opt_comp_bh l r lst: *)
(*       listOfParList (bh_bsml_opt_comp l (scatter lst) r) = *)
(*       bh_comp k gl opl gr opr l lst r. *)
(* Proof. *)
(*   rewrite bh_bsml_opt_comp_bh_bsml_comp.  *)
(*   apply bh_bsml_bh_gen. *)
(* Qed. *)


Lemma bh_bsml_opt_comp_bh_aux : forall l r plst p,
  get (bh_bsml_opt_comp l plst r) p =
  bh_comp k gl opl gr opr
         (opl l (gl (fold_left  (fun x i => x++(get plst i))  (firstn (proj1_sig  p) (proj1_sig processors)) nil )))
         (get plst p)
        (opr(gr (fold_left  (fun x i => x++(get plst i))  (skipn (proj1_sig p +1) (proj1_sig processors)) nil) ) r) .
Proof.
  intros l r plst p.
  destruct gl_hom as [assoc gl_props]; destruct gl_props as [ right_id_opl gl_props]; destruct gl_props as [left_id_opl opl_gl_app].
  unfold bh_bsml_opt_comp. 
  (* revert l r plst. *)
  (* destruct p as [p p_proc]. *)
  (* induction p;   intros l r plst. *)
  Opaque minus.
  autorewrite with bsml bsmlbase.
   rewrite <- bh_seq_bh_seq_opt.
  rewrite bh_seq_bh.
rewrite <- loc_lefts_def'.
rewrite <- loc_rights_def'.
  f_equal.
Qed.
Hint Rewrite bh_bsml_opt_comp_bh_aux : bsml.

    Instance opl_gl_homo: Homomorphism' opl gl.
    Proof.
      firstorder.
    Defined.
    Instance opr_gr_homo: Homomorphism' opr gr.
    Proof.
      firstorder.
    Defined.

    Theorem  bh_bsml_opt_comp_bh : forall (l : L) (r : R) (plst : par (list A)),
      listOfParList (bh_bsml_opt_comp l plst r)  =
      bh_comp k gl opl gr opr l (listOfParList ( plst)) r .
    Proof.
      intros l r plst.

      (* transforming left hand side for using bh_bsml_opt_comp_bh_aux *)
      set (lst:=(listOfParList plst)).
      assert (lst =(listOfParList plst)) by reflexivity.
      revert l r  H; generalize lst; clear lst; intro lst; revert plst.
      induction lst.

      (*case where the parallel list is empty *)
          intros plst l r H'. 
          simpl.
          rewrite listOfParList_fold_left_app_map_get. 
          rewrite (map_ext _ (fun p : {pid : nat | pid < bsp_p} =>
            bh_comp k gl opl gr opr
            (opl l
              (gl
                (fold_left
                  (fun (x : list A) (i : processor) => x ++ get plst i)
                  (firstn (proj1_sig p) (proj1_sig processors)) 
                  []))) ([])
            (opr
              (gr
                (fold_left
                  (fun (x : list A) (i : processor) => x ++ get plst i)
                  (skipn (proj1_sig p + 1) (proj1_sig processors)) 
                  [])) r))).
          (* rewrite <- fold_left_f_inv. *)
          (* apply fold_left_const. *)
          Opaque processors.
          simpl.
          Transparent processors.
          
          induction (proj1_sig processors).
              reflexivity. 
             simpl. rewrite IHl0. reflexivity.
         intro p.
         rewrite (bh_bsml_opt_comp_bh_aux l r plst p).
         f_equal.
         symmetry in H'.
         apply (parallel_list_nil _ plst H').

      (* Case where the parallel list is not empty (a::lst) *)
      intros plst l r H. 
      (* rewriting homomorphism properties *)
      change (a :: lst)with([a] ++ lst); rewrite bh_hom;auto.
      (* introducing parallel vector decomposition into three part :
         << [] |...| [] |  a::sublst | lst_{p+1} | ... | lst_{maxPid} >>
         and introducing the plst' parallel vector 
         << [] |...| [] |  sublst | lst_{p+1} | ... | lst_{maxPid} >>
         *)

      generate_sub_par_list plst.      
      symmetry in lopl_plst'.
      (* using the induction hypothesis  we can say that bh_comp apply to lopl plst is 
         like bh_comp _ [a] _ ++  bh_bsml _ plst' _  ,
         [a] being used in the "from left" element of the parallel computations over plst'
         
*)
      rewrite <-  (IHlst _ _ _ lopl_plst').
      clear IHlst.

      (* using decomp_lopl tactics, we switch to a  representation of the sequentialisation of the result vector splitted in three parts
         - the first one containing the result of the computation where local list were empty
         - the second, concerning the value at prossesor [p] where is stocked the first element of the sequentialised list (a::lst) in plst'
         -the third concerning the rest of the parallel list, where plst and plst' have the same values.
         *)
      (* Set Ltac Debug. *)
      (*    Opaque proj1_sig. *)
      decomp_lopl p.

      (* pretending that the result of bh over empty list is empty (proved later), 
         the concatenation give an empty list *)
      rewrite  (fold_left_all_nil (map (fun x : processor => get (bh_bsml_opt_comp l plst r) x) lg)).     
      (* produce a subgoal, which ask thatthe result of bh over empty list is empty (for plst)         
         See at the end of the proof
         *)
      rewrite  (fold_left_all_nil (map (fun x : processor => get (bh_bsml_opt_comp (opl l (gl [a])) plst' r) x) lg)).     
      (*produce a subgoal, which ask that the result of bh over empty list is empty (for plst')
         See at the end of the proof
         *)

Unset Ltac Debug.
      simpl.

      (* at proc p,  get (bh de plst) p = (bh_comp over [a])++ get (bh de plst') p  *)
      rewrite  bh_bsml_opt_comp_bh_aux .
      rewrite  bh_bsml_opt_comp_bh_aux.
      rewrite get_plst_p.
      change (a ::sublst)with([a] ++ sublst). rewrite bh_hom;auto.
      Opaque processors.
      simpl.
      unfold  is_homomorphism  in *.
      destruct gr_hom as [opr_assoc H']; destruct H' as [opr_id_r H'] ;destruct H' as [opr_idl gr_app].
      destruct gl_hom as [opl_assoc H']; destruct H' as [opl_id_l H'] ;destruct H' as [opl_idl gl_app].
      (* Simplifying result of communication using known facts about elements at left and right of [p] *)

      decomp_one_lopl p plst.   
      decomp_one_lopl p plst'.
      f_equal.                  
            (* Value computed at the place of the first element is the same as the first value computed sequentially  *)
             f_equal.
                (* incomming value from the left ok  *)
                rewrite right_neutral. reflexivity.             
                (* incomming value from the left ok  *)
             rewrite <- (AlgebraClasses.associative (op:=opr)).
             rewrite <- gr_app.
             f_equal. f_equal.
             rewrite lopl_plst'.
             decomp_lopl p.
             reflexivity.
             (* Values computed after the first element are the same *)
             f_equal.
             (* Values computed at proc p are the same *)
                   f_equal.
                           (*Incomming data from left are the same  *)
                       rewrite right_neutral.
                       rewrite right_neutral.
                       reflexivity.
    (* After proc p, values computed  are the same *)
                  f_equal.
                       apply map_assumption.  intros p' in_p'_ld.
                       (* getting result of local application of bh *)
                       rewrite bh_bsml_opt_comp_bh_aux.
                       rewrite  bh_bsml_opt_comp_bh_aux.
                       (* reusing decomposition around [p] to simplify communication  *)
                       decomp_one_lopl p plst.
                       decomp_one_lopl p plst'.
                       clear length_lg_p0 length_lg0 length_lg_p1 length_lg1  length_lg_p2 length_lg2 . 
                       assert ((proj1_sig p) < (proj1_sig p')) by (
                         apply (splitedProcessorsRightSup  _ _ _ processors_decomp _ in_p'_ld)).
                       replace (proj1_sig p') with  (proj1_sig p +((proj1_sig p')- (proj1_sig p))) by omega.
                       autorewrite with length in length_lg .
                       replace  (firstn (proj1_sig p + (proj1_sig p' - proj1_sig p))
                         (map (fun x : processor => get plst x) lg ++
                           map (fun x : processor => get plst x) (p :: ld)))
                         with  
                           (map (fun x : processor => get plst x) lg ++ firstn  (proj1_sig p' - proj1_sig p) (
                             map (fun x : processor => get plst x) (p :: ld))).
                       Focus 2.
                               rewrite firstn_app2. 
                               autorewrite with length.
                               rewrite length_lg.
                               replace ( proj1_sig p + (proj1_sig p' - proj1_sig p) - proj1_sig p) with  (proj1_sig p' - proj1_sig p) by (clear length_lg_p length_lg ;omega).
                               reflexivity.
                               autorewrite with length.
                               rewrite length_lg.
                               omega.
                        replace  (firstn (proj1_sig p + (proj1_sig p' - proj1_sig p))
                          (map (fun x : processor => get plst' x) lg ++
                            map (fun x : processor => get plst' x) (p :: ld)))
                        with  
                          (map (fun x : processor => get plst' x) lg ++ firstn  (proj1_sig p' - proj1_sig p) (
                            map (fun x : processor => get plst' x) (p :: ld))).
                       Focus 2.
                               rewrite firstn_app2. 
                               autorewrite with length.
                               rewrite length_lg.
                               replace ( proj1_sig p + (proj1_sig p' - proj1_sig p) - proj1_sig p) with  (proj1_sig p' - proj1_sig p) by (clear length_lg_p length_lg ;omega).
                               reflexivity.
                               autorewrite with length.
                               rewrite length_lg.
                               omega.
                      repeat rewrite <-  map_app.
                      replace (lg ++ p :: ld) with ((lg ++ [p]) ++ ld) by (rewrite <- app_assoc; reflexivity).
                      repeat rewrite (map_app _ (lg ++ [p])).
                      assert ( H1:((proj1_sig p + (proj1_sig p' - proj1_sig p) + 1 -(length lg + length [p]))) =  (proj1_sig p' - proj1_sig p)) by (rewrite length_lg; simpl; clear length_lg_p length_lg ;omega).                 
                      clear length_lg_p length_lg. (* Speed up omega tactic call *)
                      replace (
                                 skipn (proj1_sig p + (proj1_sig p' - proj1_sig p) + 1)
                                 (map (fun x : processor => get plst x) (lg ++ [p]) ++
                                   map (fun x : processor => get plst x) ld)) 
                      with
                                 (skipn (proj1_sig p' - proj1_sig p)
                                     (map (fun x : processor => get plst x)  ld)) 
                      by (
                                       rewrite (skipn_app2);[
                                         autorewrite with length;
                                           rewrite H1; reflexivity|
                                             autorewrite with length;
                                             omega]).
                       replace (
                         skipn (proj1_sig p + (proj1_sig p' - proj1_sig p) + 1)
                         (map (fun x : processor => get plst' x) (lg ++ [p]) ++
                           map (fun x : processor => get plst' x) ld)) 
                       with
                         (skipn (proj1_sig p' - proj1_sig p)
                           (map (fun x : processor => get plst' x)  ld)) 
                       by(
                         rewrite (skipn_app2);[
                           autorewrite with length;
                             rewrite H1; reflexivity|
                               autorewrite with length;  omega]
                       ).
                       repeat rewrite fold_left_app.
                       simpl_decomposed_lists.
                       f_equal.
                       change (p :: ld) with([p]++ld).
                       replace  (proj1_sig p' - proj1_sig p)with  (S (proj1_sig p' - proj1_sig p -1)) by (omega).
                       simpl. 
                       repeat rewrite (fold_left_app_acc _ (get _ p)).
                       rewrite get_plst_p.
                       rewrite get_plst'_p. 
                       simpl.
                       simpl_decomposed_lists.
                       repeat rewrite associative.
                       rewrite <- gl_app.             
                       reflexivity.
            
                       (* Subgoal produce by rewrite 
                          the result of bh over empty list is empty (for plst') *)
             intros l0 H0.
             rewrite in_map_iff in H0.
             destruct H0 as [p' H0]; destruct H0 as [l0_def p'def].
             rewrite bh_bsml_opt_comp_bh_aux in l0_def.
             replace (get plst' p') with (@nil A)  in l0_def by(simpl_decomposed_lists; reflexivity).
             simpl in  l0_def.
             symmetry;assumption.

             (* Subgoal produce by rewrite 
                the result of bh over empty list is empty (for plst) *)
             intros l0 H0.
             rewrite in_map_iff in H0.
             destruct H0 as [p' H0]; destruct H0 as [l0_def p'def].
             rewrite bh_bsml_opt_comp_bh_aux in l0_def.
             replace (get plst p') with (@nil A)  in l0_def by(simpl_decomposed_lists; reflexivity).
             simpl in  l0_def.
             symmetry;assumption.
   Qed.
   
   Lemma  bh_bsml_opt_comp_length :
     forall l r plst,
      length (listOfParList (bh_bsml_opt_comp l plst r))  =
      length (listOfParList ( plst)) .
    Proof.
      intros l r plst.
      rewrite bh_bsml_opt_comp_bh.
      autorewrite with length.
      reflexivity.
    Qed.

 End BSML_BH.
 
 Hint Rewrite bh_bsml_opt_comp_length : length. 

End TYPE.

Module Make(Bs : PRIMITIVES) 
  (Bsml_prop : Properties.TYPE Bs) 
  (ParSig : ParSig.TYPE Bs Bsml_prop) 
  (Bsml_base : Base.TYPE Bs Bsml_prop ParSig)  
  (Parallel : Parallelisation )
  (Partition : Lists.TYPE Bs Bsml_prop ParSig Bsml_base Parallel).
  Include TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition.
End Make.
