Require Import Skel.BH.bh_bsml.
Require Import Datatypes.Lists.ListsDistribution.
Require Import List.

Module Type BH_Cost  (Bs : PRIMITIVES) 
                                            (Bsml_prop : Properties.TYPE Bs) 
                                            (ParSig : ParSig.TYPE Bs Bsml_prop) 
                                            (Bsml_base : Base.TYPE Bs Bsml_prop ParSig)  
                                            (Parallel : Parallelisation )
                                            (Partition : Lists.TYPE Bs Bsml_prop ParSig Bsml_base Parallel)
                                            (Bsml_bh : bh_bsml.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition) 
                                            (distrib : ListDistributionByList Bs Bsml_prop ParSig Bsml_base Parallel Partition).

  Import Bsml_bh. 
  Import distrib.  

  Lemma BH_PreserveDistribution :
    forall (L A R B : Type) (k : L -> A -> R -> B) (gl : list A -> L)
    (opl : L -> L -> L) (gr : list A -> R) (opr : R -> R -> R),
    bh.is_homomorphism A L gl opl ->
    bh.is_homomorphism A R gr opr ->
    forall (l : L) (r : R) (plst : Bs.par (list A)),
      proj1_sig (compute_distribution (bh_bsml_opt_comp L A R B k gl opl gr opr l plst r) )
      =
      proj1_sig (compute_distribution  plst)
      .
Proof.
  intros L A R B k gl opl gr opr H H0 l r plst.
  unfold compute_distribution.
  apply distribution_eq_length_irrel.
  simpl in *.
  apply map_ext.
  intros a.
  autorewrite with bsml.
  rewrite bh_bsml_opt_comp_bh_aux; try assumption .
  autorewrite with length.
  reflexivity.
Qed.
End BH_Cost.

Module Make  (Bs : PRIMITIVES) 
                                            (Bsml_prop : Properties.TYPE Bs) 
                                            (ParSig : ParSig.TYPE Bs Bsml_prop) 
                                            (Bsml_base : Base.TYPE Bs Bsml_prop ParSig)  
                                            (Parallel : Parallelisation )
                                            (Partition : Lists.TYPE Bs Bsml_prop ParSig Bsml_base Parallel)
                                            (Bsml_bh : bh_bsml.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition) 
                                            (distrib : ListDistributionByList Bs Bsml_prop ParSig Bsml_base Parallel Partition).
Include BH_Cost Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh distrib.
End Make.
(*  *)