Require Import List.
Require Import CertifiedBSML.QuickStart.
Require Import SDPP.CorrectParallelisation.Structure.
Require Import SDPP.Datatypes.Lists.Lists.

Require Import LIFO.List.Filter.

Module Type TYPE
  (BsmlPrimitives: Primitives.Specification.PRIMITIVES)
  (BsmlPropertiesM : Primitives.Properties.TYPE(BsmlPrimitives))
  (ParSigM : Support.ParSig.TYPE(BsmlPrimitives)(BsmlPropertiesM))
  (BsmlBaseM : StandardLibrary.Base.TYPE(BsmlPrimitives)(BsmlPropertiesM)(ParSigM))
  (Parallel : Parallelisation)
  (ListPart : Lists.TYPE   BsmlPrimitives BsmlPropertiesM ParSigM BsmlBaseM   Parallel).
  Import BsmlPrimitives BsmlPropertiesM ParSigM BsmlBaseM Parallel.
  Existing Instance RelationClasses.eq_equivalence.

(**  **map  *)
Program Definition mapPar (E A : Type) (f : E -> A) plst : par (list A) := 
  parfun (map f) plst.
Implicit Arguments mapPar [E A].

Global Instance  mapParCorrect (E A : Type) (f : E -> A) : Parallel (map f) (mapPar f).
constructor.
intros par_a. simpl.
unfold mapPar.
listOfParListInduction.
autorewrite with bsmlbase.
reflexivity.
repeat rewrite map_app.
rewrite IHp'.
autorewrite with bsmlbase.
reflexivity.
Qed.              

(**  **Filter  *)

Program Definition filterPar  (E:Type)(f:E -> bool)(pl:par (list E)) : par (list E):=
  parfun (filter f) pl.
Implicit Arguments filterPar  [E].

Global Instance  FilterParCorrect (E : Type) (select : E -> bool) : Parallel (filter select) (filterPar select).
(* begin hide *) 
Proof.
  constructor;intros par_a; simpl.  
  listOfParListInduction.
  unfold filterPar.
  autorewrite with  bsmlbase.
  reflexivity.
  rewrite filter_app.
  rewrite IHp'.
  f_equal.
  unfold filterPar;
    autorewrite with bsmlbase.
    reflexivity.
Qed.

End TYPE.

Module Make
  (BsmlPrimitives: Primitives.Specification.PRIMITIVES)
  (BsmlPropertiesM : Primitives.Properties.TYPE(BsmlPrimitives))
  (ParSigM : Support.ParSig.TYPE(BsmlPrimitives)(BsmlPropertiesM))
  (BsmlBaseM : StandardLibrary.Base.TYPE(BsmlPrimitives)(BsmlPropertiesM)(ParSigM))
  (Parallel : Parallelisation)
  (ListPart  : Lists.TYPE   BsmlPrimitives BsmlPropertiesM ParSigM BsmlBaseM   Parallel).
Include TYPE BsmlPrimitives BsmlPropertiesM ParSigM BsmlBaseM Parallel ListPart.
End Make.
(*  *)