
Require Export CertifiedBSML.QuickStart.
Require Export CorrectParallelisation.Structure.
Require Export Datatypes.Lists.Lists.

Require Import List.
Require Import LIFO.List.Notations.
Require Import LIFO.List.Last.
Require Import LIFO.List.Fold.
Require Import LIFO.Algebra.AlgebraClasses.
Require Import LIFO.Option.Option.

Module Type TYPE(Bs : PRIMITIVES) 
  (Bsml_prop : Properties.TYPE Bs) 
  (ParSig : ParSig.TYPE Bs Bsml_prop) 
  (Bsml_base : Base.TYPE Bs Bsml_prop ParSig)  
  (Parallel : Parallelisation )
  (Partition : Lists.TYPE Bs Bsml_prop ParSig Bsml_base Parallel).
  
  Export Bs.  
  Export Bsml_base.
  Export Bsml_prop.

  Export Parallel.
  Export Partition.

Section Last_Parametrized_by_A.



Variable A:Type.

Program Definition option_parallel_last lv : par (option A)    :=  
  let local_last := parfun  (@last_option A ) lv in 
    let mesg:= put (parfun (fun  a _ => a) local_last ) in 
      parfun (fun f => fold_right (fun p acc => right_option (f p) acc)  None (proj1_sig processors)) mesg
.

Program Definition parallel_last lv d : par A :=  
  parfun (fun opt => match opt with None  => d | Some a => a end ) (option_parallel_last lv).


End Last_Parametrized_by_A.


Global Instance partition_A_default A (default : A) : Partitionnable A | 1000  :=
{
  join :=fun v => get v firstProcessor;
    partition := fun a => proj1_sig (mkpar (fun p =>if processorEqDec p firstProcessor then  a else default))
}.
Proof.
  (* exists A; reflexivity. *)
  intros a.
  autorewrite with bsml.
  destruct (processorEqDec firstProcessor firstProcessor)as [H|H] ; [ reflexivity |  contradict H; reflexivity]. 
Defined.

(* Instance partition_A A : Partitionnable A | 1000  := *)
(* { *)
(*   join :=fun v => proj v firstProcessor; *)
(*     partition := fun a =>replicate a *)
(* }. *)
(* Proof. *)
(*   (* exists A; reflexivity. *) *)
(*   intros a. *)
(*   unfold replicate.  *)
(*   autorewrite with bsml. *)
(*   reflexivity. *)
(* Defined. *)

(* Global Instance partition_option_default A : Partitionnable (option A) | 950.  *)
(* apply (@partition_A_default (option A) None). *)
(* Defined. *)

Global Instance partition_option_default A : Partitionnable (option A) | 950:=
{
  join :=fun v => get v firstProcessor;
    partition := fun a => proj1_sig (mkpar (fun p =>if processorEqDec p firstProcessor then  a else None ))
}.
Proof.
  (* exists A; reflexivity. *)
  intros a.
  autorewrite with bsml.
  destruct (processorEqDec firstProcessor firstProcessor)as [H|H] ; [ reflexivity |  contradict H; reflexivity]. 
Defined.

Global Instance parallel_last_option_ok A : Parallel (@last_option A)  (option_parallel_last A).
Proof.
  Opaque processors.
  constructor.
  intros par_a.
  simpl.
  unfold option_parallel_last.
  simpl.

  unfold parfun.
  autorewrite with bsml.
  rewrite   Map.fold_right_f_inv.


  set (l:=listOfParList par_a).
  assert (lopl: l =listOfParList par_a) by reflexivity.
  induction  (l) using rev_ind. 
  
  apply fold_right_const.
  intros a H.
  simpl.
  rewrite in_map_iff in H.
  destruct H as [p' H].
  destruct H as [a_value in_p'_all_proc].
  rewrite <- a_value.
  autorewrite with bsml.
  rewrite parallel_list_nil.
  reflexivity.
  auto.
  
  generate_sub_par_list_rev par_a.
  
decomp_one_lopl p par_a.
  change (p :: ld) with ([p]++ld).
  rewrite map_app.
  repeat rewrite fold_right_app.   
  rewrite (fold_right_const _ _ _ None _  ).
  Focus 2.
      intros a H.
      rewrite in_map_iff in H.
      destruct H as [p' H].
      destruct H as [a_value in_p'_ld].
      rewrite <- a_value.
      autorewrite with bsml.
      simpl_decomposed_lists_rev.
      reflexivity.               

      simpl.  
      replace     (get
        (proj1_sig
           (put
              (proj1_sig
                 (apply
                    (proj1_sig
                       (replicate (fun (a : option A) (_ : processor) => a)))
                    (proj1_sig
                       (apply (proj1_sig (replicate (@last_option _))) par_a))))))
        firstProcessor p)
        with ( Some x) by  (
          autorewrite with bsml;
            simpl_decomposed_lists_rev;
            simpl;symmetry;
              apply last_option_app_r).

      rewrite  fold_right_const.
      autorewrite with bsml.
      simpl_decomposed_lists_rev.
      autorewrite with last.
      reflexivity.

      intros a H.
      rewrite in_map_iff in H.
      destruct H as [p' H].
      destruct H as [a_value in_p'_lg].
      rewrite <- a_value.
      autorewrite with bsml.
      unfold right_option.
      simpl_decomposed_lists_rev.
      autorewrite with last.
      reflexivity.               
Qed.

Global Instance parallel_last_ok A neutral : 
  Parallel 
  (BPart:=partition_A_default A neutral)
  (fun lst => List.last lst neutral)  
  (fun plst => parallel_last A plst neutral).
Proof.
  constructor.
  unfold parallel_last.
  intros par_a. simpl.
  replace  ( get
     (proj1_sig
        (apply
           (proj1_sig
              (replicate
                 (fun opt : option A =>
                  match opt as opt0 return (opt0 = opt -> A) with
                  | Some a => fun _ : Some a = opt => a
                  | None => fun _ : None = opt => neutral
                  end eq_refl))) (option_parallel_last A par_a)))
     firstProcessor)
    with 
    (
      (fun opt : option A => match opt with
                               | Some a => a
                               | None => neutral
                             end) 
      (get
        (option_parallel_last A  par_a)
        firstProcessor)
     ).
destruct ( parallel_last_option_ok A ).
simpl in parallel_spec_match0.
rewrite parallel_spec_match0.
induction (listOfParList par_a) using rev_ind.
reflexivity.  
      autorewrite with last.
reflexivity.
      autorewrite with bsml last.
      destruct (get (option_parallel_last A par_a) firstProcessor);reflexivity.
Qed.
End TYPE.

Module  Make(Bs : PRIMITIVES) 
  (Bsml_prop : Properties.TYPE Bs) 
  (ParSig : ParSig.TYPE Bs Bsml_prop) 
  (Bsml_base : Base.TYPE Bs Bsml_prop ParSig)  
  (Parallel : Parallelisation )
  (Partition : Lists.TYPE Bs Bsml_prop ParSig Bsml_base Parallel).
  Include TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition.
End Make.

