Require Import List.
Require Import Arith.
Require Import Omega.

Require Export SDPP.Skel.map_around. 
Require Export SDPP.Skel.scan. 
Require Export SDPP.Skel.LastPar. 

Require Import LIFO.List.Inits_tails.
Require Export LIFO.List.Homomorphism.
Require Export LIFO.List.Length.
Require Export LIFO.List.Map.
Require Export LIFO.List.Misc.
Require Export LIFO.List.Notations.
Require Export LIFO.List.Prefix_Suffix.
Require Import LIFO.List.Triple_lists.

Require Export LIFO.Algebra.AlgebraClasses.

Module Type TYPE(Bs : PRIMITIVES) 
  (Bsml_prop : Properties.TYPE Bs) 
  (ParSig : ParSig.TYPE Bs Bsml_prop) 
  (Bsml_base : Base.TYPE Bs Bsml_prop ParSig)  
  (Parallel : Parallelisation )
  (Partition : Lists.TYPE Bs Bsml_prop ParSig Bsml_base Parallel)
  (Bsml_bh : bh_bsml.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition) 
  (Bh_Classes : bh_classes.TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh)
  (Ma : map_around.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh Bh_Classes)
  (LastParallel : LastPar.TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition).

  Import Ma LastParallel Parallel.

Definition homomorphism_prefixes {L A: Type}(op : L->L->L) (h : list A -> L) l :=
    mapAround (fun x => op (h (fst3 x)) (h [(id (snd3 x))])) l.

  Lemma last_homomorphism_prefixes_correct: 
    forall (A B :Type) (op:B->B->B) h, Homomorphism' op h ->
      forall l, h l = last (homomorphism_prefixes op h  l) (h (@nil A)).
  Proof.
    unfold homomorphism_prefixes.
    intros A B op h H  l.
    set (f1:=h) ; set (f2:=@id A); set (f3:=fun l:list A => @nil A);set( combine := fun (v: B *A *list A) => match v with  |((x,y),z) =>op x (h [y])end).
    destruct H as [H1 H2 H3].
    destruct H2 as [UnitL UnitR].
    unfold mapAround.
    rewrite (map_zip3  _ _ _ _ f1 f2 f3 combine).
    Focus 2. reflexivity.
    subst f1.
    induction l.
    auto.

    assert(forall (A:Type) (a:A) (l:list A), a::l = (a::nil)++l) as Happ.
      intros A0 a0 l0.
      rewrite <- (app_nil_l l0).
      apply app_comm_cons.
    rewrite Happ.
    rewrite H3.
    rewrite  IHl.
    
    rewrite (last_map _ _ (h nil,a,nil) (h nil) _ combine).
    rewrite (last_map _ _ (h nil,a,nil) (h nil) _ combine).
    repeat(rewrite last_zip3).
    repeat(rewrite (last_map _ _ (@nil A))).
    repeat(rewrite (last_map _ _ a)).

    Focus 2.
      rewrite map_length. rewrite map_length. rewrite inits_length. reflexivity.
    Focus 2.
      rewrite map_length. rewrite map_length. rewrite tails_length. reflexivity.
    Focus 2.
      rewrite map_length. rewrite map_length. rewrite inits_length. reflexivity.
    Focus 2.
      rewrite map_length. rewrite map_length. rewrite tails_length. reflexivity.

    simpl.

    rewrite (Happ A a l).
    rewrite decompInits.
    (* unfold f1. *)
    repeat(rewrite map_app). 

    case_eq(is_nil l); intro Hl.

    apply is_nil_true in Hl.
    rewrite Hl.
    simpl. rewrite left_neutral. rewrite right_neutral. reflexivity.

    apply is_nil_false in Hl.
    destruct Hl as [a' Hl].
    destruct Hl as [la' Hl].
    rewrite Hl.
    assert(forall (A:Type) (a:A) (l:list A),
      S(length l) = length(inits(a::l))) as lengthInits.
      intros. rewrite inits_length. reflexivity.
    assert(forall (A:Type)  (a:A) (l:list A),
      S(length l) = length(tails(a::l))) as lengthTails.
      intros. rewrite tails_length. reflexivity. 
    set(Hl1 := lengthInits A a nil).

       apply lengthS in Hl1. destruct Hl1 as [a0 Hl1]. destruct Hl1 as [la0 Hl1].
    set(Hl2 := lengthTails A a nil).
    apply lengthS in Hl2. destruct Hl2 as [a1 Hl2]. destruct Hl2 as [la1 Hl2].
    set(Hl3 := lengthInits A a' la').
    apply lengthS in Hl3. destruct Hl3 as [a2 Hl3]. destruct Hl3 as [la2 Hl3].
    set(Hl4 := lengthTails A a' la').
    apply lengthS in Hl4. destruct Hl4 as [a3 Hl4]. destruct Hl4 as [la3 Hl4].
    rewrite Hl3.
    assert(length (a2::la2)=length(a'::la')).
    rewrite <-Hl3.
    apply inits_length.
    inversion H.
    symmetry in H2. simpl. 
    repeat(rewrite <-H3).
    rewrite app_assoc.
    repeat(rewrite H3).
    rewrite <-H3.
    f_equal.
    f_equal.
    induction la2.
    auto.
    rewrite (last_map _ _ (@nil A)).
    simpl. 
    reflexivity.
  Qed.

  Definition k (A B C : Type) (op : B -> B -> B) (h: list A -> B) x y (z:C) := 
    op x (h (y::nil)).

  Definition f'(A B : Type) (op : B -> B -> B) (h: list A -> B) 
    (hom : Homomorphism' op h)  := fun t => 
    let f1 := h in
    let f2 := id in 
    let f3 := fun l => @nil A in 
    let combine v := match v with 
                     |((x,y),z) =>op x (h (y::nil)) 
                     end in
    let f := fun x => 
      combine ((f1(fst3(C:=list A) x)),(f2(snd3 x)),(f3(trd3 x))) in
      f t.



  Program Instance homomorphism_prefixes_BHable (A B : Type) 
  (op : B -> B -> B) (h: list A -> B) 
   {hom : Homomorphism' op h} : MapAround_BHable 
  (B:=B) (L:=B) (A:=A) (R:=list A) 
  (k := fun x y z =>op x (h (y::nil)))
  (gl := h) (gr := fun l => @nil A) (opl := op) (idl := h nil)
  (opr := @app A) (idr := nil)
  (homomorphism_prefixes op h)
 :=
  {
    MA_BH_f' := (@f' A B op h hom)
  }.
Next Obligation.
firstorder.
Qed.
Next Obligation.
firstorder.
Qed.
Next Obligation.
firstorder.
Qed.
 
(* *******************************  *)
(* *******************************  *)
(*     Homomorphism''                            *)
(* *******************************  *)
(* *******************************  *)
  Program Instance homomorphism''_prefixes_BHable (A B : Type)
    (e:B) (op : B -> B -> B) (h: list A -> B)
    `{hom : Homomorphism'' _ _ eq e op h} : MapAround_BHable
  (B:=B) (L:=B) (A:=A) (R:=list A)
  (k := fun x y z =>op x (h (y::nil)))
  (gl := h) (gr := fun l => @nil A) (opl := op) (idl := h nil)
  (opr := @app A) (idr := nil)
  (homomorphism_prefixes op h)
 :=
  {
    MA_BH_f' := (@f' A B op h _)
  }.
Next Obligation.
firstorder.
rewrite AlgebraClasses.associative.
reflexivity.
rewrite  Homomorphism''_h_nil; rewrite left_neutral; reflexivity.
rewrite  Homomorphism''_h_nil; rewrite right_neutral; reflexivity.
Defined.
Next Obligation.
firstorder. 
Defined.
Next Obligation.
  intros A B e op h eqb opass opneutral hom ls x rs.
  firstorder.
 (*  rewrite  Homomorphism''_h_nil. firstorder. *)
 (* rewrite left_neutral; reflexivity. *)
 (* rewrite right_neutral; reflexivity. *)
Defined.
Next Obligation.
  intros A B e op h eqb opass opneutral hom.
  typeclasses eauto. 
Defined.
Next Obligation.
  intros A B e op h eqb opass opneutral hom.
  firstorder;  
  rewrite Homomorphism''_h_nil;
    firstorder.
Defined.

Next Obligation.
  firstorder.  
rewrite AlgebraClasses.associative.
reflexivity.
rewrite  Homomorphism''_h_nil; rewrite left_neutral; reflexivity.
rewrite  Homomorphism''_h_nil; rewrite right_neutral; reflexivity.
Defined.

Next Obligation.
  firstorder.  
Defined.

(* *******************************  *)
(* *******************************  *)
(*    fin  Homomorphism''                       *)
(* *******************************  *)
(* *******************************  *)

Import Partition.
Section parametrized.
  Variable A B : Type.
  Variable op : B -> B -> B.
  Variable  h: list A -> B.

 


  Definition homomorphismParallelKernel   {homo_h :Homomorphism' op h } :=
    parallel   (f:= (homomorphism_prefixes op h)). 
  
  Instance homoKernel   {homo_h :Homomorphism' op h }  : Parallel (R:=eq) (homomorphism_prefixes op h) homomorphismParallelKernel.
  constructor.
  intros par_a.
  unfold homomorphismParallelKernel.
  rewrite  spec_match.
  reflexivity.
  Qed.

  (* Global Instance homo_result_part   {e:B} {neutral op :Neutral op e } : Partitionnable B:=partition_A_default B e.  *)
  Global Instance homo_result_part   {homo_h :Homomorphism' op h } : Partitionnable B:=partition_A_default B (h []).

  Definition homoPar  {homo_h :Homomorphism' op h } plst := let neutral  := (h [])  in  parallel_last _ (homomorphismParallelKernel plst) neutral .
  

  
  Global Instance homomorphism_Parallel  {homo_h :Homomorphism' op h } : Parallel h (homoPar).
  constructor.
  intros par_a.
   rewrite  (last_homomorphism_prefixes_correct _  _ op h).
  unfold homoPar.
  rewrite <-  spec_match.
  match goal with [|- context [last ?e ?n]] => change (last e n) with ((fun l  => last l n) e) end.
  rewrite <- spec_match.
  reflexivity.
  apply homo_h.
  Defined.

  (* necessite passage de Homomorphism' à  Homomorphism'' (si instance passerelle -> boucle sur neutral) *)

 


  Definition homomorphism''ParallelKernel   `{homo_h :Homomorphism''  _ _ _ e op h } :=
    parallel   (f:= (homomorphism_prefixes op h)). 
  
  Instance homo''Kernel    `{homo_h :Homomorphism''  _ _ _ e op h }  : Parallel (R:=eq) (homomorphism_prefixes op h) homomorphism''ParallelKernel.
  constructor.
  intros par_a.
  unfold homomorphism''ParallelKernel.
  rewrite  spec_match.
  reflexivity.
  Qed.

  (* Global Instance homo_result_part   {e:B} {neutral op :Neutral op e } : Partitionnable B:=partition_A_default B e.  *)
   Instance homo''_result_part   `{Neutral _  eq (equ:=_) op e}  : Partitionnable B:=partition_A_default B e.

  Definition homo''Par  `{homo_h :Homomorphism''  _ _  _ (eqb:= _) e op h } plst  :=  parallel_last _ (homomorphism''ParallelKernel plst) e .
  Program Definition homo''Par'    (e : B) (opass : Associative op) (opneutral : Neutral op e)
    {homo_h :Homomorphism''  e op h }  : (parallel_type (Partitionnable:=(list_partitionnable (A:=A))))-> (parallel_type (Partitionnable:=homo''_result_part (e:=e))).  
  simpl. 
  intros e opass opneutral homo_h X.
  exact (homo''Par X). 
Defined.

Global Instance homomorphism''_Parallel  `{homo_h :Homomorphism'' _ _ _  (eqb:= _)  e op h } : Parallel (BPart:=(homo''_result_part (e:=e)))    h (@homo''Par' e  _ _ homo_h ).
  constructor.
  intros par_a.
   rewrite  (last_homomorphism_prefixes_correct _  _ op h).
  unfold homoPar.
  rewrite <-  spec_match.
  match goal with [|- context [last ?e ?n]] => change (last e n) with ((fun l  => last l n) e) end.
  rewrite <- spec_match.
  simpl.
  rewrite   Homomorphism''_h_nil.  reflexivity.
  
  firstorder.
  rewrite AlgebraClasses.associative.
  reflexivity.
  rewrite  Homomorphism''_h_nil; rewrite left_neutral; reflexivity.
  rewrite  Homomorphism''_h_nil; rewrite right_neutral; reflexivity.
Defined.

End parametrized.

End TYPE.


Module Make(Bs : PRIMITIVES) 
  (Bsml_prop : Properties.TYPE Bs) 
  (ParSig : ParSig.TYPE Bs Bsml_prop) 
  (Bsml_base : Base.TYPE Bs Bsml_prop ParSig)  
  (Parallel : Parallelisation )
  (Partition : Lists.TYPE Bs Bsml_prop ParSig Bsml_base Parallel)
  (Bsml_bh : bh_bsml.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition) 
  (Bh_Classes : bh_classes.TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh)
  (Ma : map_around.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh Bh_Classes)
  (LastParallel : LastPar.TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition).

Include TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh Bh_Classes Ma LastParallel.

End Make.