Require Import List.
Require Import Arith.
Require Import Omega.

Require Export Skel.BH.bh.
Require Export Skel.BH.bh_bsml.
Require Export Skel.BH.bh_classes.

Require Export CertifiedBSML.QuickStart.

Require Import LIFO.List.Triple_lists.
Require Import LIFO.List.Map. 
Require Import LIFO.List.Inits_tails. 
Require Import LIFO.List.Nth. 
Require Import LIFO.List.Homomorphism.

Require Import LIFO.Algebra.AlgebraClasses.

Module Type TYPE (Bs : PRIMITIVES) 
  (Bsml_prop : Properties.TYPE Bs) 
  (ParSig : ParSig.TYPE Bs Bsml_prop) 
  (Bsml_base : Base.TYPE Bs Bsml_prop ParSig)  
  (Parallel : Parallelisation )
  (Partition : Lists.TYPE Bs Bsml_prop ParSig Bsml_base Parallel)
  (Bsml_bh : bh_bsml.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition) 
  (Bh_Classes : bh_classes.TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh).


(** ** Definitions *)
Section SPEC.

  Variables (A B: Type).
(** mapAround function *)
Definition mapAround  (f : list A * A  * list A-> B) (lst: list A) : list B :=
  map f (zip3 (inits lst) lst (tails lst)).

  (**  BH variables *)
  Variable (L R : Type).
(*   Variable (k: . *)
(*   Variable (gl: list A -> L). *)
  Variable (opl: L -> L -> L).
(*   Variable. *)
  Variable (opr: R -> R -> R).
  Variable idl : L.
  Variable idr : R.
  Variable identl_idl : forall l:L, opl l idl=l.
  Variable identr_idl : forall l:L, opl idl l=l.
  Variable identl_idr : forall r:R, opr r idr=r.
  Variable identr_idr : forall r:R, opr idr r=r.
(*   Variable gl_nil_idl : gl nil = idl. *)
(*   Variable gr_nil_idr : gr nil = idr. *)
  
  Set Implicit Arguments. 


(** Theorem 3.1 derivation of mapAround into bh *)
Theorem mapAround_to_bh :
  forall (f :((list A)*A*(list A)) -> B)  (k :L -> A -> R -> B) (gl: list A -> L)  (gr: list A -> R)  xs ,
    (gl nil = idl) ->
    (gr nil = idr) ->
    (forall ls x rs,
      f (ls,x,rs) = k (gl ls) x (gr rs))
    -> is_homomorphism A L gl opl
    -> is_homomorphism A R gr opr
    -> mapAround f xs = bh_comp  k gl opl gr opr idl xs idr. 

Proof.
  intros f k gl gr xs gl_nil_idl gr_nil_idr decompf homogl homogr .
    unfold is_homomorphism in *.
    destructList xs as xsCase. 
(** destruct the list in three cases :  empty list, singleton or (non-empty) list concatenation*)
      (** case lst = nil*)          
      unfold mapAround ;subst;simpl;reflexivity.
      (** case lst = [a]*) 
      revert a xsCase; intros x xsSingleton; generalize idl idr gl_nil_idl gr_nil_idr ;subst; intros idl idr gl_nil_idl gr_nil_idr.
      unfold mapAround, bh_comp ; simpl; unfold bh_nth;simpl.
      decompose [and] homogl ;decompose [and] homogr.
      rewrite decompf. subst.
      rewrite  (H1 (gl nil)).
      rewrite  (H4 (gr nil)).
      reflexivity.
      (** case lst = a::la ++ b::lb]*)          
       generalize idl idr gl_nil_idl gr_nil_idr ;subst; intros idl idr gl_nil_idl gr_nil_idr.
      unfold mapAround.
      (*rewrite homomorphisms decompositions f ++*) 
      apply (nthSameLengthEqual)with (d:=f(la,a,la)).
      rewrite decompTails. rewrite decompInits. zip3Rewrite. 
        rewrite map_app. (* apply (nthSameLengthEqual)with (d:=f(la,a,la)).           *)
          autorewrite with length; repeat rewrite Min.min_idempotent.
          reflexivity.
          autorewrite with length; repeat rewrite Min.min_idempotent.
          reflexivity.
          rewrite inits_length;reflexivity.
          intros n len. generalize len.
          rewrite map_length in len; rewrite zip3_length in len;rewrite inits_length in len;rewrite tails_length in len;repeat rewrite Min.min_idempotent in len.
          intro len2.
          rewrite (nth_bh k gl opl gr opr a (f (la, a, la)) idl idr ((a :: la) ++ b :: lb) n len ). 
          rewrite map_length in len2.  
          rewrite (map_nth' f _ (f(la,a,la)) (((a :: la) ++ b :: lb),a,nil) n len2).
          subst.
          rewrite identr_idl.
          rewrite identl_idr.

          zip3_nth_rewrite.
          rewrite decompf.
          rewrite inits_nth.
          rewrite tails_nth.
          reflexivity.
        Qed.

End SPEC.

Implicit Arguments mapAround [ [A] [B] ].

Export Bh_Classes.  

  (** MapAround_BHable Class  
   this Type Classes defined the sufficient condition to express a specification written using mapAround as a BH.
   It includes :
    - properties over operators
    - a proof that the specification is a mapAround
    *)
  Class MapAround_BHable {L A R B : Type} {k: L -> A -> R -> B} {gl: list A -> L} {opl: L -> L -> L} {idl :L}
  {gr: list A -> R} {opr: R -> R -> R} {idr : R} (f : list A -> list B) :Prop :=
  { MA_BH_f' : (list A)*A*(list A)  ->  B;
    f_is_mapAround : forall l, f l = mapAround MA_BH_f' l;
      MA_BH_f'_to_k :  (forall ls x rs, MA_BH_f' (ls,x,rs) = k (gl ls) x (gr rs));
      MA_BH_right_neutral :> Neutral opr idr;
      MA_BH_left_neutral :> Neutral opl idl;
      MA_BH_left_homomorphism : Homomorphism' opl gl ;
      MA_BH_right_homomorphism : Homomorphism' opr gr 
  }.
Hint Constructors MapAround_BHable : mapAround.
Hint Resolve @MA_BH_right_neutral @MA_BH_left_neutral  : mapAround.
Hint Resolve @MA_BH_left_homomorphism @MA_BH_right_homomorphism  : mapAround.

  (** With any instance of MapAround_BHable we can build an instance of BH.*)
  Global Program Instance mapAround_to_BH `(MapAround_BHable) : 
   BH_Special
  f idl idr:=
 {  BHS_k:=k;
  BHS_gl:=gl;
  BHS_opl:=opl;
  (* BHS_idl:=idl; *)
  BHS_gr:=gr; 
  BHS_opr:=opr(* ; *)
  (* BHS_idr:=idr *)
 }.

  Solve Obligations using (eauto with mapAround). 
  Next Obligation. (* f l = bh_comp k gl opl gr opr idl l idr *)
(*   intros L A R B k gl opl idl gr opr idr f H l. *)
    intros L A R B k gl opl idl gr opr idr f H lst.
    elim H. 
    intros f'0 f_is_mapAround0 MapAround_BH_f'_to_k0 MapAround_to_BH_right_neutral0;
    intros MapAround_to_BH_left_neutral0 MapAround_to_BH_left_homomorphism0;
    intros MapAround_to_BH_right_homomorphism0.
    rewrite f_is_mapAround0. 
    destructList lst as lCase. 
(* destruct the list in three cases :  empty list, singleton or (non-empty) list concatenation*)
      (* case lst = nil*)          
    unfold mapAround ;subst;simpl;reflexivity.
      (* case lst = [a]*)          
    revert a lCase; intros x lSingleton; subst. 
    unfold mapAround, bh_comp ; simpl; unfold bh_nth;simpl.
    rewrite MapAround_BH_f'_to_k0.
    rewrite (right_neutral (e:=idr)).
    rewrite (left_neutral (e:=idl)).
    reflexivity.
      (* case lst = a::la ++ b::lb]*)          
    (* elim lCase;clear lCase; intros b lCase;  *)
    (*   elim lCase;clear lCase;intros a lCase;  *)
    (*     elim lCase;clear lCase;intros la lCase;  *)
    (*       elim lCase;clear lCase;intros lb lCase; *)
            subst. 
    unfold mapAround.
      (*rewrite homomorphisms decompositions f ++*)
    apply (nthSameLengthEqual)with (d:=f'0(la,a,la)). 
    rewrite decompTails. rewrite decompInits. zip3Rewrite. 
    rewrite map_app. (* apply (nth_equal)with (d:=f(la,a,la)).           *)
    rewrite bh_length; repeat rewrite app_length; repeat rewrite map_length; repeat rewrite zip3_length; 
      repeat rewrite map_length; repeat rewrite inits_length; repeat rewrite tails_length; repeat rewrite Min.min_idempotent.
    reflexivity.
    repeat rewrite app_length;repeat rewrite map_length; repeat rewrite zip3_length ; 
      repeat rewrite map_length; repeat rewrite inits_length; repeat rewrite tails_length; repeat rewrite min_eq.
    reflexivity.
    rewrite inits_length;reflexivity.
    intros n len. generalize len.
    rewrite map_length in len; rewrite zip3_length in len;rewrite inits_length in len;rewrite tails_length in len;repeat rewrite Min.min_idempotent in len.
    intro len2.
    rewrite (nth_bh k gl opl gr opr a (f'0 (la, a, la)) idl idr ((a :: la) ++ b :: lb) n len ). 
    rewrite map_length in len2.  
    rewrite (map_nth' f'0 _ (f'0(la,a,la)) (((a :: la) ++ b :: lb),a,nil) n len2).
    subst.
    rewrite (right_neutral (e:=idr)).
    rewrite (left_neutral (e:=idl)).

    zip3_nth_rewrite.
    rewrite MapAround_BH_f'_to_k0.
    rewrite inits_nth.
    rewrite tails_nth.
    reflexivity.
  Qed.
  (** With any instance of MapAround_BHable we can build an instance of BH.*)

  Import Parallel.
  Instance mapAround_Parallelization `(MapAround_BHable) :   Parallel.Parallelizable (R:=eq) f.
  Proof.
    typeclasses eauto.
  Defined.

  (*  TODO: de manière général, gl et gr sont des quasi-homomorphisme, 
     je ne sais pas trop comment le faire apparaitre pour que la recherche d'instance unifie bien ...*)

End TYPE.

Module MAKE (Bs : PRIMITIVES) 
  (Bsml_prop : Properties.TYPE Bs) 
  (ParSig : ParSig.TYPE Bs Bsml_prop) 
  (Bsml_base : Base.TYPE Bs Bsml_prop ParSig)  
  (Parallel : Parallelisation )
  (Partition : Lists.TYPE Bs Bsml_prop ParSig Bsml_base Parallel)
  (Bsml_bh : bh_bsml.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition) 
  (Bh_Classes : bh_classes.TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh).

  Include TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh Bh_Classes.
End MAKE.