(** Corollary 3.2 derivation of scan into bh *)
Require Import List.
Require Import Arith.
Require Import Omega.

Require Export SDPP.Skel.map_around. 


Require Import LIFO.List.Triple_lists.
Require Import LIFO.List.Inits_tails.
Require Export   LIFO.List.Prefix_Suffix.
Require Export LIFO.List.Homomorphism.
Require Export LIFO.Algebra.AlgebraClasses.

Module Type TYPE(Bs : PRIMITIVES) 
  (Bsml_prop : Properties.TYPE Bs) 
  (ParSig : ParSig.TYPE Bs Bsml_prop) 
  (Bsml_base : Base.TYPE Bs Bsml_prop ParSig)  
  (Parallel : Parallelisation )
  (Partition : Lists.TYPE Bs Bsml_prop ParSig Bsml_base Parallel)
  (Bsml_bh : bh_bsml.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition) 
  (Bh_Classes : bh_classes.TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh)
  (Ma : map_around.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh Bh_Classes).

  Export Ma Bh_Classes Parallel.

  Section SCAN. 

    Variable (A : Type).
    Variable (op : A-> A ->A).
    Variable idr : A.
   
    (** scan function *)
    (* Require Import LIFO.List.Notations.
    Eval compute in prefix nat [1;2;3].
    *)

    Fixpoint sum  ls :=
      match ls with
        nil => idr
        | a::l=> op a (sum l)
      end.
    (* Lemma sum_is_fold_left :  forall ls, sum ls =  fold_left op ls idr.  *)
    (*   intros ls. induction ls. *)
    (*   reflexivity. *)
    (*   simpl. *)
    (*   rewrite left_neutral. *)
    (*   rewrite (Fold.fold_left_assoc_with_id_prop (e:=idr)). *)
    (*   f_equal; apply IHls. *)
    (* Qed. *)

    Program Instance sum_is_an_homomorphism {ident_idr : Neutral op idr}{ associative_op : Associative op} : Homomorphism' op sum.
    Next Obligation.
      intros ident_idr associative_op x y.
      induction x.
      simpl. destruct ident_idr. rewrite (left_neutral). reflexivity.
      simpl. rewrite IHx. destruct associative_op. rewrite associative. trivial.
    Qed.

    Definition scan  (lst: list A) : list A :=
      map  sum ( prefix lst).
    
    
    
    Program Instance scan_BHable {ident_idr : Neutral op idr}{ associative_op : Associative op} : MapAround_BHable scan
    (L:=A)
    (A:=A)
    (R:=list A)
    (k:= fun x y z => op x y)
    (gl:= sum)
    (gr := fun _ => nil)
    (opl := op)
    (idl := idr)
    (opr := (app (A:=A)))
    (idr := nil) := 
    { 
      MA_BH_f' := fun t => op (sum(fst3 t)) (snd3 t)
    }.
    Next Obligation.
      unfold scan. unfold mapAround. 
      induction l using rev_ind.
      reflexivity.      
     
      rewrite prefix_app.
      rewrite map_app.
      rewrite IHl.
      rewrite decompInits.
      rewrite decompTails.
      rewrite decompZip3.
      rewrite map_app.
      f_equal.
      
      apply map_zip3_indep_3.
      reflexivity.      
      autorewrite with length. reflexivity.
      simpl.
      unfold rev'.  simpl. 
      rewrite app_nil_r. 
      rewrite Homomorphism'_list_destruct.
      simpl.
      rewrite right_neutral. 
      reflexivity.      
   autorewrite with length. reflexivity.
   autorewrite with length. reflexivity.
 Defined.      

 Next Obligation.
      constructor;
      typeclasses eauto || auto.
    Qed.


 Next Obligation.
   intros ident_idr associative_op.
      constructor; typeclasses eauto || auto.
Qed.

    Typeclasses eauto := debug.
    Definition scanPar {ident_idr : Neutral op idr}{ associative_op : Associative op} := (parallel (R:=eq) (f:=scan)(Parallelizable:= BHS_Parallelizable (mapAround_to_BH scan_BHable) )).

    Global Instance scanParallel {ident_idr : Neutral op idr}{ associative_op : Associative op} : Parallel (R:=eq) scan scanPar := BHS_Parallel (mapAround_to_BH scan_BHable) .
  
  End SCAN.

End TYPE.

Module Make(Bs : PRIMITIVES) 
  (Bsml_prop : Properties.TYPE Bs) 
  (ParSig : ParSig.TYPE Bs Bsml_prop) 
  (Bsml_base : Base.TYPE Bs Bsml_prop ParSig)  
  (Parallel : Parallelisation )
  (Partition : Lists.TYPE Bs Bsml_prop ParSig Bsml_base Parallel)
  (Bsml_bh : bh_bsml.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition) 
  (Bh_Classes : bh_classes.TYPE Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh)
  (Ma : map_around.TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh Bh_Classes).

  Include TYPE  Bs Bsml_prop ParSig Bsml_base Parallel Partition Bsml_bh Bh_Classes Ma.

End Make.