open ModuleInstances_defunct
open HeatBH

open Tools
open Nat_tools

open Bsml
open Stdlib
open Base
open Comm

open Stdlib.Base
open Bsml_nat
open Bsml_nat.Bsml_natural

module HeatEquation_instance_float = (ModuleInstances_defunct.HeatEquationBHInstanceOpt (Bsml_natural))

let heat (leftBound:float) (rightBound:float) (dt:float) (dx:float) (gamma:float) (u:(float list)par) = 
 HeatEquation_instance_float.coq_HeatEqPar' leftBound rightBound dt dx gamma u


let _ = 
  let iter = try int_of_string (Sys.argv.(1)) with _ -> 200 in
  let size = try int_of_string (Sys.argv.(2)) with _ -> 10000 in
  let leftBound = 20. 
  and rightBound = 30. 
  and gamma = 0.1
  and f x = (Random.float 60.) -. 30.
  and dx = 10. /. (float size) in
  let dt = (dx*.dx)/.(10.*.gamma) in
  let u = mkpar(fun pid -> List.map f (Tools.from_to (if pid < nat_of_int (size mod Bsml.bsp_p) then 0 else 1) (size/Bsml.bsp_p))) in
  begin
    Gc.set { (Gc.get ()) with Gc.minor_heap_size = 1024*1024*1024*100};
    let start_minor_collection_vector= << let stat =Gc.stat() in stat.Gc.minor_collections>> in
    let gc_cost= << ref 0. >> in 
    let cost = << ref 0. >> in
			  Bsml.start_timing();
			  let u' = ref u in 
			  for i=1 to iter do
			    u' := heat leftBound rightBound dt dx gamma !u';
			    Bsml.stop_timing();
			    let cost_vec=Bsml.get_cost() in
			    ignore(<< $cost$:=( $cost_vec$ +. ! ($cost$)) >>);
			    Bsml.start_timing();
			    Gc.full_major();
			    Bsml.stop_timing();
			    ignore( << $gc_cost$ := $Bsml.get_cost()$ +.  ! $gc_cost$>>);
			    Bsml.start_timing();
			  done;
			  Bsml.stop_timing();
			  ignore(<< $cost$:=( $Bsml.get_cost()$ +. ! ($cost$)) >>);
			  ignore (<< if ( $this$ = 0 ) then (print_float ! $cost$ ; print_string ";")else () >> );
			  let word_vector= << let stat =Gc.stat() in stat.Gc.minor_words +.  stat.Gc.major_words -. stat.Gc.promoted_words>> in
			  let minor_collection_vector= << let stat =Gc.stat() in stat.Gc.minor_collections>> in
			  ignore (<< if ( $this$ = 0 ) then ( print_float $word_vector$; print_string ";";print_int ($minor_collection_vector$ - $start_minor_collection_vector$); print_string ";") else () >> );
			  Bsml.start_timing();
			  Gc.full_major();
			  Bsml.stop_timing();
			    ignore( << $gc_cost$ := $Bsml.get_cost()$ +.  ! $gc_cost$>>);
			    ignore (<< if ( $this$ = 0 ) then ( print_float ! $gc_cost$ )  else () >> )
  end   
