open Datatypes
open List0
open Bh_classes_opt
open Bounded_set
open Map_around
open Spec2bh

type __ = Obj.t

module type TOWER_PARAMETERS = 
 sig 
  type coq_A 
  
  val eq_A_dec : coq_A -> coq_A -> bool
  
  val coq_R_dec : coq_A -> coq_A -> bool
  
  val coq_AngleL : coq_A -> coq_A -> coq_A -> coq_A -> coq_A
  
  val coq_AngleR : coq_A -> coq_A -> coq_A -> coq_A -> coq_A
  
  val preprocess_L_x : coq_A -> coq_A
  
  val preprocess_L_h : coq_A -> coq_A
  
  val preprocess_R_x : coq_A -> coq_A
  
  val preprocess_R_h : coq_A -> coq_A
 end

module Tower = 
 functor (Bs:sig 
  val max_pid : nat
  
  val bsp_p : nat
  
  type processor = nat
  
  val nat_of_processor : processor -> nat
  
  val processor_of_nat : nat -> processor
  
  val processor0 : processor
  
  val processor_max : processor
  
  type 'x par 
  
  val get : 'a1 par -> processor -> 'a1
  
  val mkpar_specification : (processor -> 'a1) -> 'a1 par
  
  val apply_specification : ('a1 -> 'a2) par -> 'a1 par -> 'a2 par
  
  val put_specification : (processor -> 'a1) par -> (processor -> 'a1) par
  
  val proj_specification : 'a1 par -> (processor -> 'a1)
 end) ->
 functor (Coq_tp:sig 
  type coq_A 
  
  val eq_A_dec : coq_A -> coq_A -> bool
  
  val coq_R_dec : coq_A -> coq_A -> bool
  
  val coq_AngleL : coq_A -> coq_A -> coq_A -> coq_A -> coq_A
  
  val coq_AngleR : coq_A -> coq_A -> coq_A -> coq_A -> coq_A
  
  val preprocess_L_x : coq_A -> coq_A
  
  val preprocess_L_h : coq_A -> coq_A
  
  val preprocess_R_x : coq_A -> coq_A
  
  val preprocess_R_h : coq_A -> coq_A
 end) ->
 struct 
  module Bc = Bh_classes(Bs)
  
  module Map_around = Map_around(Bc)
  
  module BoundedSigInst = 
   struct 
    type coq_A = Coq_tp.coq_A
   end
  
  module BoundedA = BoundedSets(BoundedSigInst)
  
  (** val coq_Max :
      BoundedA.coq_BoundedSet -> BoundedA.coq_BoundedSet ->
      BoundedA.coq_BoundedSet **)
  
  let coq_Max a b =
    if BoundedA.coq_BSOrder_dec Coq_tp.coq_R_dec a b then b else a
  
  (** val maxAngleL :
      BoundedSigInst.coq_A -> BoundedSigInst.coq_A -> (BoundedSigInst.coq_A,
      BoundedSigInst.coq_A) prod list -> BoundedA.coq_BoundedSet **)
  
  let rec maxAngleL xL hL = function
    | Coq_nil -> BoundedA.Min
    | Coq_cons (p, l) ->
        let Coq_pair (x, h) = p in
        coq_Max (BoundedA.Elmt (Coq_tp.coq_AngleL xL hL x h))
          (maxAngleL xL hL l)
  
  (** val maxAngleR :
      BoundedSigInst.coq_A -> BoundedSigInst.coq_A -> (BoundedSigInst.coq_A,
      BoundedSigInst.coq_A) prod list -> BoundedA.coq_BoundedSet **)
  
  let rec maxAngleR xR hR = function
    | Coq_nil -> BoundedA.Min
    | Coq_cons (p, l) ->
        let Coq_pair (x, h) = p in
        coq_Max (BoundedA.Elmt (Coq_tp.coq_AngleR xR hR x h))
          (maxAngleR xR hR l)
  
  (** val visibleL :
      BoundedSigInst.coq_A -> BoundedSigInst.coq_A -> (BoundedSigInst.coq_A,
      BoundedSigInst.coq_A) prod list -> BoundedSigInst.coq_A ->
      BoundedSigInst.coq_A -> bool **)
  
  let visibleL xL hL xs xi hi =
    if BoundedA.coq_BSOrder_dec Coq_tp.coq_R_dec (maxAngleL xL hL xs)
         (BoundedA.Elmt
         (Coq_tp.coq_AngleL xL hL (Coq_tp.preprocess_L_x xi)
           (Coq_tp.preprocess_L_h hi)))
    then true
    else false
  
  (** val visibleR :
      BoundedSigInst.coq_A -> BoundedSigInst.coq_A -> (BoundedSigInst.coq_A,
      BoundedSigInst.coq_A) prod list -> BoundedSigInst.coq_A ->
      BoundedSigInst.coq_A -> bool **)
  
  let visibleR xR hR xs xi hi =
    if BoundedA.coq_BSOrder_dec Coq_tp.coq_R_dec (maxAngleR xR hR xs)
         (BoundedA.Elmt
         (Coq_tp.coq_AngleR xR hR (Coq_tp.preprocess_R_x xi)
           (Coq_tp.preprocess_R_h hi)))
    then true
    else false
  
  (** val visibleLR :
      BoundedSigInst.coq_A -> BoundedSigInst.coq_A -> BoundedSigInst.coq_A ->
      BoundedSigInst.coq_A -> (((BoundedSigInst.coq_A, BoundedSigInst.coq_A)
      prod list, (BoundedSigInst.coq_A, BoundedSigInst.coq_A) prod) prod,
      (BoundedSigInst.coq_A, BoundedSigInst.coq_A) prod list) prod -> bool **)
  
  let visibleLR xL hL xR hR = function
    | Coq_pair (lsxh, rs) ->
        let Coq_pair (ls, xh) = lsxh in
        let Coq_pair (x, h) = xh in
        if visibleL xL hL ls x h then visibleR xR hR rs x h else false
  
  (** val tower :
      BoundedSigInst.coq_A -> BoundedSigInst.coq_A -> BoundedSigInst.coq_A ->
      BoundedSigInst.coq_A -> (BoundedSigInst.coq_A, BoundedSigInst.coq_A)
      prod list -> bool list **)
  
  let tower xL hL xR hR xs =
    mapAround (visibleLR xL hL xR hR) xs
  
  (** val coq_Parallel_tower :
      BoundedSigInst.coq_A -> BoundedSigInst.coq_A -> BoundedSigInst.coq_A ->
      BoundedSigInst.coq_A -> (BoundedSigInst.coq_A, BoundedSigInst.coq_A)
      prod list Bc.par -> bool list Bs.par **)
  
  let coq_Parallel_tower xL hL xR hR l =
    let comms =
      Bc.Bsml_bh.bsml_comp_bh_comm (maxAngleL xL hL) 
        (maxAngleR xR hR) BoundedA.Min l BoundedA.Min
    in
    Bc.Bsml_bh.Partition.Bsml_prop.apply
      (Bc.Bsml_bh.Partition.Bsml_prop.apply
        (Bc.Bsml_bh.Partition.Bsml_prop.apply
          (Bc.Bsml_bh.Partition.Bsml_prop.replicate (fun x x0 x1 ->
            Bc.Bsml_bh.bh_seq (fun maxl xh maxr ->
              let Coq_pair (xi, hi) = xh in
              if BoundedA.coq_BSOrder_dec Coq_tp.coq_R_dec maxl
                   (BoundedA.Elmt
                   (Coq_tp.coq_AngleL xL hL (Coq_tp.preprocess_L_x xi)
                     (Coq_tp.preprocess_L_h hi)))
              then if BoundedA.coq_BSOrder_dec Coq_tp.coq_R_dec maxr
                        (BoundedA.Elmt
                        (Coq_tp.coq_AngleR xR hR (Coq_tp.preprocess_R_x xi)
                          (Coq_tp.preprocess_R_h hi)))
                   then true
                   else false
              else false) (maxAngleL xL hL) coq_Max 
              (maxAngleR xR hR) coq_Max x x0 x1))
          (Bc.Bsml_bh.loc_lefts coq_Max BoundedA.Min l BoundedA.Min comms))
        l) (Bc.Bsml_bh.loc_rights coq_Max BoundedA.Min l BoundedA.Min comms)
 end

