open Bsml
open Stdlib.Base
open List

let map3 f l1 l2 l3 = 
  let rec aux acc = function
    |[], _ , _  -> acc
    | _, [], _  -> acc
    | _, _ , [] -> acc
    |h1::t1,h2::t2,h3::t3->
       aux ((f h1 h2 h3)::acc) (t1,t2,t3) in
    aux [] (l1,l2,l3)

let sequence n1 n2 = 
  let rec aux acc n2 = 
    if (n1>n2) then acc else aux (n2::acc) (n2-1) in
  aux [] n2

let compose f g x = f (g x)

let procs_left  = mkpar(fun i->sequence 0 (i - 1))

let procs_right = mkpar(fun i->sequence (i + 1) (bsp_p - 1))

let tower_building (xl,hl) l (xr,hr) htower =
  let tan_l (x,h) = (h -. hl) /. (x -. xl) in
  let tan_r (x,h) = (h -. hr) /. (xr -. x) in
  let maxtans_l =
    fold_left (fun (m,acc) p -> 
		 let (x,h ) =  p in 
		 let visible =  (m  < tan_l (x,h+htower))in
		 let m = max (tan_l p) m in 
		   m, visible :: acc) (neg_infinity,[]) l in
  let maxtans_r =
    let inverse = fold_right (fun p (m,acc) -> 
				let (x,h ) =  p in 
				let visible =  (m  < tan_r (x, h+htower)) in
				let m = max (tan_r p) m in 
				  m, visible :: acc) l (neg_infinity,[]) in
      rev inverse
  in
    map2 (fun (ml,vl) (mr,vr) => vl && vr) maxtans_r maxtans_l
